@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/include/ldap_pvt.h,v 1.82.2.1 2005/01/20 18:03:48 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/include/ldap_pvt.h,v 1.91 2006/02/10 07:21:10 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  * 
- * Copyright 1998-2005 The OpenLDAP Foundation.
+ * Copyright 1998-2006 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -172,6 +172,9 @@ struct sasl_security_properties; /* avoid pulling in <sasl.h> */
 LDAP_F (int) ldap_pvt_sasl_secprops LDAP_P((
 	const char *in,
 	struct sasl_security_properties *secprops ));
+LDAP_F (void) ldap_pvt_sasl_secprops_unparse LDAP_P((
+	struct sasl_security_properties *secprops,
+	struct berval *out ));
 
 LDAP_F (void *) ldap_pvt_sasl_mutex_new LDAP_P((void));
 LDAP_F (int) ldap_pvt_sasl_mutex_lock LDAP_P((void *mutex));
@@ -209,6 +212,13 @@ ldap_pvt_find_wildcard LDAP_P((	const char *s ));
 LDAP_F( ber_slen_t )
 ldap_pvt_filter_value_unescape LDAP_P(( char *filter ));
 
+LDAP_F( ber_len_t )
+ldap_bv2escaped_filter_value_len LDAP_P(( struct berval *in ));
+
+LDAP_F( int )
+ldap_bv2escaped_filter_value_x LDAP_P(( struct berval *in, struct berval *out,
+	int inplace, void *ctx ));
+
 /* string.c */
 LDAP_F( char * )
 ldap_pvt_str2upper LDAP_P(( char *str ));
@@ -232,7 +242,7 @@ LDAP_F (int) ldap_pvt_tls_set_option LDAP_P(( struct ldap *ld,
 
 LDAP_F (void) ldap_pvt_tls_destroy LDAP_P(( void ));
 LDAP_F (int) ldap_pvt_tls_init LDAP_P(( void ));
-LDAP_F (int) ldap_pvt_tls_init_def_ctx LDAP_P(( void ));
+LDAP_F (int) ldap_pvt_tls_init_def_ctx LDAP_P(( int is_server ));
 LDAP_F (int) ldap_pvt_tls_accept LDAP_P(( Sockbuf *sb, void *ctx_arg ));
 LDAP_F (int) ldap_pvt_tls_inplace LDAP_P(( Sockbuf *sb ));
 LDAP_F (void *) ldap_pvt_tls_sb_ctx LDAP_P(( Sockbuf *sb ));
@@ -256,29 +266,25 @@ LDAP_END_DECL
  * Multiple precision stuff
  * 
  * May use OpenSSL's BIGNUM if built with TLS,
- * or GNU's multiple precision library.
+ * or GNU's multiple precision library. But if
+ * long long is available, that's big enough
+ * and much more efficient.
  *
  * If none is available, unsigned long data is used.
  */
-#ifdef HAVE_BIGNUM
+
+#if USE_MP_BIGNUM
 /*
  * Use OpenSSL's BIGNUM
  */
-#if defined(HAVE_OPENSSL_CRYPTO_H)
 #include <openssl/crypto.h>
-#elif HAVE_CRYPTO_H
-#include <crypto.h>
-#endif /* HAVE_OPENSSL_CRYPTO_H || HAVE_CRYPTO_H */
-#ifdef HAVE_OPENSSL_BN_H
 #include <openssl/bn.h>
-#elif HAVE_BN_H
-#include <bn.h>
-#endif /* HAVE_OPENSSL_BN_H || HAVE_BN_H */
 
-typedef	BIGNUM*		ldap_pvt_mp_t;
+typedef	BIGNUM* ldap_pvt_mp_t;
+#define	LDAP_PVT_MP_INIT	(NULL)
 
 #define	ldap_pvt_mp_init(mp) \
-	(mp) = BN_new()
+	do { (mp) = BN_new(); } while (0)
 
 /* FIXME: we rely on mpr being initialized */
 #define	ldap_pvt_mp_init_set(mpr,mpv) \
@@ -293,15 +299,15 @@ typedef	BIGNUM*		ldap_pvt_mp_t;
 #define ldap_pvt_mp_clear(mp) \
 	do { BN_free((mp)); (mp) = 0; } while (0)
 
-#elif defined(HAVE_GMP)
+#elif USE_MP_GMP
 /*
  * Use GNU's multiple precision library
  */
-#ifdef HAVE_GMP_H
 #include <gmp.h>
-#endif
 
 typedef mpz_t		ldap_pvt_mp_t;
+#define	LDAP_PVT_MP_INIT	{ 0 }
+
 #define ldap_pvt_mp_init(mp) \
 	mpz_init((mp))
 
@@ -317,30 +323,42 @@ typedef mpz_t		ldap_pvt_mp_t;
 #define ldap_pvt_mp_clear(mp) \
 	mpz_clear((mp))
 
-#else /* ! HAVE_BIGNUM && ! HAVE_GMP */
+#else
 /*
- * Use unsigned long
+ * Use unsigned long long
  */
 
-typedef	unsigned long	ldap_pvt_mp_t;
+#if USE_MP_LONG_LONG
+typedef	unsigned long long	ldap_pvt_mp_t;
+#define	LDAP_PVT_MP_INIT	(0LL)
+#elif USE_MP_LONG
+typedef	unsigned long		ldap_pvt_mp_t;
+#define	LDAP_PVT_MP_INIT	(0L)
+#elif HAVE_LONG_LONG
+typedef	unsigned long long	ldap_pvt_mp_t;
+#define	LDAP_PVT_MP_INIT	(0LL)
+#else
+typedef	unsigned long		ldap_pvt_mp_t;
+#define	LDAP_PVT_MP_INIT	(0L)
+#endif
 
 #define ldap_pvt_mp_init(mp) \
-	(mp) = 0
+	do { (mp) = 0; } while (0)
 
 #define	ldap_pvt_mp_init_set(mpr,mpv) \
-	(mpr) = (mpv)
+	do { (mpr) = (mpv); } while (0)
 
 #define	ldap_pvt_mp_add(mpr,mpv) \
-	(mpr) += (mpv)
+	do { (mpr) += (mpv); } while (0)
 
 #define	ldap_pvt_mp_add_ulong(mp,v) \
-	(mp) += (v)
+	do { (mp) += (v); } while (0)
 
 #define ldap_pvt_mp_clear(mp) \
-	(mp) = 0
+	do { (mp) = 0; } while (0)
 
-#endif /* ! HAVE_BIGNUM && ! HAVE_GMP */
+#endif /* MP */
 
 #include "ldap_pvt_uc.h"
 
-#endif
+#endif /* _LDAP_PVT_H */
