@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/include/ldap_rq.h,v 1.9.2.1 2005/01/20 18:03:48 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/include/ldap_rq.h,v 1.14 2006/01/03 22:12:06 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2005 The OpenLDAP Foundation.
+ * Copyright 1998-2006 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -27,6 +27,8 @@ typedef struct re_s {
 	LDAP_STAILQ_ENTRY(re_s) rnext;
 	ldap_pvt_thread_start_t *routine;
 	void *arg;
+	char *tname;
+	char *tspec;
 } re_t;
 
 typedef struct runqueue_s {
@@ -35,11 +37,20 @@ typedef struct runqueue_s {
 	ldap_pvt_thread_mutex_t	rq_mutex;
 } runqueue_t;
 
-LDAP_F( void )
+LDAP_F( struct re_s* )
 ldap_pvt_runqueue_insert(
 	struct runqueue_s* rq,
 	time_t interval,
 	ldap_pvt_thread_start_t* routine,
+	void *arg,
+	char *tname,
+	char *tspec
+);
+
+LDAP_F( struct re_s* )
+ldap_pvt_runqueue_find(
+	struct runqueue_s* rq,
+	ldap_pvt_thread_start_t* routine,
 	void *arg
 );
 
