@@ -1,8 +1,8 @@
 /* sockbuf.c - i/o routines with support for adding i/o layers. */
-/* $OpenLDAP: pkg/ldap/libraries/liblber/sockbuf.c,v 1.60 2004/06/30 20:28:15 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/libraries/liblber/sockbuf.c,v 1.65 2006/01/03 22:12:07 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2004 The OpenLDAP Foundation.
+ * Copyright 1998-2006 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -46,7 +46,7 @@
 #define LBER_MIN_BUFF_SIZE		4096
 #endif
 #ifndef LBER_MAX_BUFF_SIZE
-#define LBER_MAX_BUFF_SIZE		65536
+#define LBER_MAX_BUFF_SIZE		(65536*256)
 #endif
 #ifndef LBER_DEFAULT_READAHEAD
 #define LBER_DEFAULT_READAHEAD	16384
@@ -57,8 +57,6 @@ ber_sockbuf_alloc( void )
 {
 	Sockbuf			*sb;
 
-	ber_int_options.lbo_valid = LBER_INITIALIZED;
-
 	sb = LBER_CALLOC( 1, sizeof( Sockbuf ) );
 
 	if( sb == NULL ) return NULL;
@@ -104,13 +102,13 @@ ber_sockbuf_ctrl( Sockbuf *sb, int opt, void *arg )
 
 		case LBER_SB_OPT_GET_FD:
 			if ( arg != NULL ) {
-				*((int *)arg) = sb->sb_fd;
+				*((ber_socket_t *)arg) = sb->sb_fd;
 			}
 			ret = ( sb->sb_fd == AC_SOCKET_INVALID ? -1 : 1);
 			break;
 
 		case LBER_SB_OPT_SET_FD:
-			sb->sb_fd = *((int *)arg);
+			sb->sb_fd = *((ber_socket_t *)arg);
 			ret = 1;
 			break;
 
@@ -911,6 +909,7 @@ sb_dgram_read( Sockbuf_IO_Desc *sbiod, void *buf, ber_len_t len )
 	addrlen = sizeof( struct sockaddr );
 	src = buf;
 	buf += addrlen;
+	len -= addrlen;
 	rc = recvfrom( sbiod->sbiod_sb->sb_fd, buf, len, 0, src, &addrlen );
 
 	return rc > 0 ? rc+sizeof(struct sockaddr) : rc;
