@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/libraries/librewrite/subst.c,v 1.19 2004/03/10 18:21:39 ando Exp $ */
+/* $OpenLDAP: pkg/ldap/libraries/librewrite/subst.c,v 1.22 2006/01/03 22:12:11 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2000-2004 The OpenLDAP Foundation.
+ * Copyright 2000-2006 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -453,8 +453,8 @@ rewrite_subst_destroy(
 	int			n;
 	struct rewrite_subst	*subst;
 
-	assert( psubst );
-	assert( *psubst );
+	assert( psubst != NULL );
+	assert( *psubst != NULL );
 
 	subst = *psubst;
 
