@@ -1,8 +1,8 @@
 /* cache.c - routines to maintain an in-core cache of entries */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/cache.c,v 1.88.2.2 2005/01/20 18:03:58 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/cache.c,v 1.120 2006/04/05 05:17:11 hyc Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2000-2005 The OpenLDAP Foundation.
+ * Copyright 2000-2006 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -26,16 +26,19 @@
 
 #include "back-bdb.h"
 
+#include "ldap_rq.h"
+
 #ifdef BDB_HIER
 #define bdb_cache_lru_add	hdb_cache_lru_add
 #endif
+static void bdb_cache_lru_add( struct bdb_info *bdb, EntryInfo *ei );
 
 static int	bdb_cache_delete_internal(Cache *cache, EntryInfo *e, int decr);
 #ifdef LDAP_DEBUG
+#ifdef SLAPD_UNUSED
 static void	bdb_lru_print(Cache *cache);
 #endif
-
-static int bdb_txn_get( Operation *op, DB_ENV *env, DB_TXN **txn, int reset );
+#endif
 
 static EntryInfo *
 bdb_cache_entryinfo_new( Cache *cache )
@@ -183,8 +186,10 @@ bdb_cache_entryinfo_destroy( EntryInfo *e )
 	} \
 	(cache)->c_lruhead = (ei); \
 	(ei)->bei_lruprev = NULL; \
-	if ( (cache)->c_lrutail == NULL ) { \
-		(cache)->c_lrutail = (ei); \
+	if ( !ldap_pvt_thread_mutex_trylock( &(cache)->lru_tail_mutex )) { \
+		if ( (cache)->c_lrutail == NULL ) \
+			(cache)->c_lrutail = (ei); \
+		ldap_pvt_thread_mutex_unlock( &(cache)->lru_tail_mutex ); \
 	} \
 } while(0)
 
@@ -249,6 +254,13 @@ bdb_entryinfo_add_internal(
 	} else {
 		bdb->bi_cache.c_eiused++;
 		ber_dupbv( &ei2->bei_nrdn, &ei->bei_nrdn );
+
+		/* This is a new leaf node. But if parent had no kids, then it was
+		 * a leaf and we would be decrementing that. So, only increment if
+		 * the parent already has kids.
+		 */
+		if ( ei->bei_parent->bei_kids || !ei->bei_parent->bei_id )
+			bdb->bi_cache.c_leaves++;
 		avl_insert( &ei->bei_parent->bei_kids, ei2, bdb_rdn_cmp,
 			avl_dup_error );
 #ifdef BDB_HIER
@@ -308,6 +320,11 @@ bdb_cache_find_ndn(
 		if ( !ei2 ) {
 			int len = ei.bei_nrdn.bv_len;
 				
+			if ( BER_BVISEMPTY( ndn )) {
+				*res = eip;
+				return LDAP_SUCCESS;
+			}
+
 			ei.bei_nrdn.bv_len = ndn->bv_len -
 				(ei.bei_nrdn.bv_val - ndn->bv_val);
 			bdb_cache_entryinfo_unlock( eip );
@@ -364,26 +381,25 @@ bdb_cache_find_ndn(
 /* Walk up the tree from a child node, looking for an ID that's already
  * been linked into the cache.
  */
-static int
+int
 hdb_cache_find_parent(
 	Operation *op,
 	DB_TXN *txn,
+	u_int32_t	locker,
 	ID id,
 	EntryInfo **res )
 {
 	struct bdb_info *bdb = (struct bdb_info *) op->o_bd->be_private;
 	EntryInfo ei, eip, *ei2 = NULL, *ein = NULL, *eir = NULL;
-	char ndn[SLAP_LDAPDN_MAXLEN];
-	ID parent;
 	int rc;
-	int addlru = 1;
+	int addlru = 0;
 
 	ei.bei_id = id;
 	ei.bei_kids = NULL;
 	ei.bei_ckids = 0;
 
 	for (;;) {
-		rc = hdb_dn2id_parent( op, txn, &ei, &eip.bei_id );
+		rc = hdb_dn2id_parent( op, txn, locker, &ei, &eip.bei_id );
 		if ( rc ) break;
 
 		/* Save the previous node, if any */
@@ -424,10 +440,8 @@ hdb_cache_find_parent(
 				ein->bei_ckids++;
 				bdb_cache_entryinfo_unlock( ein );
 			}
+			addlru = 0;
 
-			if ( !eir ) {
-				addlru = 0;
-			}
 		}
 
 		/* If this is the first time, save this node
@@ -446,8 +460,16 @@ hdb_cache_find_parent(
 			ei2 = &bdb->bi_cache.c_dntree;
 		}
 		bdb->bi_cache.c_eiused++;
+		if ( ei2 && ( ei2->bei_kids || !ei2->bei_id ))
+				bdb->bi_cache.c_leaves++;
 		ldap_pvt_thread_rdwr_wunlock( &bdb->bi_cache.c_rwlock );
 
+		if ( addlru ) {
+			ldap_pvt_thread_mutex_lock( &bdb->bi_cache.lru_head_mutex );
+			bdb_cache_lru_add( bdb, ein );
+		}
+		addlru = 1;
+
 		/* Got the parent, link in and we're done. */
 		if ( ei2 ) {
 			bdb_cache_entryinfo_lock( ei2 );
@@ -509,53 +531,55 @@ int hdb_cache_load(
 }
 #endif
 
-/* caller must have lru_mutex locked. mutex
+/* caller must have lru_head_mutex locked. mutex
  * will be unlocked on return.
  */
 static void
 bdb_cache_lru_add(
 	struct bdb_info *bdb,
-	u_int32_t	locker,
 	EntryInfo *ei )
 {
 	DB_LOCK		lock, *lockp;
+	EntryInfo *elru, *elprev;
+	int count = 0;
 
-	if ( locker ) {
+	LRU_ADD( &bdb->bi_cache, ei );
+	ldap_pvt_thread_mutex_unlock( &bdb->bi_cache.lru_head_mutex );
+
+	/* See if we're above the cache size limit */
+	if ( bdb->bi_cache.c_cursize <= bdb->bi_cache.c_maxsize )
+		return;
+
+	if ( bdb->bi_cache.c_locker ) {
 		lockp = &lock;
 	} else {
 		lockp = NULL;
 	}
 
-	/* See if we're above the cache size limit */
-	if ( bdb->bi_cache.c_cursize > bdb->bi_cache.c_maxsize ) {
-		EntryInfo *elru, *elprev;
-		int i = 0;
+	ldap_pvt_thread_mutex_lock( &bdb->bi_cache.lru_tail_mutex );
 
-		/* Look for an unused entry to remove */
-		for (elru = bdb->bi_cache.c_lrutail; elru; elru = elprev, i++ ) {
-			elprev = elru->bei_lruprev;
+	/* Look for an unused entry to remove */
+	for (elru = bdb->bi_cache.c_lrutail; elru; elru = elprev ) {
+		elprev = elru->bei_lruprev;
 
-			/* Too many probes, not enough idle, give up */
-			if (i > 10)
-				break;
+		/* If we can successfully writelock it, then
+		 * the object is idle.
+		 */
+		if ( bdb_cache_entry_db_lock( bdb->bi_dbenv,
+				bdb->bi_cache.c_locker, elru, 1, 1, lockp ) == 0 ) {
 
-			/* If we can successfully writelock it, then
-			 * the object is idle.
+			int stop = 0;
+
+			/* If this node is in the process of linking into the cache,
+			 * or this node is being deleted, skip it.
 			 */
-			if ( bdb_cache_entry_db_lock( bdb->bi_dbenv,
-					bdb->bi_cache.c_locker, elru, 1, 1, lockp ) == 0 ) {
-				int stop = 0;
-
-				/* If there's no entry, or this node is in
-				 * the process of linking into the cache,
-				 * or this node is being deleted, skip it.
-				 */
-				if ( !elru->bei_e || (elru->bei_state &
-					( CACHE_ENTRY_NOT_LINKED | CACHE_ENTRY_DELETED ))) {
-					bdb_cache_entry_db_unlock( bdb->bi_dbenv, lockp );
-					continue;
-				}
-				LRU_DELETE( &bdb->bi_cache, elru );
+			if ( elru->bei_state &
+				( CACHE_ENTRY_NOT_LINKED | CACHE_ENTRY_DELETED )) {
+				bdb_cache_entry_db_unlock( bdb->bi_dbenv, lockp );
+				continue;
+			}
+			/* Free entry for this node if it's present */
+			if ( elru->bei_e ) {
 				elru->bei_e->e_private = NULL;
 #ifdef SLAP_ZONE_ALLOC
 				bdb_entry_return( bdb, elru->bei_e, elru->bei_zseq );
@@ -563,18 +587,41 @@ bdb_cache_lru_add(
 				bdb_entry_return( elru->bei_e );
 #endif
 				elru->bei_e = NULL;
-				ldap_pvt_thread_rdwr_wlock( &bdb->bi_cache.c_rwlock );
-				--bdb->bi_cache.c_cursize;
-				if (bdb->bi_cache.c_cursize <= bdb->bi_cache.c_maxsize)
-					stop = 1;
-				ldap_pvt_thread_rdwr_wunlock( &bdb->bi_cache.c_rwlock );
-				bdb_cache_entry_db_unlock( bdb->bi_dbenv, lockp );
-				if (stop) break;
+				count++;
 			}
+			/* ITS#4010 if we're in slapcat, and this node is a leaf
+			 * node, free it.
+			 *
+			 * FIXME: we need to do this for slapd as well, (which is
+			 * why we compute bi_cache.c_leaves now) but at the moment
+			 * we can't because it causes unresolvable deadlocks. 
+			 */
+			if ( slapMode & SLAP_TOOL_READONLY ) {
+				if ( !elru->bei_kids ) {
+					/* This does LRU_DELETE for us */
+					bdb_cache_delete_internal( &bdb->bi_cache, elru, 0 );
+					bdb_cache_delete_cleanup( &bdb->bi_cache, elru );
+				}
+				/* Leave node on LRU list for a future pass */
+			} else {
+				LRU_DELETE( &bdb->bi_cache, elru );
+			}
+			bdb_cache_entry_db_unlock( bdb->bi_dbenv, lockp );
+
+			if ( count == bdb->bi_cache.c_minfree ) {
+				ldap_pvt_thread_rdwr_wlock( &bdb->bi_cache.c_rwlock );
+				bdb->bi_cache.c_cursize -= bdb->bi_cache.c_minfree;
+				if ( bdb->bi_cache.c_maxsize - bdb->bi_cache.c_cursize >=
+					bdb->bi_cache.c_minfree )
+					stop = 1;
+				count = 0;
+				ldap_pvt_thread_rdwr_wunlock( &bdb->bi_cache.c_rwlock );
+			}
+			if (stop) break;
 		}
 	}
-	LRU_ADD( &bdb->bi_cache, ei );
-	ldap_pvt_thread_mutex_unlock( &bdb->bi_cache.lru_mutex );
+
+	ldap_pvt_thread_mutex_unlock( &bdb->bi_cache.lru_tail_mutex );
 }
 
 EntryInfo *
@@ -657,7 +704,7 @@ again:	ldap_pvt_thread_rdwr_rlock( &bdb->bi_cache.c_rwlock );
 	/* See if the ID exists in the database; add it to the cache if so */
 	if ( !*eip ) {
 #ifndef BDB_HIER
-		rc = bdb_id2entry( op->o_bd, tid, id, &ep );
+		rc = bdb_id2entry( op->o_bd, tid, locker, id, &ep );
 		if ( rc == 0 ) {
 			rc = bdb_cache_find_ndn( op, tid,
 				&ep->e_nname, eip );
@@ -672,13 +719,13 @@ again:	ldap_pvt_thread_rdwr_rlock( &bdb->bi_cache.c_rwlock );
 			}
 		}
 #else
-		rc = hdb_cache_find_parent(op, tid, id, eip );
-		if ( rc == 0 && *eip ) islocked = 1;
+		rc = hdb_cache_find_parent(op, tid, locker, id, eip );
+		if ( rc == 0 ) islocked = 1;
 #endif
 	}
 
 	/* Ok, we found the info, do we have the entry? */
-	if ( *eip && rc == 0 ) {
+	if ( rc == 0 ) {
 		if ( (*eip)->bei_state & CACHE_ENTRY_DELETED ) {
 			rc = DB_NOTFOUND;
 		} else {
@@ -705,31 +752,14 @@ load1:
 				bdb_cache_entry_db_unlock( bdb->bi_dbenv, lock );
 			} else if ( rc == 0 ) {
 				if ( load ) {
-					DB_TXN *ltid;
-					u_int32_t locker2 = locker;
-
-					/* We don't wrap entire read operations in txn's, but
-					 * we need our cache entry lock and any DB page locks
-					 * to be associated, in order for deadlock detection
-					 * to work properly. So if we need to read from the DB,
-					 * we use a long-lived per-thread txn for this step.
-					 */
-					if ( !ep && !tid ) {
-						rc = bdb_txn_get( op, bdb->bi_dbenv, &ltid, 0 );
-						if ( ltid )
-							locker2 = TXN_ID( ltid );
-					} else {
-						ltid = tid;
-					}
-					/* Give up original read lock, obtain write lock with
-					 * (possibly) new locker ID.
+					/* Give up original read lock, obtain write lock
 					 */
 				    if ( rc == 0 ) {
-						rc = bdb_cache_entry_db_relock( bdb->bi_dbenv, locker2,
+						rc = bdb_cache_entry_db_relock( bdb->bi_dbenv, locker,
 							*eip, 1, 0, lock );
 					}
 					if ( rc == 0 && !ep) {
-						rc = bdb_id2entry( op->o_bd, ltid, id, &ep );
+						rc = bdb_id2entry( op->o_bd, tid, locker, id, &ep );
 					}
 					if ( rc == 0 ) {
 						ep->e_private = *eip;
@@ -751,22 +781,6 @@ load1:
 						/* Otherwise, release the lock. */
 						bdb_cache_entry_db_unlock( bdb->bi_dbenv, lock );
 					}
-					if ( locker2 != locker ) {
-						/* If we're using the per-thread txn, release all
-						 * of its page locks now.
-						 */
-						DB_LOCKREQ list;
-						list.op = DB_LOCK_PUT_ALL;
-						list.obj = NULL;
-						bdb->bi_dbenv->lock_vec( bdb->bi_dbenv, locker2,
-							0, &list, 1, NULL );
-						/* If this txn was deadlocked, we must abort it
-						 * and invalidate this per-thread txn.
-						 */
-						if ( rc == DB_LOCK_DEADLOCK ) {
-							bdb_txn_get( op, bdb->bi_dbenv, &ltid, 1 );
-						}
-					}
 				} else if ( !(*eip)->bei_e ) {
 					/* Some other thread is trying to load the entry,
 					 * give it a chance to finish.
@@ -814,22 +828,23 @@ load1:
 			ldap_pvt_thread_rdwr_wunlock( &bdb->bi_cache.c_rwlock );
 		}
 
-		ldap_pvt_thread_mutex_lock( &bdb->bi_cache.lru_mutex );
+		ldap_pvt_thread_mutex_lock( &bdb->bi_cache.lru_head_mutex );
 
 		/* If the LRU list has only one entry and this is it, it
 		 * doesn't need to be added again.
 		 */
 		if ( bdb->bi_cache.c_lruhead == bdb->bi_cache.c_lrutail &&
 			bdb->bi_cache.c_lruhead == *eip ) {
-			ldap_pvt_thread_mutex_unlock( &bdb->bi_cache.lru_mutex );
+			ldap_pvt_thread_mutex_unlock( &bdb->bi_cache.lru_head_mutex );
 		} else {
-
 			/* if entry is on LRU list, remove from old spot */
 			if ( (*eip)->bei_lrunext || (*eip)->bei_lruprev ) {
+				ldap_pvt_thread_mutex_lock( &bdb->bi_cache.lru_tail_mutex );
 				LRU_DELETE( &bdb->bi_cache, *eip );
+				ldap_pvt_thread_mutex_unlock( &bdb->bi_cache.lru_tail_mutex );
 			}
-			/* lru_mutex is unlocked for us */
-			bdb_cache_lru_add( bdb, locker, *eip );
+			/* lru_head_mutex is unlocked for us */
+			bdb_cache_lru_add( bdb, *eip );
 		}
 	}
 
@@ -873,9 +888,11 @@ bdb_cache_add(
 	u_int32_t locker )
 {
 	EntryInfo *new, ei;
-	struct berval rdn = e->e_name;
 	DB_LOCK lock;
 	int rc;
+#ifdef BDB_HIER
+	struct berval rdn = e->e_name;
+#endif
 
 	ei.bei_id = e->e_id;
 	ei.bei_parent = eip;
@@ -893,7 +910,8 @@ bdb_cache_add(
 
 #ifdef BDB_HIER
 	if ( nrdn->bv_len != e->e_nname.bv_len ) {
-		char *ptr = strchr( rdn.bv_val, ',' );
+		char *ptr = ber_bvchr( &rdn, ',' );
+		assert( ptr != NULL );
 		rdn.bv_len = ptr - rdn.bv_val;
 	}
 	ber_dupbv( &ei.bei_rdn, &rdn );
@@ -923,10 +941,10 @@ bdb_cache_add(
 	ldap_pvt_thread_rdwr_wunlock( &bdb->bi_cache.c_rwlock );
 
 	/* set lru mutex */
-	ldap_pvt_thread_mutex_lock( &bdb->bi_cache.lru_mutex );
+	ldap_pvt_thread_mutex_lock( &bdb->bi_cache.lru_head_mutex );
 
-	/* lru_mutex is unlocked for us */
-	bdb_cache_lru_add( bdb, locker, new );
+	/* lru_head_mutex is unlocked for us */
+	bdb_cache_lru_add( bdb, new );
 
 	return rc;
 }
@@ -961,20 +979,22 @@ bdb_cache_modify(
  */
 int
 bdb_cache_modrdn(
+	struct bdb_info *bdb,
 	Entry *e,
 	struct berval *nrdn,
 	Entry *new,
 	EntryInfo *ein,
-	DB_ENV *env,
 	u_int32_t locker,
 	DB_LOCK *lock )
 {
 	EntryInfo *ei = BEI(e), *pei;
-	struct berval rdn;
 	int rc;
+#ifdef BDB_HIER
+	struct berval rdn;
+#endif
 
 	/* Get write lock on data */
-	rc =  bdb_cache_entry_db_relock( env, locker, ei, 1, 0, lock );
+	rc =  bdb_cache_entry_db_relock( bdb->bi_dbenv, locker, ei, 1, 0, lock );
 	if ( rc ) return rc;
 
 	/* If we've done repeated mods on a cached entry, then e_attrs
@@ -1004,7 +1024,8 @@ bdb_cache_modrdn(
 
 	rdn = e->e_name;
 	if ( nrdn->bv_len != e->e_nname.bv_len ) {
-		char *ptr = strchr(rdn.bv_val, ',');
+		char *ptr = ber_bvchr(&rdn, ',');
+		assert( ptr != NULL );
 		rdn.bv_len = ptr - rdn.bv_val;
 	}
 	ber_dupbv( &ei->bei_rdn, &rdn );
@@ -1019,12 +1040,11 @@ bdb_cache_modrdn(
 	}
 #ifdef BDB_HIER
 	{
-		int max = ei->bei_modrdns;
 		/* Record the generation number of this change */
-		for ( pei = ein; pei->bei_parent; pei = pei->bei_parent ) {
-			if ( pei->bei_modrdns > max ) max = pei->bei_modrdns;
-		}
-		ei->bei_modrdns = max + 1;
+		ldap_pvt_thread_mutex_lock( &bdb->bi_modrdns_mutex );
+		bdb->bi_modrdns++;
+		ei->bei_modrdns = bdb->bi_modrdns;
+		ldap_pvt_thread_mutex_unlock( &bdb->bi_modrdns_mutex );
 	}
 #endif
 	avl_insert( &ein->bei_kids, ei, bdb_rdn_cmp, avl_dup_error );
@@ -1049,7 +1069,7 @@ bdb_cache_delete(
 	EntryInfo *ei = BEI(e);
 	int	rc;
 
-	assert( e->e_private );
+	assert( e->e_private != NULL );
 
 	/* Set this early, warn off any queriers */
 	ei->bei_state |= CACHE_ENTRY_DELETED;
@@ -1070,10 +1090,18 @@ bdb_cache_delete(
 		e->e_id, 0, 0 );
 
 	/* set lru mutex */
-	ldap_pvt_thread_mutex_lock( &cache->lru_mutex );
+	ldap_pvt_thread_mutex_lock( &cache->lru_tail_mutex );
+
+	/* set cache write lock */
+	ldap_pvt_thread_rdwr_wlock( &cache->c_rwlock );
+
 	rc = bdb_cache_delete_internal( cache, e->e_private, 1 );
+
+	/* free cache write lock */
+	ldap_pvt_thread_rdwr_wunlock( &cache->c_rwlock );
+
 	/* free lru mutex */
-	ldap_pvt_thread_mutex_unlock( &cache->lru_mutex );
+	ldap_pvt_thread_mutex_unlock( &cache->lru_tail_mutex );
 
 	/* Leave entry info locked */
 
@@ -1123,9 +1151,6 @@ bdb_cache_delete_internal(
 {
 	int rc = 0;	/* return code */
 
-	/* set cache write lock */
-	ldap_pvt_thread_rdwr_wlock( &cache->c_rwlock );
-
 	/* Lock the parent's kids tree */
 	bdb_cache_entryinfo_lock( e->bei_parent );
 
@@ -1139,27 +1164,25 @@ bdb_cache_delete_internal(
 	{
 		rc = -1;
 	}
+	if ( e->bei_parent->bei_kids )
+		cache->c_leaves--;
 
 	/* id tree */
 	if ( avl_delete( &cache->c_idtree, (caddr_t) e, bdb_id_cmp ) == NULL ) {
 		rc = -1;
 	}
 
-	if (rc != 0) {
-		return rc;
+	if ( rc == 0 ){
+		cache->c_eiused--;
+
+		/* lru */
+		LRU_DELETE( cache, e );
+		if ( e->bei_e ) cache->c_cursize--;
 	}
 
-	cache->c_eiused--;
-
-	/* lru */
-	LRU_DELETE( cache, e );
-	if ( e->bei_e ) cache->c_cursize--;
-
-	/* free cache write lock */
-	ldap_pvt_thread_rdwr_wunlock( &cache->c_rwlock );
 	bdb_cache_entryinfo_unlock( e->bei_parent );
 
-	return( 0 );
+	return( rc );
 }
 
 static void
@@ -1186,7 +1209,7 @@ bdb_cache_release_all( Cache *cache )
 	/* set cache write lock */
 	ldap_pvt_thread_rdwr_wlock( &cache->c_rwlock );
 	/* set lru mutex */
-	ldap_pvt_thread_mutex_lock( &cache->lru_mutex );
+	ldap_pvt_thread_mutex_lock( &cache->lru_tail_mutex );
 
 	Debug( LDAP_DEBUG_TRACE, "====> bdb_cache_release_all\n", 0, 0, 0 );
 
@@ -1196,16 +1219,22 @@ bdb_cache_release_all( Cache *cache )
 		cache->c_lruhead = cache->c_eifree->bei_lrunext;
 		bdb_cache_entryinfo_destroy(cache->c_eifree);
 	}
+	cache->c_cursize = 0;
+	cache->c_eiused = 0;
+	cache->c_leaves = 0;
+	cache->c_idtree = NULL;
 	cache->c_lruhead = NULL;
 	cache->c_lrutail = NULL;
+	cache->c_dntree.bei_kids = NULL;
 
 	/* free lru mutex */
-	ldap_pvt_thread_mutex_unlock( &cache->lru_mutex );
+	ldap_pvt_thread_mutex_unlock( &cache->lru_tail_mutex );
 	/* free cache write lock */
 	ldap_pvt_thread_rdwr_wunlock( &cache->c_rwlock );
 }
 
 #ifdef LDAP_DEBUG
+#ifdef SLAPD_UNUSED
 static void
 bdb_lru_print( Cache *cache )
 {
@@ -1223,77 +1252,22 @@ bdb_lru_print( Cache *cache )
 	}
 }
 #endif
-
-static void
-bdb_txn_free( void *key, void *data )
-{
-	DB_TXN *txn = data;
-	TXN_ABORT( txn );
-}
-
-/* Obtain a long-lived transaction for the current thread.
- * If reset == 1, remove the current transaction. */
-static int
-bdb_txn_get( Operation *op, DB_ENV *env, DB_TXN **txn, int reset )
-{
-	int i, rc, lockid;
-	void *ctx, *data = NULL;
-
-	/* If no op was provided, try to find the ctx anyway... */
-	if ( op ) {
-		ctx = op->o_threadctx;
-	} else {
-		ctx = ldap_pvt_thread_pool_context();
-	}
-
-	/* Shouldn't happen unless we're single-threaded */
-	if ( !ctx ) {
-		*txn = NULL;
-		return 0;
-	}
-
-	if ( reset ) {
-		TXN_ABORT( *txn );
-		return ldap_pvt_thread_pool_setkey( ctx, ((char *)env)+1, NULL, NULL );
-	}
-
-	if ( ldap_pvt_thread_pool_getkey( ctx, ((char *)env)+1, &data, NULL ) ||
-		data == NULL ) {
-		for ( i=0, rc=1; rc != 0 && i<4; i++ ) {
-			rc = TXN_BEGIN( env, NULL, txn, 0 );
-			if (rc) ldap_pvt_thread_yield();
-		}
-		if ( rc != 0) {
-			return rc;
-		}
-		if ( ( rc = ldap_pvt_thread_pool_setkey( ctx, ((char *)env)+1,
-			*txn, bdb_txn_free ) ) ) {
-			TXN_ABORT( *txn );
-			Debug( LDAP_DEBUG_ANY, "bdb_txn_get: err %s(%d)\n",
-				db_strerror(rc), rc, 0 );
-
-			return rc;
-		}
-	} else {
-		*txn = data;
-	}
-	return 0;
-}
+#endif
 
 #ifdef BDB_REUSE_LOCKERS
 static void
 bdb_locker_id_free( void *key, void *data )
 {
 	DB_ENV *env = key;
-	int lockid = (int) data;
+	u_int32_t lockid = (long)data;
 	int rc;
 
 	rc = XLOCK_ID_FREE( env, lockid );
 	if ( rc == EINVAL ) {
 		DB_LOCKREQ lr;
 		Debug( LDAP_DEBUG_ANY,
-			"bdb_locker_id_free: %d err %s(%d)\n",
-			lockid, db_strerror(rc), rc );
+			"bdb_locker_id_free: %lu err %s(%d)\n",
+			(unsigned long) lockid, db_strerror(rc), rc );
 		/* release all locks held by this locker. */
 		lr.op = DB_LOCK_PUT_ALL;
 		lr.obj = NULL;
@@ -1303,9 +1277,10 @@ bdb_locker_id_free( void *key, void *data )
 }
 
 int
-bdb_locker_id( Operation *op, DB_ENV *env, int *locker )
+bdb_locker_id( Operation *op, DB_ENV *env, u_int32_t *locker )
 {
-	int i, rc, lockid;
+	int i, rc;
+	u_int32_t lockid;
 	void *data;
 	void *ctx;
 
@@ -1332,7 +1307,7 @@ bdb_locker_id( Operation *op, DB_ENV *env, int *locker )
 		if ( rc != 0) {
 			return rc;
 		}
-		data = (void *)lockid;
+		data = (void *)((long)lockid);
 		if ( ( rc = ldap_pvt_thread_pool_setkey( ctx, env,
 			data, bdb_locker_id_free ) ) ) {
 			XLOCK_ID_FREE( env, lockid );
@@ -1342,12 +1317,12 @@ bdb_locker_id( Operation *op, DB_ENV *env, int *locker )
 			return rc;
 		}
 	} else {
-		lockid = (int)data;
+		lockid = (long)data;
 	}
 	*locker = lockid;
 	return 0;
 }
-#endif
+#endif /* BDB_REUSE_LOCKERS */
 
 void
 bdb_cache_delete_entry(
