@@ -1,8 +1,8 @@
 /* init.c - initialize bdb backend */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/init.c,v 1.177.2.5 2005/02/02 17:32:44 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/init.c,v 1.247 2006/04/02 17:54:25 hallvard Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2000-2005 The OpenLDAP Foundation.
+ * Copyright 2000-2006 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -20,10 +20,12 @@
 #include <ac/string.h>
 #include <ac/unistd.h>
 #include <ac/stdlib.h>
-
+#include <ac/errno.h>
+#include <sys/stat.h>
 #include "back-bdb.h"
 #include <lutil.h>
 #include <ldap_rq.h>
+#include "alock.h"
 
 static const struct bdbi_database {
 	char *file;
@@ -36,8 +38,6 @@ static const struct bdbi_database {
 	{ NULL, NULL, 0, 0 }
 };
 
-struct berval bdb_uuid = BER_BVNULL;
-
 typedef void * db_malloc(size_t);
 typedef void * db_realloc(void *, size_t);
 
@@ -46,9 +46,9 @@ bdb_db_init( BackendDB *be )
 {
 	struct bdb_info	*bdb;
 
-	Debug( LDAP_DEBUG_ANY,
-		LDAP_XSTRING(bdb_db_init) ": Initializing "
-		BDB_UCTYPE " database\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE,
+		LDAP_XSTRING(bdb_db_init) ": Initializing " BDB_UCTYPE " database\n",
+		0, 0, 0 );
 
 	/* allocate backend-database-specific stuff */
 	bdb = (struct bdb_info *) ch_calloc( 1, sizeof(struct bdb_info) );
@@ -59,6 +59,7 @@ bdb_db_init( BackendDB *be )
 	bdb->bi_dbenv_mode = SLAPD_DEFAULT_DB_MODE;
 
 	bdb->bi_cache.c_maxsize = DEFAULT_CACHE_SIZE;
+	bdb->bi_cache.c_minfree = 1;
 
 	bdb->bi_lock_detect = DB_LOCK_DEFAULT;
 	bdb->bi_search_stack_depth = DEFAULT_SEARCH_STACK_DEPTH;
@@ -66,35 +67,43 @@ bdb_db_init( BackendDB *be )
 
 	ldap_pvt_thread_mutex_init( &bdb->bi_database_mutex );
 	ldap_pvt_thread_mutex_init( &bdb->bi_lastid_mutex );
-	ldap_pvt_thread_mutex_init( &bdb->bi_cache.lru_mutex );
+#ifdef BDB_HIER
+	ldap_pvt_thread_mutex_init( &bdb->bi_modrdns_mutex );
+#endif
+	ldap_pvt_thread_mutex_init( &bdb->bi_cache.lru_head_mutex );
+	ldap_pvt_thread_mutex_init( &bdb->bi_cache.lru_tail_mutex );
 	ldap_pvt_thread_mutex_init( &bdb->bi_cache.c_dntree.bei_kids_mutex );
 	ldap_pvt_thread_rdwr_init ( &bdb->bi_cache.c_rwlock );
+	ldap_pvt_thread_rdwr_init( &bdb->bi_idl_tree_rwlock );
+	ldap_pvt_thread_mutex_init( &bdb->bi_idl_tree_lrulock );
 
 	be->be_private = bdb;
+	be->be_cf_ocs = be->bd_info->bi_cf_ocs;
 
 	return 0;
 }
 
-static void *
-bdb_checkpoint( void *ctx, void *arg )
-{
-	struct re_s *rtask = arg;
-	struct bdb_info *bdb = rtask->arg;
-	
-	TXN_CHECKPOINT( bdb->bi_dbenv, bdb->bi_txn_cp_kbyte,
-		bdb->bi_txn_cp_min, 0 );
-	return NULL;
-}
+static int
+bdb_db_close( BackendDB *be );
 
 static int
 bdb_db_open( BackendDB *be )
 {
 	int rc, i;
 	struct bdb_info *bdb = (struct bdb_info *) be->be_private;
+	struct stat stat1, stat2;
 	u_int32_t flags;
-#ifdef HAVE_EBCDIC
 	char path[MAXPATHLEN];
-#endif
+	char *dbhome;
+	int do_recover = 0, do_alock_recover = 0;
+	int alockt, quick = 0;
+
+	if ( be->be_suffix == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"bdb_db_open: need suffix\n",
+			0, 0, 0 );
+		return -1;
+	}
 
 	Debug( LDAP_DEBUG_ARGS,
 		"bdb_db_open: %s\n",
@@ -107,116 +116,146 @@ bdb_db_open( BackendDB *be )
 		return -1;
 	}
 #endif
-	/* we should check existance of dbenv_home and db_directory */
+
+	/* Check existence of dbenv_home. Any error means trouble */
+	rc = stat( bdb->bi_dbenv_home, &stat1 );
+	if( rc !=0 ) {
+		Debug( LDAP_DEBUG_ANY,
+			"bdb_db_open: Cannot access database directory %s (%d)\n",
+			bdb->bi_dbenv_home, errno, 0 );
+			return -1;
+	}
+
+	/* Perform database use arbitration/recovery logic */
+	alockt = (slapMode & SLAP_TOOL_READONLY) ? ALOCK_LOCKED : ALOCK_UNIQUE;
+	if ( slapMode & SLAP_TOOL_QUICK ) {
+		alockt |= ALOCK_NOSAVE;
+		quick = 1;
+	}
+
+	rc = alock_open( &bdb->bi_alock_info, 
+				"slapd", 
+				bdb->bi_dbenv_home, alockt );
+
+	/* alockt is TRUE if the existing environment was created in Quick mode */
+	alockt = (rc & ALOCK_NOSAVE) ? 1 : 0;
+	rc &= ~ALOCK_NOSAVE;
+
+	if( rc == ALOCK_RECOVER ) {
+		Debug( LDAP_DEBUG_ANY,
+			"bdb_db_open: unclean shutdown detected;"
+			" attempting recovery.\n", 
+			0, 0, 0 );
+		do_alock_recover = 1;
+		do_recover = DB_RECOVER;
+	} else if( rc == ALOCK_BUSY ) {
+		Debug( LDAP_DEBUG_ANY,
+			"bdb_db_open: database already in use\n", 
+			0, 0, 0 );
+		return -1;
+	} else if( rc != ALOCK_CLEAN ) {
+		Debug( LDAP_DEBUG_ANY,
+			"bdb_db_open: alock package is unstable\n", 
+			0, 0, 0 );
+		return -1;
+	}
+
+	/*
+	 * The DB_CONFIG file may have changed. If so, recover the
+	 * database so that new settings are put into effect. Also
+	 * note the possible absence of DB_CONFIG in the log.
+	 */
+	if( stat( bdb->bi_db_config_path, &stat1 ) == 0 ) {
+		if ( !do_recover ) {
+			char *ptr = lutil_strcopy(path, bdb->bi_dbenv_home);
+			*ptr++ = LDAP_DIRSEP[0];
+			strcpy( ptr, "__db.001" );
+			if( stat( path, &stat2 ) == 0 ) {
+				if( stat2.st_mtime < stat1.st_mtime ) {
+					Debug( LDAP_DEBUG_ANY,
+						"bdb_db_open: DB_CONFIG for suffix %s has changed.\n"
+						"Performing database recovery to activate new settings.\n",
+						be->be_suffix[0].bv_val, 0, 0 );
+					do_recover = DB_RECOVER;
+				}
+			}
+		}
+	}
+	else {
+		Debug( LDAP_DEBUG_ANY,
+			"bdb_db_open: Warning - No DB_CONFIG file found "
+			"in directory %s: (%d)\n"
+			"Expect poor performance for suffix %s.\n",
+			bdb->bi_dbenv_home, errno, be->be_suffix[0].bv_val );
+	}
+
+	/* Always let slapcat run, regardless of environment state.
+	 * This can be used to cause a cache flush after an unclean
+	 * shutdown.
+	 */
+	if ( do_recover && ( slapMode & SLAP_TOOL_READONLY )) {
+		Debug( LDAP_DEBUG_ANY,
+			"bdb_db_open: Recovery skipped in read-only mode. "
+			"Run manual recovery if errors are encountered.\n",
+			0, 0, 0 );
+		do_recover = 0;
+		quick = alockt;
+	}
+
+	/* An existing environment in Quick mode has nothing to recover. */
+	if ( alockt && do_recover ) {
+		Debug( LDAP_DEBUG_ANY,
+			"bdb_db_open: cannot recover, database must be reinitialized.\n", 
+			0, 0, 0 );
+		rc = -1;
+		goto fail;
+	}
 
 	rc = db_env_create( &bdb->bi_dbenv, 0 );
 	if( rc != 0 ) {
 		Debug( LDAP_DEBUG_ANY,
 			"bdb_db_open: db_env_create failed: %s (%d)\n",
 			db_strerror(rc), rc, 0 );
-		return rc;
+		goto fail;
 	}
 
-	flags = DB_INIT_MPOOL | DB_THREAD | DB_CREATE;
-
-	if ( !( slapMode & SLAP_TOOL_QUICK ))
-		flags |= DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_TXN;
-	
-#if 0
-	/* Never do automatic recovery, must perform it manually.
-	 * Otherwise restarting with gentlehup will corrupt the
-	 * database.
-	 */
-	if( !(slapMode & SLAP_TOOL_MODE) ) flags |= DB_RECOVER;
+#ifdef HAVE_EBCDIC
+	strcpy( path, bdb->bi_dbenv_home );
+	__atoe( path );
+	dbhome = path;
+#else
+	dbhome = bdb->bi_dbenv_home;
 #endif
 
-	/* If a key was set, use shared memory for the BDB environment */
-	if ( bdb->bi_shm_key ) {
-		bdb->bi_dbenv->set_shm_key( bdb->bi_dbenv, bdb->bi_shm_key );
-		flags |= DB_SYSTEM_MEM;
+	/* If existing environment is clean but doesn't support
+	 * currently requested modes, remove it.
+	 */
+	if ( !do_recover && ( alockt ^ quick )) {
+		rc = bdb->bi_dbenv->remove( bdb->bi_dbenv, dbhome, DB_FORCE );
+		if ( rc ) {
+			Debug( LDAP_DEBUG_ANY,
+				"bdb_db_open: dbenv remove failed: %s (%d)\n",
+				db_strerror(rc), rc, 0 );
+			bdb->bi_dbenv = NULL;
+			goto fail;
+		}
+		rc = db_env_create( &bdb->bi_dbenv, 0 );
+		if( rc != 0 ) {
+			Debug( LDAP_DEBUG_ANY,
+				"bdb_db_open: db_env_create failed: %s (%d)\n",
+				db_strerror(rc), rc, 0 );
+			goto fail;
+		}
 	}
 
 	bdb->bi_dbenv->set_errpfx( bdb->bi_dbenv, be->be_suffix[0].bv_val );
 	bdb->bi_dbenv->set_errcall( bdb->bi_dbenv, bdb_errcall );
+
 	bdb->bi_dbenv->set_lk_detect( bdb->bi_dbenv, bdb->bi_lock_detect );
 
 	/* One long-lived TXN per thread, two TXNs per write op */
 	bdb->bi_dbenv->set_tx_max( bdb->bi_dbenv, connection_pool_max * 3 );
 
-#ifdef SLAP_ZONE_ALLOC
-	if ( bdb->bi_cache.c_maxsize ) {
-		bdb->bi_cache.c_zctx = slap_zn_mem_create(
-								SLAP_ZONE_INITSIZE,
-								SLAP_ZONE_MAXSIZE,
-								SLAP_ZONE_DELTA,
-								SLAP_ZONE_SIZE);
-	}
-#endif
-
-	if ( bdb->bi_idl_cache_max_size ) {
-		bdb->bi_idl_tree = NULL;
-		ldap_pvt_thread_rdwr_init( &bdb->bi_idl_tree_rwlock );
-		ldap_pvt_thread_mutex_init( &bdb->bi_idl_tree_lrulock );
-		bdb->bi_idl_cache_size = 0;
-	}
-
-#ifdef BDB_SUBDIRS
-	{
-		char dir[MAXPATHLEN], *ptr;
-		
-		if (bdb->bi_dbenv_home[0] == '.') {
-			/* If home is a relative path, relative subdirs
-			 * are just concat'd by BDB. We don't want the
-			 * path to be concat'd twice, e.g.
-			 * ./test-db/./test-db/tmp
-			 */
-			ptr = dir;
-		} else {
-			ptr = lutil_strcopy( dir, bdb->bi_dbenv_home );
-			*ptr++ = LDAP_DIRSEP[0];
-#ifdef HAVE_EBCDIC
-			__atoe( dir );
-#endif
-		}
-
-		strcpy( ptr, BDB_TMP_SUBDIR );
-#ifdef HAVE_EBCDIC
-		__atoe( ptr );
-#endif
-		rc = bdb->bi_dbenv->set_tmp_dir( bdb->bi_dbenv, dir );
-		if( rc != 0 ) {
-			Debug( LDAP_DEBUG_ANY,
-				"bdb_db_open: set_tmp_dir(%s) failed: %s (%d)\n",
-				dir, db_strerror(rc), rc );
-			return rc;
-		}
-
-		strcpy( ptr, BDB_LG_SUBDIR );
-#ifdef HAVE_EBCDIC
-		__atoe( ptr );
-#endif
-		rc = bdb->bi_dbenv->set_lg_dir( bdb->bi_dbenv, dir );
-		if( rc != 0 ) {
-			Debug( LDAP_DEBUG_ANY,
-				"bdb_db_open: set_lg_dir(%s) failed: %s (%d)\n",
-				dir, db_strerror(rc), rc );
-			return rc;
-		}
-
-		strcpy( ptr, BDB_DATA_SUBDIR );
-#ifdef HAVE_EBCDIC
-		__atoe( ptr );
-#endif
-		rc = bdb->bi_dbenv->set_data_dir( bdb->bi_dbenv, dir );
-		if( rc != 0 ) {
-			Debug( LDAP_DEBUG_ANY,
-				"bdb_db_open: set_data_dir(%s) failed: %s (%d)\n",
-				dir, db_strerror(rc), rc );
-			return rc;
-		}
-	}
-#endif
-
 	if( bdb->bi_dbenv_xflags != 0 ) {
 		rc = bdb->bi_dbenv->set_flags( bdb->bi_dbenv,
 			bdb->bi_dbenv_xflags, 1);
@@ -224,38 +263,62 @@ bdb_db_open( BackendDB *be )
 			Debug( LDAP_DEBUG_ANY,
 				"bdb_db_open: dbenv_set_flags failed: %s (%d)\n",
 				db_strerror(rc), rc, 0 );
-			return rc;
+			goto fail;
 		}
 	}
 
+#define	BDB_TXN_FLAGS	(DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_TXN)
+
 	Debug( LDAP_DEBUG_TRACE,
 		"bdb_db_open: dbenv_open(%s)\n",
 		bdb->bi_dbenv_home, 0, 0);
 
-#ifdef HAVE_EBCDIC
-	strcpy( path, bdb->bi_dbenv_home );
-	__atoe( path );
-	rc = bdb->bi_dbenv->open( bdb->bi_dbenv,
-		path,
-		flags,
-		bdb->bi_dbenv_mode );
-#else
-	rc = bdb->bi_dbenv->open( bdb->bi_dbenv,
-		bdb->bi_dbenv_home,
-		flags,
-		bdb->bi_dbenv_mode );
-#endif
-	if( rc != 0 ) {
+	flags = DB_INIT_MPOOL | DB_CREATE | DB_THREAD;
+
+	if ( !quick )
+		flags |= BDB_TXN_FLAGS;
+
+	/* If a key was set, use shared memory for the BDB environment */
+	if ( bdb->bi_shm_key ) {
+		bdb->bi_dbenv->set_shm_key( bdb->bi_dbenv, bdb->bi_shm_key );
+		flags |= DB_SYSTEM_MEM;
+	}
+	rc = bdb->bi_dbenv->open( bdb->bi_dbenv, dbhome,
+			flags | do_recover, bdb->bi_dbenv_mode );
+
+	if ( rc ) {
 		Debug( LDAP_DEBUG_ANY,
-			"bdb_db_open: dbenv_open failed: %s (%d)\n",
-			db_strerror(rc), rc, 0 );
-		return rc;
+			"bdb_db_open: Database cannot be %s, err %d. "
+			"Restore from backup!\n",
+				do_recover ? "recovered" : "opened", rc, 0);
+		goto fail;
+	}
+
+	if ( do_alock_recover && alock_recover (&bdb->bi_alock_info) != 0 ) {
+		Debug( LDAP_DEBUG_ANY,
+			"bdb_db_open: alock_recover failed\n",
+			0, 0, 0 );
+		rc = -1;
+		goto fail;
+	}
+
+#ifdef SLAP_ZONE_ALLOC
+	if ( bdb->bi_cache.c_maxsize ) {
+		bdb->bi_cache.c_zctx = slap_zn_mem_create(
+			SLAP_ZONE_INITSIZE, SLAP_ZONE_MAXSIZE,
+			SLAP_ZONE_DELTA, SLAP_ZONE_SIZE);
+	}
+#endif
+
+	if ( bdb->bi_idl_cache_max_size ) {
+		bdb->bi_idl_tree = NULL;
+		bdb->bi_idl_cache_size = 0;
 	}
 
 	flags = DB_THREAD | bdb->bi_db_opflags;
 
 #ifdef DB_AUTO_COMMIT
-	if ( !( slapMode & SLAP_TOOL_QUICK ))
+	if ( !quick )
 		flags |= DB_AUTO_COMMIT;
 #endif
 
@@ -273,10 +336,14 @@ bdb_db_open( BackendDB *be )
 			Debug( LDAP_DEBUG_ANY,
 				"bdb_db_open: db_create(%s) failed: %s (%d)\n",
 				bdb->bi_dbenv_home, db_strerror(rc), rc );
-			return rc;
+			goto fail;
 		}
 
 		if( i == BDB_ID2ENTRY ) {
+			if ( slapMode & SLAP_TOOL_MODE )
+				db->bdi_db->mpf->set_priority( db->bdi_db->mpf,
+					DB_PRIORITY_VERY_LOW );
+
 			rc = db->bdi_db->set_pagesize( db->bdi_db,
 				BDB_ID2ENTRY_PAGESIZE );
 			if ( slapMode & SLAP_TOOL_READMAIN ) {
@@ -330,7 +397,8 @@ bdb_db_open( BackendDB *be )
 			Debug( LDAP_DEBUG_ANY,
 				"bdb_db_open: db_open(%s) failed: %s (%d)\n",
 				buf, db_strerror(rc), rc );
-			return rc;
+			db->bdi_db->close( db->bdi_db, 0 );
+			goto fail;
 		}
 
 		flags &= ~(DB_CREATE | DB_RDONLY);
@@ -347,25 +415,20 @@ bdb_db_open( BackendDB *be )
 		Debug( LDAP_DEBUG_ANY,
 			"bdb_db_open: last_id(%s) failed: %s (%d)\n",
 			bdb->bi_dbenv_home, db_strerror(rc), rc );
-		return rc;
+		goto fail;
 	}
 
-	if ( !( slapMode & SLAP_TOOL_QUICK )) {
+	if ( !quick ) {
 		XLOCK_ID(bdb->bi_dbenv, &bdb->bi_cache.c_locker);
 	}
 
-	/* If we're in server mode and time-based checkpointing is enabled,
-	 * submit a task to perform periodic checkpoints.
-	 */
-	if ( slapMode & SLAP_SERVER_MODE && bdb->bi_txn_cp &&
-		bdb->bi_txn_cp_min )  {
-		ldap_pvt_thread_mutex_lock( &slapd_rq.rq_mutex );
-		ldap_pvt_runqueue_insert( &slapd_rq, bdb->bi_txn_cp_min*60,
-			bdb_checkpoint, bdb );
-		ldap_pvt_thread_mutex_unlock( &slapd_rq.rq_mutex );
-	}
+	bdb->bi_flags |= BDB_IS_OPEN;
 
 	return 0;
+
+fail:
+	bdb_db_close( be );
+	return rc;
 }
 
 static int
@@ -376,7 +439,12 @@ bdb_db_close( BackendDB *be )
 	struct bdb_db_info *db;
 	bdb_idl_cache_entry_t *entry, *next_entry;
 
-	while( bdb->bi_ndatabases-- ) {
+	bdb->bi_flags &= ~BDB_IS_OPEN;
+
+	ber_bvarray_free( bdb->bi_db_config );
+	bdb->bi_db_config = NULL;
+
+	while( bdb->bi_databases && bdb->bi_ndatabases-- ) {
 		db = bdb->bi_databases[bdb->bi_ndatabases];
 		rc = db->bdi_db->close( db->bdi_db, 0 );
 		/* Lower numbered names are not strdup'd */
@@ -385,13 +453,13 @@ bdb_db_close( BackendDB *be )
 		free( db );
 	}
 	free( bdb->bi_databases );
-	bdb_attr_index_destroy( bdb->bi_attrs );
+	bdb->bi_databases = NULL;
 
 	bdb_cache_release_all (&bdb->bi_cache);
 
 	if ( bdb->bi_idl_cache_max_size ) {
-		ldap_pvt_thread_rdwr_wlock ( &bdb->bi_idl_tree_rwlock );
 		avl_free( bdb->bi_idl_tree, NULL );
+		bdb->bi_idl_tree = NULL;
 		entry = bdb->bi_idl_lru_head;
 		while ( entry != NULL ) {
 			next_entry = entry->idl_lru_next;
@@ -401,33 +469,25 @@ bdb_db_close( BackendDB *be )
 			free( entry );
 			entry = next_entry;
 		}
-		ldap_pvt_thread_rdwr_wunlock ( &bdb->bi_idl_tree_rwlock );
+		bdb->bi_idl_lru_head = bdb->bi_idl_lru_tail = NULL;
 	}
 
-	if ( !( slapMode & SLAP_TOOL_QUICK )) {
-		XLOCK_ID_FREE(bdb->bi_dbenv, bdb->bi_cache.c_locker);
-	}
-
-	return 0;
-}
-
-static int
-bdb_db_destroy( BackendDB *be )
-{
-	int rc;
-	struct bdb_info *bdb = (struct bdb_info *) be->be_private;
-	Operation *ps = NULL;
-	Operation *psn = NULL;
-	void *saved_tmpmemctx = NULL;
-
 	/* close db environment */
 	if( bdb->bi_dbenv ) {
-		/* force a checkpoint */
+		/* Free cache locker if we enabled locking */
 		if ( !( slapMode & SLAP_TOOL_QUICK )) {
+			XLOCK_ID_FREE(bdb->bi_dbenv, bdb->bi_cache.c_locker);
+			bdb->bi_cache.c_locker = 0;
+		}
+
+		/* force a checkpoint, but not if we were ReadOnly,
+		 * and not in Quick mode since there are no transactions there.
+		 */
+		if ( !( slapMode & ( SLAP_TOOL_QUICK|SLAP_TOOL_READONLY ))) {
 			rc = TXN_CHECKPOINT( bdb->bi_dbenv, 0, 0, DB_FORCE );
 			if( rc != 0 ) {
 				Debug( LDAP_DEBUG_ANY,
-					"bdb_db_destroy: txn_checkpoint failed: %s (%d)\n",
+					"bdb_db_close: txn_checkpoint failed: %s (%d)\n",
 					db_strerror(rc), rc, 0 );
 			}
 		}
@@ -436,23 +496,43 @@ bdb_db_destroy( BackendDB *be )
 		bdb->bi_dbenv = NULL;
 		if( rc != 0 ) {
 			Debug( LDAP_DEBUG_ANY,
-				"bdb_db_destroy: close failed: %s (%d)\n",
+				"bdb_db_close: close failed: %s (%d)\n",
 				db_strerror(rc), rc, 0 );
 			return rc;
 		}
 	}
 
+	rc = alock_close( &bdb->bi_alock_info );
+	if( rc != 0 ) {
+		Debug( LDAP_DEBUG_ANY,
+			"bdb_db_close: alock_close failed\n", 0, 0, 0 );
+		return -1;
+	}
+
+	return 0;
+}
+
+static int
+bdb_db_destroy( BackendDB *be )
+{
+	struct bdb_info *bdb = (struct bdb_info *) be->be_private;
+
 	if( bdb->bi_dbenv_home ) ch_free( bdb->bi_dbenv_home );
+	if( bdb->bi_db_config_path ) ch_free( bdb->bi_db_config_path );
+
+	bdb_attr_index_destroy( bdb );
 
 	ldap_pvt_thread_rdwr_destroy ( &bdb->bi_cache.c_rwlock );
-	ldap_pvt_thread_mutex_destroy( &bdb->bi_cache.lru_mutex );
+	ldap_pvt_thread_mutex_destroy( &bdb->bi_cache.lru_head_mutex );
+	ldap_pvt_thread_mutex_destroy( &bdb->bi_cache.lru_tail_mutex );
 	ldap_pvt_thread_mutex_destroy( &bdb->bi_cache.c_dntree.bei_kids_mutex );
+#ifdef BDB_HIER
+	ldap_pvt_thread_mutex_destroy( &bdb->bi_modrdns_mutex );
+#endif
 	ldap_pvt_thread_mutex_destroy( &bdb->bi_lastid_mutex );
 	ldap_pvt_thread_mutex_destroy( &bdb->bi_database_mutex );
-	if ( bdb->bi_idl_cache_max_size ) {
-		ldap_pvt_thread_rdwr_destroy( &bdb->bi_idl_tree_rwlock );
-		ldap_pvt_thread_mutex_destroy( &bdb->bi_idl_tree_lrulock );
-	}
+	ldap_pvt_thread_rdwr_destroy( &bdb->bi_idl_tree_rwlock );
+	ldap_pvt_thread_mutex_destroy( &bdb->bi_idl_tree_lrulock );
 
 	ch_free( bdb );
 	be->be_private = NULL;
@@ -464,16 +544,17 @@ int
 bdb_back_initialize(
 	BackendInfo	*bi )
 {
+	int rc;
+
 	static char *controls[] = {
 		LDAP_CONTROL_ASSERT,
 		LDAP_CONTROL_MANAGEDSAIT,
 		LDAP_CONTROL_NOOP,
 		LDAP_CONTROL_PAGEDRESULTS,
-#ifdef LDAP_CONTROL_SUBENTRIES
 		LDAP_CONTROL_SUBENTRIES,
-#endif
-#ifdef LDAP_CONTROL_X_PERMISSIVE_MODIFY
 		LDAP_CONTROL_X_PERMISSIVE_MODIFY,
+#ifdef LDAP_X_TXN
+		LDAP_CONTROL_X_TXN_SPEC,
 #endif
 		NULL
 	};
@@ -485,9 +566,7 @@ bdb_back_initialize(
 
 	bi->bi_flags |=
 		SLAP_BFLAG_INCREMENT |
-#ifdef BDB_SUBENTRIES
 		SLAP_BFLAG_SUBENTRIES |
-#endif
 		SLAP_BFLAG_ALIASES |
 		SLAP_BFLAG_REFERRALS;
 
@@ -509,16 +588,17 @@ bdb_back_initialize(
 #endif
 
 		ver = (major << 24) | (minor << 16) | patch;
-		if( ver < DB_VERSION_FULL )
-		{
+		if( ver != DB_VERSION_FULL ) {
+			/* fail if a versions don't match */
 			Debug( LDAP_DEBUG_ANY,
 				LDAP_XSTRING(bdb_back_initialize) ": "
 				"BDB library version mismatch:"
 				" expected " DB_VERSION_STRING ","
 				" got %s\n", version, 0, 0 );
+			return -1;
 		}
 
-		Debug( LDAP_DEBUG_ANY, LDAP_XSTRING(bdb_back_initialize)
+		Debug( LDAP_DEBUG_TRACE, LDAP_XSTRING(bdb_back_initialize)
 			": %s\n", version, 0, 0 );
 	}
 
@@ -532,20 +612,13 @@ bdb_back_initialize(
 	db_env_set_func_yield( ldap_pvt_thread_yield );
 #endif
 
-	{
-		static char uuidbuf[ LDAP_LUTIL_UUIDSTR_BUFSIZE ];
-
-		bdb_uuid.bv_len = lutil_uuidstr( uuidbuf, sizeof( uuidbuf ));
-		bdb_uuid.bv_val = uuidbuf;
-	}
-
 	bi->bi_open = 0;
 	bi->bi_close = 0;
 	bi->bi_config = 0;
 	bi->bi_destroy = 0;
 
 	bi->bi_db_init = bdb_db_init;
-	bi->bi_db_config = bdb_db_config;
+	bi->bi_db_config = config_generic_wrapper;
 	bi->bi_db_open = bdb_db_open;
 	bi->bi_db_close = bdb_db_close;
 	bi->bi_db_destroy = bdb_db_destroy;
@@ -586,7 +659,9 @@ bdb_back_initialize(
 	bi->bi_connection_init = 0;
 	bi->bi_connection_destroy = 0;
 
-	return 0;
+	rc = bdb_back_init_cf( bi );
+
+	return rc;
 }
 
 #if	(SLAPD_BDB == SLAPD_MOD_DYNAMIC && !defined(BDB_HIER)) || \
