@@ -1,8 +1,8 @@
 /* search.c - ldap backend search function */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-ldap/search.c,v 1.148.2.5 2005/02/02 17:32:44 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-ldap/search.c,v 1.201.2.1 2006/04/11 16:07:18 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2005 The OpenLDAP Foundation.
+ * Copyright 1999-2006 The OpenLDAP Foundation.
  * Portions Copyright 1999-2003 Howard Chu.
  * Portions Copyright 2000-2003 Pierangelo Masarati.
  * All rights reserved.
@@ -38,30 +38,128 @@
 
 static int
 ldap_build_entry( Operation *op, LDAPMessage *e, Entry *ent,
-	 struct berval *bdn, int flags );
-#define LDAP_BUILD_ENTRY_PRIVATE	0x01
+	 struct berval *bdn );
+
+/*
+ * Quick'n'dirty rewrite of filter in case of error, to deal with
+ * <draft-zeilenga-ldap-t-f>.
+ */
+static int
+ldap_back_munge_filter(
+	Operation	*op,
+	struct berval	*filter )
+{
+	ldapinfo_t	*li = (ldapinfo_t *) op->o_bd->be_private;
+
+	char		*ptr;
+	int		gotit = 0;
+
+	Debug( LDAP_DEBUG_ARGS, "=> ldap_back_munge_filter \"%s\"\n",
+			filter->bv_val, 0, 0 );
+
+	for ( ptr = strstr( filter->bv_val, "(?=" ); 
+			ptr;
+			ptr = strstr( ptr, "(?=" ) )
+	{
+		static struct berval
+			bv_true = BER_BVC( "(?=true)" ),
+			bv_false = BER_BVC( "(?=false)" ),
+			bv_undefined = BER_BVC( "(?=undefined)" ),
+			bv_t = BER_BVC( "(&)" ),
+			bv_f = BER_BVC( "(|)" ),
+			bv_T = BER_BVC( "(objectClass=*)" ),
+			bv_F = BER_BVC( "(!(objectClass=*))" );
+		struct berval	*oldbv = NULL,
+				*newbv = NULL,
+				oldfilter = BER_BVNULL;
+
+		if ( strncmp( ptr, bv_true.bv_val, bv_true.bv_len ) == 0 ) {
+			oldbv = &bv_true;
+			if ( li->li_flags & LDAP_BACK_F_SUPPORT_T_F ) {
+				newbv = &bv_t;
+
+			} else {
+				newbv = &bv_T;
+			}
+
+		} else if ( strncmp( ptr, bv_false.bv_val, bv_false.bv_len ) == 0 )
+		{
+			oldbv = &bv_false;
+			if ( li->li_flags & LDAP_BACK_F_SUPPORT_T_F ) {
+				newbv = &bv_f;
+
+			} else {
+				newbv = &bv_F;
+			}
+
+		} else if ( strncmp( ptr, bv_undefined.bv_val, bv_undefined.bv_len ) == 0 )
+		{
+			oldbv = &bv_undefined;
+			newbv = &bv_F;
+
+		} else {
+			gotit = 0;
+			goto done;
+		}
+
+		oldfilter = *filter;
+		if ( newbv->bv_len > oldbv->bv_len ) {
+			filter->bv_len += newbv->bv_len - oldbv->bv_len;
+			if ( filter->bv_val == op->ors_filterstr.bv_val ) {
+				filter->bv_val = op->o_tmpalloc( filter->bv_len + 1,
+						op->o_tmpmemctx );
+
+				AC_MEMCPY( filter->bv_val, op->ors_filterstr.bv_val,
+						op->ors_filterstr.bv_len + 1 );
+
+			} else {
+				filter->bv_val = op->o_tmprealloc( filter->bv_val,
+						filter->bv_len + 1, op->o_tmpmemctx );
+			}
+
+			ptr = filter->bv_val + ( ptr - oldfilter.bv_val );
+		}
+
+		AC_MEMCPY( &ptr[ newbv->bv_len ],
+				&ptr[ oldbv->bv_len ], 
+				oldfilter.bv_len - ( ptr - filter->bv_val ) - oldbv->bv_len + 1 );
+		AC_MEMCPY( ptr, newbv->bv_val, newbv->bv_len );
+
+		ptr += newbv->bv_len;
+		gotit = 1;
+	}
+
+done:;
+	Debug( LDAP_DEBUG_ARGS, "<= ldap_back_munge_filter \"%s\" (%d)\n",
+			filter->bv_val, gotit, 0 );
+
+	return gotit;
+}
 
 int
 ldap_back_search(
 		Operation	*op,
 		SlapReply	*rs )
 {
-	struct ldapconn *lc;
+	ldapconn_t	*lc;
 	struct timeval	tv;
+	time_t		stoptime = (time_t)-1;
 	LDAPMessage	*res,
 			*e;
 	int		rc = 0,
 			msgid; 
-	struct berval	match = BER_BVNULL;
+	struct berval	match = BER_BVNULL,
+			filter = BER_BVNULL;
 	int		i;
 	char		**attrs = NULL;
-	int		dontfreetext = 0;
-	int		freeconn = 0;
-	int		do_retry = 1;
+	int		freetext = 0;
+	int		do_retry = 1, dont_retry = 0;
 	LDAPControl	**ctrls = NULL;
+	/* FIXME: shouldn't this be null? */
+	const char	*save_matched = rs->sr_matched;
 
 	lc = ldap_back_getconn( op, rs, LDAP_BACK_SENDERR );
-	if ( !lc ) {
+	if ( !lc || !ldap_back_dobind( lc, op, rs, LDAP_BACK_SENDERR ) ) {
 		return rs->sr_err;
 	}
 
@@ -69,9 +167,6 @@ ldap_back_search(
 	 * FIXME: in case of values return filter, we might want
 	 * to map attrs and maybe rewrite value
 	 */
-	if ( !ldap_back_dobind( lc, op, rs, LDAP_BACK_SENDERR ) ) {
-		return rs->sr_err;
-	}
 
 	/* should we check return values? */
 	if ( op->ors_deref != -1 ) {
@@ -82,9 +177,10 @@ ldap_back_search(
 	if ( op->ors_tlimit != SLAP_NO_LIMIT ) {
 		tv.tv_sec = op->ors_tlimit;
 		tv.tv_usec = 0;
+		stoptime = op->o_time + op->ors_tlimit;
 
 	} else {
-		tv.tv_sec = 0;
+		LDAP_BACK_TV_SET( &tv );
 	}
 
 	if ( op->ors_attrs ) {
@@ -107,25 +203,53 @@ ldap_back_search(
 	ctrls = op->o_ctrls;
 	rc = ldap_back_proxy_authz_ctrl( lc, op, rs, &ctrls );
 	if ( rc != LDAP_SUCCESS ) {
-		dontfreetext = 1;
 		goto finish;
 	}
-	
+
+	/* deal with <draft-zeilenga-ldap-t-f> filters */
+	filter = op->ors_filterstr;
 retry:
-	rs->sr_err = ldap_search_ext( lc->lc_ld, op->o_req_ndn.bv_val,
-			op->ors_scope, op->ors_filterstr.bv_val,
+	rs->sr_err = ldap_search_ext( lc->lc_ld, op->o_req_dn.bv_val,
+			op->ors_scope, filter.bv_val,
 			attrs, op->ors_attrsonly, ctrls, NULL,
 			tv.tv_sec ? &tv : NULL,
 			op->ors_slimit, &msgid );
 
 	if ( rs->sr_err != LDAP_SUCCESS ) {
-fail:;
-		rc = ldap_back_op_result( lc, op, rs, msgid, LDAP_BACK_SENDERR );
-		if ( freeconn ) {
-			ldap_back_freeconn( op, lc );
-			lc = NULL;
+		switch ( rs->sr_err ) {
+		case LDAP_SERVER_DOWN:
+			if ( do_retry ) {
+				do_retry = 0;
+				if ( ldap_back_retry( &lc, op, rs, LDAP_BACK_DONTSEND ) ) {
+					goto retry;
+				}
+			}
+
+			if ( lc == NULL ) {
+				/* reset by ldap_back_retry ... */
+				rs->sr_err = slap_map_api2result( rs );
+
+			} else {
+				rc = ldap_back_op_result( lc, op, rs, msgid, 0, LDAP_BACK_DONTSEND );
+			}
+				
+			goto finish;
+
+		case LDAP_FILTER_ERROR:
+			if ( ldap_back_munge_filter( op, &filter ) ) {
+				goto retry;
+			}
+
+			/* invalid filters return success with no data */
+			rs->sr_err = LDAP_SUCCESS;
+			rs->sr_text = NULL;
+			goto finish;
+		
+		default:
+			rs->sr_err = slap_map_api2result( rs );
+			rs->sr_text = NULL;
+			goto finish;
 		}
-		goto finish;
 	}
 
 	/* We pull apart the ber result, stuff it into a slapd entry, and
@@ -133,69 +257,76 @@ fail:;
 	 * but this is necessary for version matching, and for ACL processing.
 	 */
 
-	for ( rc = 0; rc != -1; rc = ldap_result( lc->lc_ld, msgid, 0, &tv, &res ) )
+	for ( rc = 0; rc != -1; rc = ldap_result( lc->lc_ld, msgid, LDAP_MSG_ONE, &tv, &res ) )
 	{
 		/* check for abandon */
 		if ( op->o_abandon ) {
+			if ( rc > 0 ) {
+				ldap_msgfree( res );
+			}
 			ldap_abandon_ext( lc->lc_ld, msgid, NULL, NULL );
-			rc = 0;
+			rc = SLAPD_ABANDON;
 			goto finish;
 		}
 
 		if ( rc == 0 ) {
-			tv.tv_sec = 0;
-			tv.tv_usec = 100000;
+			LDAP_BACK_TV_SET( &tv );
 			ldap_pvt_thread_yield();
 
-		} else if ( rc == LDAP_RES_SEARCH_ENTRY ) {
-			Entry		ent = {0};
-			struct berval	bdn;
-			int		abort = 0;
+			/* check time limit */
+			if ( op->ors_tlimit != SLAP_NO_LIMIT
+					&& slap_get_time() > stoptime )
+			{
+				ldap_abandon_ext( lc->lc_ld, msgid, NULL, NULL );
+				rc = rs->sr_err = LDAP_TIMELIMIT_EXCEEDED;
+				goto finish;
+			}
+			continue;
+
+		} else {
+			/* don't retry any more */
+			dont_retry = 1;
+		}
+
+
+		if ( rc == LDAP_RES_SEARCH_ENTRY ) {
+			Entry		ent = { 0 };
+			struct berval	bdn = BER_BVNULL;
 
 			do_retry = 0;
 
 			e = ldap_first_entry( lc->lc_ld, res );
-			rc = ldap_build_entry( op, e, &ent, &bdn,
-					LDAP_BUILD_ENTRY_PRIVATE );
-		       if ( rc == LDAP_SUCCESS ) {
+			rc = ldap_build_entry( op, e, &ent, &bdn );
+			if ( rc == LDAP_SUCCESS ) {
 				rs->sr_entry = &ent;
 				rs->sr_attrs = op->ors_attrs;
 				rs->sr_operational_attrs = NULL;
 				rs->sr_flags = 0;
-				abort = send_search_entry( op, rs );
-				while ( ent.e_attrs ) {
-					Attribute	*a;
-					BerVarray	v;
-
-					a = ent.e_attrs;
-					ent.e_attrs = a->a_next;
-
-					v = a->a_vals;
-					if ( a->a_vals != &slap_dummy_bv ) {
-						ber_bvarray_free( a->a_vals );
-					}
-					if ( a->a_nvals != v ) {
-						ber_bvarray_free( a->a_nvals );
-					}
-					ch_free( a );
+				rs->sr_err = LDAP_SUCCESS;
+				rc = rs->sr_err = send_search_entry( op, rs );
+				if ( !BER_BVISNULL( &ent.e_name ) ) {
+					assert( ent.e_name.bv_val != bdn.bv_val );
+					op->o_tmpfree( ent.e_name.bv_val, op->o_tmpmemctx );
+					BER_BVZERO( &ent.e_name );
 				}
-				
-				if ( ent.e_dn && ( ent.e_dn != bdn.bv_val ) ) {
-					free( ent.e_dn );
-				}
-				if ( ent.e_ndn ) {
-					free( ent.e_ndn );
+				if ( !BER_BVISNULL( &ent.e_nname ) ) {
+					op->o_tmpfree( ent.e_nname.bv_val, op->o_tmpmemctx );
+					BER_BVZERO( &ent.e_nname );
 				}
+				entry_clean( &ent );
 			}
 			ldap_msgfree( res );
-			if ( abort ) {
-				ldap_abandon_ext( lc->lc_ld, msgid, NULL, NULL );
+			if ( rc != LDAP_SUCCESS ) {
+				if ( rc == LDAP_UNAVAILABLE ) {
+					rc = rs->sr_err = LDAP_OTHER;
+				} else {
+					ldap_abandon_ext( lc->lc_ld, msgid, NULL, NULL );
+				}
 				goto finish;
 			}
 
 		} else if ( rc == LDAP_RES_SEARCH_REFERENCE ) {
 			char		**references = NULL;
-			int		cnt;
 
 			do_retry = 0;
 			rc = ldap_parse_reference( lc->lc_ld, res,
@@ -205,25 +336,35 @@ fail:;
 				continue;
 			}
 
-			if ( references == NULL ) {
-				continue;
+			/* FIXME: there MUST be at least one */
+			if ( references && references[ 0 ] && references[ 0 ][ 0 ] ) {
+				int		cnt;
+
+				for ( cnt = 0; references[ cnt ]; cnt++ )
+					/* NO OP */ ;
+
+				/* FIXME: there MUST be at least one */
+				rs->sr_ref = ch_malloc( ( cnt + 1 ) * sizeof( struct berval ) );
+
+				for ( cnt = 0; references[ cnt ]; cnt++ ) {
+					ber_str2bv( references[ cnt ], 0, 0, &rs->sr_ref[ cnt ] );
+				}
+				BER_BVZERO( &rs->sr_ref[ cnt ] );
+
+				/* ignore return value by now */
+				( void )send_search_reference( op, rs );
+
+			} else {
+				Debug( LDAP_DEBUG_ANY,
+					"%s ldap_back_search: "
+					"got SEARCH_REFERENCE "
+					"with no referrals\n",
+					op->o_log_prefix, 0, 0 );
 			}
 
-			for ( cnt = 0; references[ cnt ]; cnt++ )
-				/* NO OP */ ;
-				
-			rs->sr_ref = ch_calloc( cnt + 1, sizeof( struct berval ) );
-
-			for ( cnt = 0; references[ cnt ]; cnt++ ) {
-				ber_str2bv( references[ cnt ], 0, 0, &rs->sr_ref[ cnt ] );
-			}
-
-			/* ignore return value by now */
-			( void )send_search_reference( op, rs );
-
 			/* cleanup */
 			if ( references ) {
-				ldap_value_free( references );
+				ber_memvfree( (void **)references );
 				ch_free( rs->sr_ref );
 				rs->sr_ref = NULL;
 			}
@@ -239,59 +380,88 @@ fail:;
 			rc = ldap_parse_result( lc->lc_ld, res, &rs->sr_err,
 					&match.bv_val, (char **)&rs->sr_text,
 					&references, &rs->sr_ctrls, 1 );
+			freetext = 1;
 			if ( rc != LDAP_SUCCESS ) {
 				rs->sr_err = rc;
 			}
 			rs->sr_err = slap_map_api2result( rs );
 
-			if ( references ) {
+			if ( references && references[ 0 ] && references[ 0 ][ 0 ] ) {
 				int	cnt;
 
+				if ( rs->sr_err != LDAP_REFERRAL ) {
+					/* FIXME: error */
+					Debug( LDAP_DEBUG_ANY,
+						"%s ldap_back_search: "
+						"got referrals with %d\n",
+						op->o_log_prefix,
+						rs->sr_err, 0 );
+					rs->sr_err = LDAP_REFERRAL;
+				}
+
 				for ( cnt = 0; references[ cnt ]; cnt++ )
 					/* NO OP */ ;
 				
-				rs->sr_ref = ch_calloc( cnt + 1, sizeof( struct berval ) );
+				rs->sr_ref = ch_malloc( ( cnt + 1 ) * sizeof( struct berval ) );
 
 				for ( cnt = 0; references[ cnt ]; cnt++ ) {
-					ber_str2bv( references[ cnt ], 0, 0, &rs->sr_ref[ cnt ] );
+					/* duplicating ...*/
+					ber_str2bv( references[ cnt ], 0, 1, &rs->sr_ref[ cnt ] );
 				}
+				BER_BVZERO( &rs->sr_ref[ cnt ] );
+			}
 
-				/* cleanup */
-				if ( references ) {
-					ldap_value_free( references );
+			if ( match.bv_val != NULL ) {
+				{
+					match.bv_len = strlen( match.bv_val );
 				}
 			}
 
+			/* cleanup */
+			if ( references ) {
+				ber_memvfree( (void **)references );
+			}
+
 			rc = 0;
 			break;
 		}
 	}
 
-	if ( rc == -1 ) {
+ 	if ( rc == -1 && dont_retry == 0 ) {
 		if ( do_retry ) {
 			do_retry = 0;
-			if ( ldap_back_retry( lc, op, rs, LDAP_BACK_SENDERR ) ) {
+			if ( ldap_back_retry( &lc, op, rs, LDAP_BACK_DONTSEND ) ) {
 				goto retry;
 			}
 		}
-		/* FIXME: invalidate the connection? */
 		rs->sr_err = LDAP_SERVER_DOWN;
-		freeconn = 1;
-		goto fail;
+		rs->sr_err = slap_map_api2result( rs );
+		goto finish;
 	}
 
 	/*
 	 * Rewrite the matched portion of the search base, if required
 	 */
 	if ( !BER_BVISNULL( &match ) && !BER_BVISEMPTY( &match ) ) {
-		rs->sr_matched = match.bv_val;
+		struct berval	pmatch;
+
+		if ( dnPretty( NULL, &match, &pmatch, op->o_tmpmemctx ) == LDAP_SUCCESS ) {
+			rs->sr_matched = pmatch.bv_val;
+			LDAP_FREE( match.bv_val );
+
+		} else {
+			rs->sr_matched = match.bv_val;
+		}
 	}
+
 	if ( rs->sr_v2ref ) {
 		rs->sr_err = LDAP_REFERRAL;
 	}
 
 finish:;
-	send_ldap_result( op, rs );
+	if ( rc != SLAPD_ABANDON ) {
+		send_ldap_result( op, rs );
+	}
 
 	(void)ldap_back_proxy_authz_ctrl_free( op, &ctrls );
 
@@ -300,13 +470,22 @@ finish:;
 		rs->sr_ctrls = NULL;
 	}
 
-	if ( match.bv_val ) {
-		rs->sr_matched = NULL;
-		LDAP_FREE( match.bv_val );
+	if ( rs->sr_matched != NULL && rs->sr_matched != save_matched ) {
+		if ( rs->sr_matched != match.bv_val ) {
+			ber_memfree_x( (char *)rs->sr_matched, op->o_tmpmemctx );
+
+		} else {
+			LDAP_FREE( match.bv_val );
+		}
+		rs->sr_matched = save_matched;
+	}
+
+	if ( !BER_BVISNULL( &filter ) && filter.bv_val != op->ors_filterstr.bv_val ) {
+		op->o_tmpfree( filter.bv_val, op->o_tmpmemctx );
 	}
 
 	if ( rs->sr_text ) {
-		if ( !dontfreetext ) {
+		if ( freetext ) {
 			LDAP_FREE( (char *)rs->sr_text );
 		}
 		rs->sr_text = NULL;
@@ -321,7 +500,11 @@ finish:;
 		ch_free( attrs );
 	}
 
-	return rc;
+	if ( lc != NULL ) {
+		ldap_back_release_conn( op, rs, lc );
+	}
+
+	return rs->sr_err;
 }
 
 static int
@@ -329,18 +512,16 @@ ldap_build_entry(
 		Operation	*op,
 		LDAPMessage	*e,
 		Entry		*ent,
-		struct berval	*bdn,
-		int		flags )
+		struct berval	*bdn )
 {
 	struct berval	a;
 	BerElement	ber = *e->lm_ber;
 	Attribute	*attr, **attrp;
 	const char	*text;
 	int		last;
-	int		private = flags & LDAP_BUILD_ENTRY_PRIVATE;
 
 	/* safe assumptions ... */
-	assert( ent );
+	assert( ent != NULL );
 	BER_BVZERO( &ent->e_bv );
 
 	if ( ber_scanf( &ber, "{m{", bdn ) == LBER_ERROR ) {
@@ -380,19 +561,22 @@ ldap_build_entry(
 		if ( slap_bv2ad( &a, &attr->a_desc, &text ) 
 				!= LDAP_SUCCESS )
 		{
-			if ( slap_bv2undef_ad( &a, &attr->a_desc, &text ) 
-					!= LDAP_SUCCESS )
+			if ( slap_bv2undef_ad( &a, &attr->a_desc, &text,
+				SLAP_AD_PROXIED ) != LDAP_SUCCESS )
 			{
 				Debug( LDAP_DEBUG_ANY, 
-					"slap_bv2undef_ad(%s):	%s\n",
-					a.bv_val, text, 0 );
+					"%s ldap_build_entry: "
+					"slap_bv2undef_ad(%s): %s\n",
+					op->o_log_prefix, a.bv_val, text );
 				ch_free( attr );
 				continue;
 			}
 		}
 
 		/* no subschemaSubentry */
-		if ( attr->a_desc == slap_schema.si_ad_subschemaSubentry ) {
+		if ( attr->a_desc == slap_schema.si_ad_subschemaSubentry
+			|| attr->a_desc == slap_schema.si_ad_entryDN )
+		{
 
 			/* 
 			 * We eat target's subschemaSubentry because
@@ -400,6 +584,10 @@ ldap_build_entry(
 			 * to resolve to the appropriate backend;
 			 * later, the local subschemaSubentry is
 			 * added.
+			 *
+			 * We also eat entryDN because the frontend
+			 * will reattach it without checking if already
+			 * present...
 			 */
 			( void )ber_scanf( &ber, "x" /* [W] */ );
 
@@ -414,13 +602,7 @@ ldap_build_entry(
 			 * Note: attr->a_vals can be null when using
 			 * values result filter
 			 */
-			if ( private ) {
-				attr->a_vals = (struct berval *)&slap_dummy_bv;
-				
-			} else {
-				attr->a_vals = ch_malloc( sizeof( struct berval ) );
-				BER_BVZERO( &attr->a_vals[ 0 ] );
-			}
+			attr->a_vals = (struct berval *)&slap_dummy_bv;
 			last = 0;
 
 		} else {
@@ -520,18 +702,20 @@ ldap_back_entry_get(
 		Entry			**ent
 )
 {
-	struct ldapconn *lc;
+	ldapconn_t	*lc;
 	int		rc = 1,
 			do_not_cache;
 	struct berval	bdn;
 	LDAPMessage	*result = NULL,
 			*e = NULL;
-	char		*gattr[3];
+	char		*attr[3], **attrp = NULL;
 	char		*filter = NULL;
 	SlapReply	rs;
 	int		do_retry = 1;
 	LDAPControl	**ctrls = NULL;
 
+	*ent = NULL;
+
 	/* Tell getconn this is a privileged op */
 	do_not_cache = op->o_do_not_cache;
 	op->o_do_not_cache = 1;
@@ -543,14 +727,15 @@ ldap_back_entry_get(
 	op->o_do_not_cache = do_not_cache;
 
 	if ( at ) {
+		attrp = attr;
 		if ( oc && at != slap_schema.si_ad_objectClass ) {
-			gattr[0] = slap_schema.si_ad_objectClass->ad_cname.bv_val;
-			gattr[1] = at->ad_cname.bv_val;
-			gattr[2] = NULL;
+			attr[0] = slap_schema.si_ad_objectClass->ad_cname.bv_val;
+			attr[1] = at->ad_cname.bv_val;
+			attr[2] = NULL;
 
 		} else {
-			gattr[0] = at->ad_cname.bv_val;
-			gattr[1] = NULL;
+			attr[0] = at->ad_cname.bv_val;
+			attr[1] = NULL;
 		}
 	}
 
@@ -573,12 +758,12 @@ ldap_back_entry_get(
 	
 retry:
 	rc = ldap_search_ext_s( lc->lc_ld, ndn->bv_val, LDAP_SCOPE_BASE, filter,
-				at ? gattr : NULL, 0, ctrls, NULL,
-				LDAP_NO_LIMIT, LDAP_NO_LIMIT, &result );
+				attrp, 0, ctrls, NULL,
+				NULL, LDAP_NO_LIMIT, &result );
 	if ( rc != LDAP_SUCCESS ) {
 		if ( rc == LDAP_SERVER_DOWN && do_retry ) {
 			do_retry = 0;
-			if ( ldap_back_retry( lc, op, &rs, LDAP_BACK_DONTSEND ) ) {
+			if ( ldap_back_retry( &lc, op, &rs, LDAP_BACK_DONTSEND ) ) {
 				goto retry;
 			}
 		}
@@ -587,12 +772,17 @@ retry:
 
 	e = ldap_first_entry( lc->lc_ld, result );
 	if ( e == NULL ) {
+		/* the entry exists, but it doesn't match the filter? */
 		goto cleanup;
 	}
 
 	*ent = ch_calloc( 1, sizeof( Entry ) );
+	if ( *ent == NULL ) {
+		rc = LDAP_NO_MEMORY;
+		goto cleanup;
+	}
 
-	rc = ldap_build_entry( op, e, *ent, &bdn, 0 );
+	rc = ldap_build_entry( op, e, *ent, &bdn );
 
 	if ( rc != LDAP_SUCCESS ) {
 		ch_free( *ent );
@@ -610,6 +800,10 @@ cleanup:
 		ch_free( filter );
 	}
 
+	if ( lc != NULL ) {
+		ldap_back_release_conn( op, &rs, lc );
+	}
+
 	return rc;
 }
 
