@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-meta/unbind.c,v 1.11 2004/09/04 02:54:33 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-meta/unbind.c,v 1.30 2006/02/11 16:40:29 ando Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2004 The OpenLDAP Foundation.
+ * Copyright 1999-2006 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * Portions Copyright 1999-2003 Howard Chu.
  * All rights reserved.
@@ -24,6 +24,7 @@
 
 #include <stdio.h>
 
+#include <ac/errno.h>
 #include <ac/socket.h>
 #include <ac/string.h>
 
@@ -33,48 +34,41 @@
 
 int
 meta_back_conn_destroy(
-		Backend		*be,
-		Connection	*conn
-)
+	Backend		*be,
+	Connection	*conn )
 {
-	struct metainfo	*li = ( struct metainfo * )be->be_private;
-	struct metaconn *lc, lc_curr;
+	metainfo_t	*mi = ( metainfo_t * )be->be_private;
+	metaconn_t	*mc,
+			mc_curr = { 0 };
+	int		i;
+
 
 	Debug( LDAP_DEBUG_TRACE,
-		"=>meta_back_conn_destroy: fetching conn %ld\n%s%s",
-		conn->c_connid, "", "" );
+		"=>meta_back_conn_destroy: fetching conn=%ld DN=\"%s\"\n",
+		conn->c_connid,
+		BER_BVISNULL( &conn->c_ndn ) ? "" : conn->c_ndn.bv_val, 0 );
 	
-	lc_curr.conn = conn;
+	mc_curr.mc_conn = conn;
 	
-	ldap_pvt_thread_mutex_lock( &li->conn_mutex );
-	lc = avl_delete( &li->conntree, ( caddr_t )&lc_curr,
-			meta_back_conn_cmp );
-	ldap_pvt_thread_mutex_unlock( &li->conn_mutex );
-
-	if ( lc ) {
-		int i;
-		
+	ldap_pvt_thread_mutex_lock( &mi->mi_conninfo.lai_mutex );
+	while ( ( mc = avl_delete( &mi->mi_conninfo.lai_tree, ( caddr_t )&mc_curr, meta_back_conn_cmp ) ) != NULL )
+	{
 		Debug( LDAP_DEBUG_TRACE,
-			"=>meta_back_conn_destroy: destroying conn %ld\n%s%s",
-			lc->conn->c_connid, "", "" );
+			"=>meta_back_conn_destroy: destroying conn %ld\n",
+			LDAP_BACK_PCONN_ID( mc->mc_conn ), 0, 0 );
 		
-		/*
-		 * Cleanup rewrite session
-		 */
-		for ( i = 0; i < li->ntargets; ++i ) {
-			if ( lc->conns[ i ].ld == NULL ) {
-				continue;
-			}
+		assert( mc->mc_refcnt == 0 );
 
-			rewrite_session_delete( li->targets[ i ]->rwmap.rwm_rw, conn );
-			meta_clear_one_candidate( &lc->conns[ i ], 1 );
-		}
-
-		free( lc->conns );
-		free( lc );
+		meta_back_conn_free( mc );
 	}
+	ldap_pvt_thread_mutex_unlock( &mi->mi_conninfo.lai_mutex );
 
-	/* no response to unbind */
+	/*
+	 * Cleanup rewrite session
+	 */
+	for ( i = 0; i < mi->mi_ntargets; ++i ) {
+		rewrite_session_delete( mi->mi_targets[ i ].mt_rwmap.rwm_rw, conn );
+	}
 
 	return 0;
 }
