@@ -1,8 +1,8 @@
 /* backend.c - deals with backend subsystem */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-monitor/backend.c,v 1.33.2.1 2005/01/20 18:04:00 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-monitor/backend.c,v 1.41 2006/01/12 07:40:09 hyc Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2005 The OpenLDAP Foundation.
+ * Copyright 2001-2006 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -42,6 +42,7 @@ monitor_subsys_backend_init(
 	int			i;
 	monitor_entry_t		*mp;
 	monitor_subsys_t	*ms_database;
+	BackendInfo			*bi;
 
 	mi = ( monitor_info_t * )be->be_private;
 
@@ -56,8 +57,7 @@ monitor_subsys_backend_init(
 		return -1;
 	}
 
-	if ( monitor_cache_get( mi, &ms->mss_ndn, &e_backend ) )
-	{
+	if ( monitor_cache_get( mi, &ms->mss_ndn, &e_backend ) ) {
 		Debug( LDAP_DEBUG_ANY,
 			"monitor_subsys_backend_init: "
 			"unable to get entry \"%s\"\n",
@@ -69,35 +69,22 @@ monitor_subsys_backend_init(
 	mp->mp_children = NULL;
 	ep = &mp->mp_children;
 
-	for ( i = 0; i < nBackendInfo; i++ ) {
+	i = -1;
+	LDAP_STAILQ_FOREACH( bi, &backendInfo, bi_next ) {
 		char 		buf[ BACKMONITOR_BUFSIZE ];
-		BackendInfo 	*bi;
+		BackendDB	*be;
 		struct berval 	bv;
 		int		j;
 		Entry		*e;
 
-		bi = &backendInfo[ i ];
+		i++;
+
+		bv.bv_len = snprintf( buf, sizeof( buf ), "cn=Backend %d", i );
+		bv.bv_val = buf;
+
+		e = monitor_entry_stub( &ms->mss_dn, &ms->mss_ndn, &bv,
+			mi->mi_oc_monitoredObject, mi, NULL, NULL );
 
-		snprintf( buf, sizeof( buf ),
-				"dn: cn=Backend %d,%s\n"
-				"objectClass: %s\n"
-				"structuralObjectClass: %s\n"
-				"cn: Backend %d\n"
-				"creatorsName: %s\n"
-				"modifiersName: %s\n"
-				"createTimestamp: %s\n"
-				"modifyTimestamp: %s\n",
-				i,
-				ms->mss_dn.bv_val,
-				mi->mi_oc_monitoredObject->soc_cname.bv_val,
-				mi->mi_oc_monitoredObject->soc_cname.bv_val,
-				i,
-				mi->mi_creatorsName.bv_val,
-				mi->mi_creatorsName.bv_val,
-				mi->mi_startTime.bv_val,
-				mi->mi_startTime.bv_val );
-		
-		e = str2entry( buf );
 		if ( e == NULL ) {
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_subsys_backend_init: "
@@ -106,39 +93,42 @@ monitor_subsys_backend_init(
 			return( -1 );
 		}
 		
-		bv.bv_val = bi->bi_type;
-		bv.bv_len = strlen( bv.bv_val );
-
+		ber_str2bv( bi->bi_type, 0, 0, &bv );
 		attr_merge_normalize_one( e, mi->mi_ad_monitoredInfo,
 				&bv, NULL );
 		attr_merge_normalize_one( e_backend, mi->mi_ad_monitoredInfo,
 				&bv, NULL );
 
+		attr_merge_normalize_one( e, mi->mi_ad_monitorRuntimeConfig,
+			bi->bi_cf_ocs == NULL ? (struct berval *)&slap_false_bv :
+				(struct berval *)&slap_true_bv, NULL );
+
 		if ( bi->bi_controls ) {
 			int j;
 
 			for ( j = 0; bi->bi_controls[ j ]; j++ ) {
-				bv.bv_val = bi->bi_controls[ j ];
-				bv.bv_len = strlen( bv.bv_val );
-				attr_merge_one( e, slap_schema.si_ad_supportedControl, &bv, NULL );
+				ber_str2bv( bi->bi_controls[ j ], 0, 0, &bv );
+				attr_merge_one( e, slap_schema.si_ad_supportedControl,
+						&bv, &bv );
 			}
 		}
 
-		for ( j = 0; j < nBackendDB; j++ ) {
-			BackendDB	*be = &backendDB[ j ];
+		j = -1;
+		LDAP_STAILQ_FOREACH( be, &backendDB, be_next ) {
 			char		buf[ SLAP_LDAPDN_MAXLEN ];
 			struct berval	dn;
 			
+			j++;
+
 			if ( be->bd_info != bi ) {
 				continue;
 			}
 
 			snprintf( buf, sizeof( buf ), "cn=Database %d,%s",
 					j, ms_database->mss_dn.bv_val );
-			dn.bv_val = buf;
-			dn.bv_len = strlen( buf );
 
-			attr_merge_normalize_one( e, mi->mi_ad_seeAlso,
+			ber_str2bv( buf, 0, 0, &dn );
+			attr_merge_normalize_one( e, slap_schema.si_ad_seeAlso,
 					&dn, NULL );
 		}
 		
