@@ -1,8 +1,8 @@
 /* compare.c - monitor backend compare routine */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-monitor/compare.c,v 1.20.2.1 2005/01/20 18:04:00 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-monitor/compare.c,v 1.24.2.1 2006/04/18 17:33:11 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2005 The OpenLDAP Foundation.
+ * Copyright 2001-2006 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -32,13 +32,21 @@ monitor_back_compare( struct slap_op *op, struct slap_rep *rs)
 	monitor_info_t	*mi = ( monitor_info_t * ) op->o_bd->be_private;
 	Entry           *e, *matched = NULL;
 	Attribute	*a;
+	int		rc;
 
 	/* get entry with reader lock */
-	monitor_cache_dn2entry( op, &op->o_req_ndn, &e, &matched );
+	monitor_cache_dn2entry( op, rs, &op->o_req_ndn, &e, &matched );
 	if ( e == NULL ) {
 		rs->sr_err = LDAP_NO_SUCH_OBJECT;
 		if ( matched ) {
-			rs->sr_matched = matched->e_dn;
+			if ( !access_allowed_mask( op, matched,
+					slap_schema.si_ad_entry,
+					NULL, ACL_DISCLOSE, NULL, NULL ) )
+			{
+				/* do nothing */ ;
+			} else {
+				rs->sr_matched = matched->e_dn;
+			}
 		}
 		send_ldap_result( op, rs );
 		if ( matched ) {
@@ -46,7 +54,7 @@ monitor_back_compare( struct slap_op *op, struct slap_rep *rs)
 			rs->sr_matched = NULL;
 		}
 
-		return( 0 );
+		return rs->sr_err;
 	}
 
 	rs->sr_err = access_allowed( op, e, op->oq_compare.rs_ava->aa_desc,
@@ -75,14 +83,30 @@ monitor_back_compare( struct slap_op *op, struct slap_rep *rs)
 	}
 
 return_results:;
-	send_ldap_result( op, rs );
-	if ( rs->sr_err == LDAP_COMPARE_FALSE
-			|| rs->sr_err == LDAP_COMPARE_TRUE ) {
-		rs->sr_err = LDAP_SUCCESS;
+	rc = rs->sr_err;
+	switch ( rc ) {
+	case LDAP_COMPARE_FALSE:
+	case LDAP_COMPARE_TRUE:
+		rc = LDAP_SUCCESS;
+		break;
+
+	case LDAP_NO_SUCH_ATTRIBUTE:
+		break;
+
+	default:
+		if ( !access_allowed_mask( op, e, slap_schema.si_ad_entry,
+				NULL, ACL_DISCLOSE, NULL, NULL ) )
+		{
+			rs->sr_err = LDAP_NO_SUCH_OBJECT;
+		}
+		break;
 	}
+		
+	send_ldap_result( op, rs );
+	rs->sr_err = rc;
 
 	monitor_cache_release( mi, e );
 
-	return( rs->sr_err );
+	return rs->sr_err;
 }
 
