@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-sql/schema-map.c,v 1.44.2.2 2005/01/24 21:28:53 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-sql/schema-map.c,v 1.59 2006/01/03 22:12:24 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2005 The OpenLDAP Foundation.
+ * Copyright 1999-2006 The OpenLDAP Foundation.
  * Portions Copyright 1999 Dmitry Kovalev.
  * Portions Copyright 2002 Pierangelo Masarati.
  * Portions Copyright 2004 Mark Adamson.
@@ -27,6 +27,7 @@
 #include <sys/types.h>
 #include "ac/string.h"
 
+#include "lutil.h"
 #include "slap.h"
 #include "proto-sql.h"
 
@@ -100,19 +101,21 @@ backsql_dup_attr( void *v_m1, void *v_m2 )
 
 static int
 backsql_make_attr_query( 
+	backsql_info		*bi,
 	backsql_oc_map_rec 	*oc_map,
 	backsql_at_map_rec 	*at_map )
 {
 	struct berbuf	bb = BB_NULL;
 
-	backsql_strfcat( &bb, "lblblblbcbl", 
+	backsql_strfcat_x( &bb, NULL, "lblbbbblblbcbl", 
 			(ber_len_t)STRLENOF( "SELECT " ), "SELECT ", 
 			&at_map->bam_sel_expr, 
-			(ber_len_t)STRLENOF( " " BACKSQL_ALIASING BACKSQL_ALIASING_QUOTE ),
-				" " BACKSQL_ALIASING BACKSQL_ALIASING_QUOTE, 
+			(ber_len_t)STRLENOF( " " ), " ",
+			&bi->sql_aliasing,
+			&bi->sql_aliasing_quote, 
 			&at_map->bam_ad->ad_cname,
-			(ber_len_t)STRLENOF( BACKSQL_ALIASING_QUOTE " FROM " ),
-				BACKSQL_ALIASING_QUOTE " FROM ", 
+			&bi->sql_aliasing_quote,
+			(ber_len_t)STRLENOF( " FROM " ), " FROM ", 
 			&at_map->bam_from_tbls, 
 			(ber_len_t)STRLENOF( " WHERE " ), " WHERE ", 
 			&oc_map->bom_keytbl,
@@ -121,17 +124,16 @@ backsql_make_attr_query(
 			(ber_len_t)STRLENOF( "=?" ), "=?" );
 
 	if ( !BER_BVISNULL( &at_map->bam_join_where ) ) {
-		backsql_strfcat( &bb, "lb",
+		backsql_strfcat_x( &bb, NULL, "lb",
 				(ber_len_t)STRLENOF( " AND " ), " AND ", 
 				&at_map->bam_join_where );
 	}
 
-	backsql_strfcat( &bb, "lbl", 
-			(ber_len_t)STRLENOF( " ORDER BY " BACKSQL_ALIASING_QUOTE ),
-				" ORDER BY " BACKSQL_ALIASING_QUOTE,
-			&at_map->bam_sel_expr,
-			(ber_len_t)STRLENOF( BACKSQL_ALIASING_QUOTE ),
-				BACKSQL_ALIASING_QUOTE );
+	backsql_strfcat_x( &bb, NULL, "lbbb", 
+			(ber_len_t)STRLENOF( " ORDER BY " ), " ORDER BY ",
+			&bi->sql_aliasing_quote,
+			&at_map->bam_ad->ad_cname,
+			&bi->sql_aliasing_quote );
 
 	at_map->bam_query = bb.bb_val.bv_val;
 
@@ -139,7 +141,7 @@ backsql_make_attr_query(
 	/* Query to count how many rows will be returned. */
 	BER_BVZERO( &bb.bb_val );
 	bb.bb_len = 0;
-	backsql_strfcat( &bb, "lblbcbl", 
+	backsql_strfcat_x( &bb, NULL, "lblbcbl", 
 			(ber_len_t)STRLENOF( "SELECT COUNT(*) FROM " ),
 				"SELECT COUNT(*) FROM ", 
 			&at_map->bam_from_tbls, 
@@ -150,7 +152,7 @@ backsql_make_attr_query(
 			(ber_len_t)STRLENOF( "=?" ), "=?" );
 
 	if ( !BER_BVISNULL( &at_map->bam_join_where ) ) {
-		backsql_strfcat( &bb, "lb",
+		backsql_strfcat_x( &bb, NULL, "lb",
 				(ber_len_t)STRLENOF( " AND " ), " AND ", 
 				&at_map->bam_join_where );
 	}
@@ -162,7 +164,7 @@ backsql_make_attr_query(
 }
 
 static int
-backsql_add_sysmaps( backsql_oc_map_rec *oc_map )
+backsql_add_sysmaps( backsql_info *bi, backsql_oc_map_rec *oc_map )
 {
 	backsql_at_map_rec	*at_map;
 	char			s[] = "+9223372036854775807L";
@@ -183,12 +185,12 @@ backsql_add_sysmaps( backsql_oc_map_rec *oc_map )
 	
 	bb.bb_len = at_map->bam_from_tbls.bv_len + 1;
 	bb.bb_val = at_map->bam_from_tbls;
-	backsql_merge_from_clause( &bb, &oc_map->bom_keytbl );
+	backsql_merge_from_clause( bi, &bb, &oc_map->bom_keytbl );
 	at_map->bam_from_tbls = bb.bb_val;
 
 	BER_BVZERO( &bb.bb_val );
 	bb.bb_len = 0;
-	backsql_strfcat( &bb, "lbcblb",
+	backsql_strfcat_x( &bb, NULL, "lbcblb",
 			(ber_len_t)STRLENOF( "ldap_entries.id=ldap_entry_objclasses.entry_id AND ldap_entries.keyval=" ),
 				"ldap_entries.id=ldap_entry_objclasses.entry_id AND ldap_entries.keyval=",
 			&oc_map->bom_keytbl, 
@@ -230,7 +232,7 @@ backsql_add_sysmaps( backsql_oc_map_rec *oc_map )
 		snprintf( tmp, sizeof(tmp), 
 			"DELETE FROM ldap_entry_objclasses "
 			"WHERE entry_id=(SELECT id FROM ldap_entries "
-			"WHERE oc_map_id=%lu"
+			"WHERE oc_map_id=%lu "
 			"AND keyval=?) AND oc_name=?",
 			oc_map->bom_id );
 		at_map->bam_delete_proc = ch_strdup( tmp );
@@ -240,7 +242,7 @@ backsql_add_sysmaps( backsql_oc_map_rec *oc_map )
 	at_map->bam_expect_return = 0;
 	at_map->bam_next = NULL;
 
-	backsql_make_attr_query( oc_map, at_map );
+	backsql_make_attr_query( bi, oc_map, at_map );
 	if ( avl_insert( &oc_map->bom_attrs, at_map, backsql_cmp_attr, backsql_dup_attr ) == BACKSQL_DUPLICATE ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_add_sysmaps(): "
 				"duplicate attribute \"%s\" in objectClass \"%s\" map\n",
@@ -253,7 +255,7 @@ backsql_add_sysmaps( backsql_oc_map_rec *oc_map )
 	ch_free( at_map->bam_join_where.bv_val );
 	BER_BVZERO( &bb.bb_val );
 	bb.bb_len = 0;
-	backsql_strfcat( &bb, "lbcblb",
+	backsql_strfcat_x( &bb, NULL, "lbcblb",
 			(ber_len_t)STRLENOF( /* "ldap_entries.id=ldap_entry_objclasses.entry_id AND " */ "ldap_entries.keyval=" ),
 				/* "ldap_entries.id=ldap_entry_objclasses.entry_id AND " */ "ldap_entries.keyval=",
 			&oc_map->bom_keytbl, 
@@ -315,7 +317,6 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 	backsql_BindRowAsStrings( bas->bas_sth, &at_row );
 	for ( ; rc = SQLFetch( bas->bas_sth ), BACKSQL_SUCCESS( rc ); ) {
 		const char	*text = NULL;
-		char		*next = NULL;
 		struct berval	bv;
 		struct berbuf	bb = BB_NULL;
 
@@ -359,7 +360,7 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 		}
 
 		ber_str2bv( at_row.cols[ 2 ], 0, 0, &bv );
-		backsql_merge_from_clause( &bb, &bv );
+		backsql_merge_from_clause( bas->bas_bi, &bb, &bv );
 		at_map->bam_from_tbls = bb.bb_val;
 		if ( at_row.value_len[ 3 ] < 0 ) {
 			BER_BVZERO( &at_map->bam_join_where );
@@ -376,17 +377,13 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 		if ( at_row.value_len[ 5 ] > 0 ) {
 			at_map->bam_delete_proc = ch_strdup( at_row.cols[ 5 ] );
 		}
-		at_map->bam_param_order = strtol( at_row.cols[ 6 ], 
-				&next, 0 );
-		if ( next == at_row.cols[ 6 ] || next[0] != '\0' ) {
+		if ( lutil_atoix( &at_map->bam_param_order, at_row.cols[ 6 ], 0 ) != 0 ) {
 			/* error */
 		}
-		at_map->bam_expect_return = strtol( at_row.cols[ 7 ],
-				&next, 0 );
-		if ( next == at_row.cols[ 7 ] || next[0] != '\0' ) {
+		if ( lutil_atoix( &at_map->bam_expect_return, at_row.cols[ 7 ], 0 ) != 0 ) {
 			/* error */
 		}
-		backsql_make_attr_query( oc_map, at_map );
+		backsql_make_attr_query( bas->bas_bi, oc_map, at_map );
 		Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_attr_mapping(): "
 			"preconstructed query \"%s\"\n",
 			at_map->bam_query, 0, 0 );
@@ -404,7 +401,7 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 		{
 			struct berbuf	bb = BB_NULL;
 
-			backsql_strfcat( &bb, "bcbc",
+			backsql_strfcat_x( &bb, NULL, "bcbc",
 					&bas->bas_bi->sql_upper_func,
 					'(' /* ) */ ,
 					&at_map->bam_sel_expr,
@@ -419,7 +416,7 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 		"autoadding 'objectClass' and 'ref' mappings\n",
 		BACKSQL_OC_NAME( oc_map ), 0, 0 );
 
-	(void)backsql_add_sysmaps( oc_map );
+	(void)backsql_add_sysmaps( bas->bas_bi, oc_map );
 
 	return BACKSQL_AVL_CONTINUE;
 }
@@ -484,7 +481,12 @@ backsql_load_schema_map( backsql_info *bi, SQLHDBC dbh )
 		oc_map = (backsql_oc_map_rec *)ch_calloc( 1,
 				sizeof( backsql_oc_map_rec ) );
 
-		oc_map->bom_id = strtol( oc_row.cols[ 0 ], NULL, 0 );
+		if ( lutil_atoulx( &oc_map->bom_id, oc_row.cols[ 0 ], 0 ) != 0 ) {
+			Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): "
+				"unable to parse id=\"%s\"\n", 
+				oc_row.cols[ 0 ], 0, 0 );
+			return LDAP_OTHER;
+		}
 
 		oc_map->bom_oc = oc_find( oc_row.cols[ 1 ] );
 		if ( oc_map->bom_oc == NULL ) {
@@ -507,8 +509,12 @@ backsql_load_schema_map( backsql_info *bi, SQLHDBC dbh )
 		}
 		oc_map->bom_delete_proc = ( oc_row.value_len[ colnum ] < 0 ) ? NULL 
 			: ch_strdup( oc_row.cols[ colnum ] );
-		oc_map->bom_expect_return = strtol( oc_row.cols[ colnum + 1 ], 
-				NULL, 0 );
+		if ( lutil_atoix( &oc_map->bom_expect_return, oc_row.cols[ colnum + 1 ], 0 ) != 0 ) {
+			Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): "
+				"unable to parse expect_return=\"%s\" for objectClass \"%s\"\n", 
+				oc_row.cols[ colnum + 1 ], oc_row.cols[ 1 ], 0 );
+			return LDAP_OTHER;
+		}
 
 		colnum += 2;
 		if ( ( oc_row.ncols > colnum ) &&
@@ -799,9 +805,9 @@ backsql_supad2at( backsql_oc_map_rec *objclass, AttributeDescription *supad,
 	struct supad2at_t	va = { 0 };
 	int			rc;
 
-	assert( objclass );
-	assert( supad );
-	assert( pret );
+	assert( objclass != NULL );
+	assert( supad != NULL );
+	assert( pret != NULL );
 
 	*pret = NULL;
 
@@ -842,6 +848,12 @@ backsql_free_attr( void *v_at )
 		ch_free( at->bam_query );
 	}
 
+#ifdef BACKSQL_COUNTQUERY
+	if ( at->bam_countquery != NULL ) {
+		ch_free( at->bam_countquery );
+	}
+#endif /* BACKSQL_COUNTQUERY */
+
 	/* TimesTen */
 	if ( !BER_BVISNULL( &at->bam_sel_expr_u ) ) {
 		ch_free( at->bam_sel_expr_u.bv_val );
