@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/delete.c,v 1.126.2.2 2005/01/24 21:28:51 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/delete.c,v 1.138 2006/03/30 15:59:26 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2005 The OpenLDAP Foundation.
+ * Copyright 1998-2006 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -34,10 +34,6 @@
 
 #include "lutil.h"
 
-#ifdef LDAP_SLAPI
-#include "slapi/slapi.h"
-#endif
-
 int
 do_delete(
     Operation	*op,
@@ -94,9 +90,14 @@ do_delete(
 	op->o_bd = frontendDB;
 	rs->sr_err = frontendDB->be_delete( op, rs );
 
-cleanup:;
-	slap_graduate_commit_csn( op );
+#ifdef LDAP_X_TXN
+	if( rs->sr_err == LDAP_X_TXN_SPECIFY_OKAY ) {
+		/* skip cleanup */
+		return rs->sr_err;
+	}
+#endif
 
+cleanup:;
 	op->o_tmpfree( op->o_req_dn.bv_val, op->o_tmpmemctx );
 	op->o_tmpfree( op->o_req_ndn.bv_val, op->o_tmpmemctx );
 	return rs->sr_err;
@@ -107,6 +108,7 @@ fe_op_delete( Operation *op, SlapReply *rs )
 {
 	struct berval	pdn = BER_BVNULL;
 	int		manageDSAit;
+	BackendDB	*op_be, *bd = op->o_bd;
 	
 	manageDSAit = get_manageDSAit( op );
 
@@ -117,27 +119,29 @@ fe_op_delete( Operation *op, SlapReply *rs )
 	 */
 	op->o_bd = select_backend( &op->o_req_ndn, manageDSAit, 1 );
 	if ( op->o_bd == NULL ) {
+		op->o_bd = bd;
 		rs->sr_ref = referral_rewrite( default_referral,
 			NULL, &op->o_req_dn, LDAP_SCOPE_DEFAULT );
 
 		if (!rs->sr_ref) rs->sr_ref = default_referral;
 		if ( rs->sr_ref != NULL ) {
 			rs->sr_err = LDAP_REFERRAL;
-
-			op->o_bd = frontendDB;
 			send_ldap_result( op, rs );
-			op->o_bd = NULL;
 
 			if (rs->sr_ref != default_referral) ber_bvarray_free( rs->sr_ref );
 		} else {
-			op->o_bd = frontendDB;
 			send_ldap_error( op, rs, LDAP_UNWILLING_TO_PERFORM,
 				"no global superior knowledge" );
-			op->o_bd = NULL;
 		}
 		goto cleanup;
 	}
 
+	/* If we've got a glued backend, check the real backend */
+	op_be = op->o_bd;
+	if ( SLAP_GLUE_INSTANCE( op->o_bd )) {
+		op->o_bd = select_backend( &op->o_req_ndn, manageDSAit, 0 );
+	}
+
 	/* check restrictions */
 	if( backend_check_restrictions( op, rs, NULL ) != LDAP_SUCCESS ) {
 		send_ldap_result( op, rs );
@@ -149,33 +153,6 @@ fe_op_delete( Operation *op, SlapReply *rs )
 		goto cleanup;
 	}
 
-#if defined( LDAP_SLAPI )
-#define pb op->o_pb
-	if ( pb ) {
-		slapi_int_pblock_set_operation( pb, op );
-		slapi_pblock_set( pb, SLAPI_DELETE_TARGET, (void *)op->o_req_dn.bv_val );
-		slapi_pblock_set( pb, SLAPI_MANAGEDSAIT, (void *)manageDSAit );
-
-		rs->sr_err = slapi_int_call_plugins( op->o_bd,
-			SLAPI_PLUGIN_PRE_DELETE_FN, pb );
-		if ( rs->sr_err < 0 ) {
-			/*
-			 * A preoperation plugin failure will abort the
-			 * entire operation.
-			 */
-			Debug (LDAP_DEBUG_TRACE, "do_delete: "
-				"delete preoperation plugin failed.\n", 0, 0, 0);
-			if ( ( slapi_pblock_get( pb, SLAPI_RESULT_CODE,
-				(void *)&rs->sr_err ) != 0 ) ||
-				rs->sr_err == LDAP_SUCCESS )
-			{
-				rs->sr_err = LDAP_OTHER;
-			}
-			goto cleanup;
-		}
-	}
-#endif /* defined( LDAP_SLAPI ) */
-
 	/*
 	 * do the delete if 1 && (2 || 3)
 	 * 1) there is a delete function implemented in this backend;
@@ -185,10 +162,7 @@ fe_op_delete( Operation *op, SlapReply *rs )
 	if ( op->o_bd->be_delete ) {
 		/* do the update here */
 		int repl_user = be_isupdate( op );
-#ifndef SLAPD_MULTIMASTER
-		if ( !SLAP_SHADOW(op->o_bd) || repl_user )
-#endif
-		{
+		if ( !SLAP_SINGLE_SHADOW(op->o_bd) || repl_user ) {
 			struct berval	org_req_dn = BER_BVNULL;
 			struct berval	org_req_ndn = BER_BVNULL;
 			struct berval	org_dn = BER_BVNULL;
@@ -196,16 +170,9 @@ fe_op_delete( Operation *op, SlapReply *rs )
 			int		org_managedsait;
 			slap_callback 	cb = { NULL, slap_replog_cb, NULL, NULL };
 
-			if ( !repl_user ) {
-				struct berval csn = BER_BVNULL;
-				char csnbuf[LDAP_LUTIL_CSNSTR_BUFSIZE];
-				slap_get_csn( op, csnbuf, sizeof(csnbuf), &csn, 1 );
-			}
+			op->o_bd = op_be;
 
-#ifdef SLAPD_MULTIMASTER
-			if ( !op->o_bd->be_update_ndn.bv_len || !repl_user )
-#endif
-			{
+			if ( !op->o_bd->be_update_ndn.bv_len || !repl_user ) {
 				cb.sc_next = op->o_callback;
 				op->o_callback = &cb;
 			}
@@ -226,7 +193,7 @@ fe_op_delete( Operation *op, SlapReply *rs )
 			{
 				op->o_delete_glue_parent = 0;
 				if ( !be_issuffix( op->o_bd, &op->o_req_ndn )) {
-					slap_callback cb = { NULL };
+					slap_callback cb = { NULL, NULL, NULL, NULL };
 					cb.sc_response = slap_null_cb;
 					dnParent( &op->o_req_ndn, &pdn );
 					op->o_req_dn = pdn;
@@ -245,7 +212,6 @@ fe_op_delete( Operation *op, SlapReply *rs )
 			op->o_req_ndn = org_req_ndn;
 			op->o_delete_glue_parent = 0;
 
-#ifndef SLAPD_MULTIMASTER
 		} else {
 			BerVarray defref = op->o_bd->be_update_refs
 				? op->o_bd->be_update_refs : default_referral;
@@ -264,7 +230,6 @@ fe_op_delete( Operation *op, SlapReply *rs )
 					LDAP_UNWILLING_TO_PERFORM,
 					"shadow context; no update referral" );
 			}
-#endif
 		}
 
 	} else {
@@ -272,16 +237,7 @@ fe_op_delete( Operation *op, SlapReply *rs )
 			"operation not supported within namingContext" );
 	}
 
-#if defined( LDAP_SLAPI )
-	if ( pb != NULL && slapi_int_call_plugins( op->o_bd,
-		SLAPI_PLUGIN_POST_DELETE_FN, pb ) < 0)
-	{
-		Debug(LDAP_DEBUG_TRACE,
-			"do_delete: delete postoperation plugins failed\n",
-			0, 0, 0 );
-	}
-#endif /* defined( LDAP_SLAPI ) */
-
 cleanup:;
+	op->o_bd = bd;
 	return rs->sr_err;
 }
