@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/main.c,v 1.198.2.2 2005/01/28 17:29:48 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/main.c,v 1.239.2.2 2006/04/18 17:33:08 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2005 The OpenLDAP Foundation.
+ * Copyright 1998-2006 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -27,6 +27,7 @@
 
 #include <stdio.h>
 
+#include <ac/ctype.h>
 #include <ac/socket.h>
 #include <ac/string.h>
 #include <ac/time.h>
@@ -97,43 +98,17 @@ const char Versionstr[] =
 	OPENLDAP_PACKAGE " " OPENLDAP_VERSION " Standalone LDAP Server (slapd)";
 #endif
 
-#ifdef LOG_LOCAL4
-#define DEFAULT_SYSLOG_USER  LOG_LOCAL4
-
-typedef struct _str2intDispatch {
-	char	*stringVal;
-	int	 abbr;
-	int	 intVal;
-} STRDISP, *STRDISP_P;
-
-/* table to compute syslog-options to integer */
-static STRDISP	syslog_types[] = {
-	{ "LOCAL0", sizeof("LOCAL0"), LOG_LOCAL0 },
-	{ "LOCAL1", sizeof("LOCAL1"), LOG_LOCAL1 },
-	{ "LOCAL2", sizeof("LOCAL2"), LOG_LOCAL2 },
-	{ "LOCAL3", sizeof("LOCAL3"), LOG_LOCAL3 },
-	{ "LOCAL4", sizeof("LOCAL4"), LOG_LOCAL4 },
-	{ "LOCAL5", sizeof("LOCAL5"), LOG_LOCAL5 },
-	{ "LOCAL6", sizeof("LOCAL6"), LOG_LOCAL6 },
-	{ "LOCAL7", sizeof("LOCAL7"), LOG_LOCAL7 },
-#ifdef LOG_USER
-	{ "USER", sizeof("USER"), LOG_USER },
-#endif
-#ifdef LOG_DAEMON
-	{ "DAEMON", sizeof("DAEMON"), LOG_DAEMON },
-#endif
-	{ NULL, 0, 0 }
-};
-
-static int cnvt_str2int( char *, STRDISP_P, int );
-#endif	/* LOG_LOCAL4 */
-
 #define CHECK_NONE	0x00
 #define CHECK_CONFIG	0x01
 static int check = CHECK_NONE;
 static int version = 0;
 
 void *slap_tls_ctx;
+LDAP *slap_tls_ld;
+
+#ifdef LOG_LOCAL4
+#define DEFAULT_SYSLOG_USER	LOG_LOCAL4
+#endif /* LOG_LOCAL4 */
 
 static int
 slapd_opt_slp( const char *val, void *arg )
@@ -181,9 +156,137 @@ struct option_helper {
 	const char	*oh_usage;
 } option_helpers[] = {
 	{ BER_BVC("slp"),	slapd_opt_slp,	NULL, "slp[={on|off}] enable/disable SLP" },
-	{ BER_BVNULL }
+	{ BER_BVNULL, 0, NULL, NULL }
 };
 
+#ifdef LOG_LOCAL4
+static int
+parse_syslog_user( const char *arg, int *syslogUser )
+{
+	static slap_verbmasks syslogUsers[] = {
+		{ BER_BVC( "LOCAL0" ), LOG_LOCAL0 },
+		{ BER_BVC( "LOCAL1" ), LOG_LOCAL1 },
+		{ BER_BVC( "LOCAL2" ), LOG_LOCAL2 },
+		{ BER_BVC( "LOCAL3" ), LOG_LOCAL3 },
+		{ BER_BVC( "LOCAL4" ), LOG_LOCAL4 },
+		{ BER_BVC( "LOCAL5" ), LOG_LOCAL5 },
+		{ BER_BVC( "LOCAL6" ), LOG_LOCAL6 },
+		{ BER_BVC( "LOCAL7" ), LOG_LOCAL7 },
+#ifdef LOG_USER
+		{ BER_BVC( "USER" ), LOG_USER },
+#endif /* LOG_USER */
+#ifdef LOG_DAEMON
+		{ BER_BVC( "DAEMON" ), LOG_DAEMON },
+#endif /* LOG_DAEMON */
+		{ BER_BVNULL, 0 }
+	};
+	int i = verb_to_mask( optarg, syslogUsers );
+
+	if ( BER_BVISNULL( &syslogUsers[ i ].word ) ) {
+		Debug( LDAP_DEBUG_ANY,
+			"unrecognized syslog user \"%s\".\n",
+			optarg, 0, 0 );
+		return 1;
+	}
+
+	*syslogUser = syslogUsers[ i ].mask;
+
+	return 0;
+}
+#endif /* LOG_LOCAL4 */
+
+static int
+parse_syslog_level( const char *arg, int *levelp )
+{
+	static slap_verbmasks	str2syslog_level[] = {
+		{ BER_BVC( "EMERG" ),	LOG_EMERG },
+		{ BER_BVC( "ALERT" ),	LOG_ALERT },
+		{ BER_BVC( "CRIT" ),	LOG_CRIT },
+		{ BER_BVC( "ERR" ),	LOG_ERR },
+		{ BER_BVC( "WARNING" ),	LOG_WARNING },
+		{ BER_BVC( "NOTICE" ),	LOG_NOTICE },
+		{ BER_BVC( "INFO" ),	LOG_INFO },
+		{ BER_BVC( "DEBUG" ),	LOG_DEBUG },
+		{ BER_BVNULL, 0 }
+	};
+	int i = verb_to_mask( arg, str2syslog_level );
+	if ( BER_BVISNULL( &str2syslog_level[ i ].word ) ) {
+		Debug( LDAP_DEBUG_ANY,
+			"unknown syslog level \"%s\".\n",
+			arg, 0, 0 );
+		return 1;
+	}
+	
+	*levelp = str2syslog_level[ i ].mask;
+
+	return 0;
+}
+
+int
+parse_debug_unknowns( char **unknowns, int *levelp )
+{
+	int i, level, rc = 0;
+
+	for ( i = 0; unknowns[ i ] != NULL; i++ ) {
+		level = 0;
+		if ( str2loglevel( unknowns[ i ], &level )) {
+			fprintf( stderr,
+				"unrecognized log level \"%s\"\n", unknowns[ i ] );
+			rc = 1;
+		} else {
+			*levelp |= level;
+		}
+	}
+	return rc;
+}
+
+int
+parse_debug_level( const char *arg, int *levelp, char ***unknowns )
+{
+	int	level;
+
+	if ( arg && arg[ 0 ] != '-' && !isdigit( (unsigned char) arg[ 0 ] ) )
+	{
+		int	i;
+		char	**levels;
+
+		levels = ldap_str2charray( arg, "," );
+
+		for ( i = 0; levels[ i ] != NULL; i++ ) {
+			level = 0;
+
+			if ( str2loglevel( levels[ i ], &level ) ) {
+				/* remember this for later */
+				ldap_charray_add( unknowns, levels[ i ] );
+				fprintf( stderr,
+					"unrecognized log level \"%s\" (deferred)\n",
+					levels[ i ] );
+			} else {
+				*levelp |= level;
+			}
+		}
+
+		ldap_charray_free( levels );
+
+	} else {
+		if ( lutil_atoix( &level, arg, 0 ) != 0 ) {
+			fprintf( stderr,
+				"unrecognized log level "
+				"\"%s\"\n", arg );
+			return 1;
+		}
+
+		if ( level == 0 ) {
+			*levelp = 0;
+
+		} else {
+			*levelp |= level;
+		}
+	}
+
+	return 0;
+}
+
 static void
 usage( char *name )
 {
@@ -192,11 +295,12 @@ usage( char *name )
 	fprintf( stderr,
 		"\t-4\t\tIPv4 only\n"
 		"\t-6\t\tIPv6 only\n"
-		"\t-T {add|auth|cat|dn|index|passwd|test}\n"
+		"\t-T {acl|add|auth|cat|dn|index|passwd|test}\n"
 		"\t\t\tRun in Tool mode\n"
 		"\t-c cookie\tSync cookie of consumer\n"
 		"\t-d level\tDebug level" "\n"
 		"\t-f filename\tConfiguration file\n"
+		"\t-F dir\tConfiguration directory\n"
 #if defined(HAVE_SETUID) && defined(HAVE_SETGID)
 		"\t-g group\tGroup (id or name) to run as\n"
 #endif
@@ -245,34 +349,39 @@ int main( int argc, char **argv )
 	char *sandbox = NULL;
 #endif
 #ifdef LOG_LOCAL4
-    int	    syslogUser = DEFAULT_SYSLOG_USER;
+	int syslogUser = DEFAULT_SYSLOG_USER;
 #endif
 	
 	int g_argc = argc;
 	char **g_argv = argv;
 
-#ifdef HAVE_NT_SERVICE_MANAGER
-	char		*configfile = ".\\slapd.conf";
-#else
-	char		*configfile = SLAPD_DEFAULT_CONFIGFILE;
-#endif
-	char	    *serverName;
-	int	    serverMode = SLAP_SERVER_MODE;
+	char *configfile = NULL;
+	char *configdir = NULL;
+	char *serverName;
+	int serverMode = SLAP_SERVER_MODE;
 
 	struct sync_cookie *scp = NULL;
 	struct sync_cookie *scp_entry = NULL;
 
+	char **debug_unknowns = NULL;
+	char **syslog_unknowns = NULL;
+
+	char *serverNamePrefix = "";
+	size_t	l;
+
+	int slapd_pid_file_unlink = 0, slapd_args_file_unlink = 0;
+
 #ifdef CSRIMALLOC
 	FILE *leakfile;
 	if( ( leakfile = fopen( "slapd.leak", "w" )) == NULL ) {
 		leakfile = stderr;
 	}
 #endif
-	char	*serverNamePrefix = "";
-	size_t	l;
 
 	slap_sl_mem_init();
 
+	(void) ldap_pvt_thread_initialize();
+
 	serverName = lutil_progname( "slapd", argc, argv );
 
 	if ( strcmp( serverName, "slapd" ) ) {
@@ -288,6 +397,7 @@ int main( int argc, char **argv )
 	{
 		int *i;
 		char *newConfigFile;
+		char *newConfigDir;
 		char *newUrls;
 		char *regService = NULL;
 
@@ -319,22 +429,31 @@ int main( int argc, char **argv )
 			configfile = newConfigFile;
 			Debug ( LDAP_DEBUG_ANY, "new config file from registry is: %s\n", configfile, 0, 0 );
 		}
+
+		newConfigDir = (char*)lutil_getRegParam( regService, "ConfigDir" );
+		if ( newConfigDir != NULL ) {
+			configdir = newConfigDir;
+			Debug ( LDAP_DEBUG_ANY, "new config dir from registry is: %s\n", configdir, 0, 0 );
+		}
 	}
 #endif
 
 	while ( (i = getopt( argc, argv,
-			     "c:d:f:h:n:o:s:StT:V"
+			     "c:d:f:F:h:n:o:s:tT:V"
 #if LDAP_PF_INET6
 				"46"
 #endif
 #ifdef HAVE_CHROOT
 				"r:"
 #endif
+#ifdef LDAP_SYSLOG
+				"S:"
+#endif
 #ifdef LOG_LOCAL4
-			     "l:"
+				"l:"
 #endif
 #if defined(HAVE_SETUID) && defined(HAVE_SETGID)
-			     "u:g:"
+				"u:g:"
 #endif
 			     )) != EOF ) {
 		switch ( i ) {
@@ -356,7 +475,17 @@ int main( int argc, char **argv )
 			scp = (struct sync_cookie *) ch_calloc( 1,
 										sizeof( struct sync_cookie ));
 			ber_str2bv( optarg, 0, 1, &scp->octet_str );
-			slap_parse_sync_cookie( scp );
+			
+			/* This only parses out the rid at this point */
+			slap_parse_sync_cookie( scp, NULL );
+
+			if ( scp->rid == -1 ) {
+				Debug( LDAP_DEBUG_ANY,
+						"main: invalid cookie \"%s\"\n",
+						optarg, 0, 0 );
+				slap_sync_cookie_free( scp, 1 );
+				goto destroy;
+			}
 
 			LDAP_STAILQ_FOREACH( scp_entry, &slap_sync_cookie, sc_next ) {
 				if ( scp->rid == scp_entry->rid ) {
@@ -370,21 +499,30 @@ int main( int argc, char **argv )
 			LDAP_STAILQ_INSERT_TAIL( &slap_sync_cookie, scp, sc_next );
 			break;
 
-		case 'd':	/* set debug level and 'do not detach' flag */
+		case 'd': {	/* set debug level and 'do not detach' flag */
+			int	level = 0;
+
 			no_detach = 1;
+			if ( parse_debug_level( optarg, &level, &debug_unknowns ) ) {
+				goto destroy;
+			}
 #ifdef LDAP_DEBUG
-			slap_debug |= atoi( optarg );
+			slap_debug |= level;
 #else
-			if ( atoi( optarg ) != 0 )
+			if ( level != 0 )
 				fputs( "must compile with LDAP_DEBUG for debugging\n",
 				       stderr );
 #endif
-			break;
+			} break;
 
 		case 'f':	/* read config file */
 			configfile = ch_strdup( optarg );
 			break;
 
+		case 'F':	/* use config dir */
+			configdir = ch_strdup( optarg );
+			break;
+
 		case 'o': {
 			char		*val = strchr( optarg, '=' );
 			struct berval	opt;
@@ -420,13 +558,24 @@ int main( int argc, char **argv )
 		}
 
 		case 's':	/* set syslog level */
-			ldap_syslog = atoi( optarg );
+			if ( parse_debug_level( optarg, &ldap_syslog, &syslog_unknowns ) ) {
+				goto destroy;
+			}
 			break;
 
+#if defined(LDAP_DEBUG) && defined(LDAP_SYSLOG)
+		case 'S':
+			if ( parse_syslog_level( optarg, &ldap_syslog_level ) ) {
+				goto destroy;
+			}
+			break;
+#endif /* LDAP_DEBUG && LDAP_SYSLOG */
+
 #ifdef LOG_LOCAL4
 		case 'l':	/* set syslog local user */
-			syslogUser = cnvt_str2int( optarg,
-				syslog_types, DEFAULT_SYSLOG_USER );
+			if ( parse_syslog_user( optarg, &syslogUser ) ) {
+				goto destroy;
+			}
 			break;
 #endif
 
@@ -497,8 +646,6 @@ unhandled_option:;
 		}
 	}
 
-	(void) ldap_pvt_thread_initialize();
-
 	ber_set_option(NULL, LBER_OPT_DEBUG_LEVEL, &slap_debug);
 	ldap_set_option(NULL, LDAP_OPT_DEBUG_LEVEL, &slap_debug);
 	ldif_debug = slap_debug;
@@ -560,71 +707,21 @@ unhandled_option:;
 	lutil_passwd_init();
 	slap_op_init();
 
-#ifdef SLAPD_MODULES
-	if ( module_init() != 0 ) {
-		rc = 1;
-		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 17 );
+#ifdef HAVE_TLS
+	rc = ldap_create( &slap_tls_ld );
+	if ( rc ) {
+		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 20 );
 		goto destroy;
 	}
 #endif
 
-	if ( slap_schema_init( ) != 0 ) {
-		Debug( LDAP_DEBUG_ANY,
-		    "schema initialization error\n",
-		    0, 0, 0 );
-
-		goto destroy;
-	}
-
-	if ( slap_init( serverMode, serverName ) != 0 ) {
-		rc = 1;
+	rc = slap_init( serverMode, serverName );
+	if ( rc ) {
 		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 18 );
 		goto destroy;
 	}
 
-	if ( slap_controls_init( ) != 0 ) {
-		Debug( LDAP_DEBUG_ANY,
-		    "controls initialization error\n",
-		    0, 0, 0 );
-
-		goto destroy;
-	}
-
-#ifdef HAVE_TLS
-	/* Library defaults to full certificate checking. This is correct when
-	 * a client is verifying a server because all servers should have a
-	 * valid cert. But few clients have valid certs, so we want our default
-	 * to be no checking. The config file can override this as usual.
-	 */
-	rc = 0;
-	(void) ldap_pvt_tls_set_option( NULL, LDAP_OPT_X_TLS_REQUIRE_CERT, &rc );
-#endif
-
-#ifdef LDAP_SLAPI
-	if ( slapi_int_initialize() != 0 ) {
-		Debug( LDAP_DEBUG_ANY,
-		    "slapi initialization error\n",
-		    0, 0, 0 );
-
-		goto destroy;
-	}
-#endif /* LDAP_SLAPI */
-
-	if ( frontend_init() ) {
-		goto destroy;
-	}
-
-	if ( overlay_init() ) {
-		goto destroy;
-	}
-
-#ifdef SLAP_DYNACL
-	if ( acl_init() ) {
-		goto destroy;
-	}
-#endif /* SLAP_DYNACL */
-
-	if ( read_config( configfile, 0 ) != 0 ) {
+	if ( read_config( configfile, configdir ) != 0 ) {
 		rc = 1;
 		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 19 );
 
@@ -635,6 +732,21 @@ unhandled_option:;
 		goto destroy;
 	}
 
+	if ( debug_unknowns ) {
+		rc = parse_debug_unknowns( debug_unknowns, &slap_debug );
+		ldap_charray_free( debug_unknowns );
+		debug_unknowns = NULL;
+		if ( rc )
+			goto destroy;
+	}
+	if ( syslog_unknowns ) {
+		rc = parse_debug_unknowns( syslog_unknowns, &ldap_syslog );
+		ldap_charray_free( syslog_unknowns );
+		syslog_unknowns = NULL;
+		if ( rc )
+			goto destroy;
+	}
+
 	if ( check & CHECK_CONFIG ) {
 		fprintf( stderr, "config check succeeded\n" );
 
@@ -645,6 +757,14 @@ unhandled_option:;
 		}
 	}
 
+	if ( glue_sub_attach( ) != 0 ) {
+		Debug( LDAP_DEBUG_ANY,
+		    "subordinate config error\n",
+		    0, 0, 0 );
+
+		goto destroy;
+	}
+
 	if ( slap_schema_check( ) != 0 ) {
 		Debug( LDAP_DEBUG_ANY,
 		    "schema prep error\n",
@@ -665,16 +785,15 @@ unhandled_option:;
 	}
 
 	{
-		void *def_ctx = NULL;
-
-		/* Save existing default ctx, if any */
-		ldap_pvt_tls_get_option( NULL, LDAP_OPT_X_TLS_CTX, &def_ctx );
+		int opt = 1;
 
 		/* Force new ctx to be created */
-		ldap_pvt_tls_set_option( NULL, LDAP_OPT_X_TLS_CTX, NULL );
-
-		rc = ldap_pvt_tls_init_def_ctx();
-		if( rc != 0) {
+		rc = ldap_pvt_tls_set_option( slap_tls_ld, LDAP_OPT_X_TLS_NEWCTX, &opt );
+		if( rc == 0 ) {
+			/* The ctx's refcount is bumped up here */
+			ldap_pvt_tls_get_option( slap_tls_ld, LDAP_OPT_X_TLS_CTX, &slap_tls_ctx );
+			load_extop( &slap_EXOP_START_TLS, 0, starttls_extop );
+		} else if ( rc != LDAP_NOT_SUPPORTED ) {
 			Debug( LDAP_DEBUG_ANY,
 			    "main: TLS init def ctx failed: %d\n",
 			    rc, 0, 0 );
@@ -682,10 +801,6 @@ unhandled_option:;
 			SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 20 );
 			goto destroy;
 		}
-		/* Retrieve slapd's own ctx */
-		ldap_pvt_tls_get_option( NULL, LDAP_OPT_X_TLS_CTX, &slap_tls_ctx );
-		/* Restore previous ctx */
-		ldap_pvt_tls_set_option( NULL, LDAP_OPT_X_TLS_CTX, def_ctx );
 	}
 #endif
 
@@ -719,13 +834,61 @@ unhandled_option:;
 	mal_leaktrace(1);
 #endif
 
+	if ( slapd_pid_file != NULL ) {
+		FILE *fp = fopen( slapd_pid_file, "w" );
+
+		if ( fp == NULL ) {
+			int save_errno = errno;
+
+			Debug( LDAP_DEBUG_ANY, "unable to open pid file "
+				"\"%s\": %d (%s)\n",
+				slapd_pid_file,
+				save_errno, strerror( save_errno ) );
+
+			free( slapd_pid_file );
+			slapd_pid_file = NULL;
+
+			rc = 1;
+			goto destroy;
+		}
+		fprintf( fp, "%d\n", (int) getpid() );
+		fclose( fp );
+		slapd_pid_file_unlink = 1;
+	}
+
+	if ( slapd_args_file != NULL ) {
+		FILE *fp = fopen( slapd_args_file, "w" );
+
+		if ( fp == NULL ) {
+			int save_errno = errno;
+
+			Debug( LDAP_DEBUG_ANY, "unable to open args file "
+				"\"%s\": %d (%s)\n",
+				slapd_args_file,
+				save_errno, strerror( save_errno ) );
+
+			free( slapd_args_file );
+			slapd_args_file = NULL;
+
+			rc = 1;
+			goto destroy;
+		}
+
+		for ( i = 0; i < g_argc; i++ ) {
+			fprintf( fp, "%s ", g_argv[i] );
+		}
+		fprintf( fp, "\n" );
+		fclose( fp );
+		slapd_args_file_unlink = 1;
+	}
+
 	/*
 	 * FIXME: moved here from slapd_daemon_task()
 	 * because back-monitor db_open() needs it
 	 */
 	time( &starttime );
 
-	if ( slap_startup( NULL )  != 0 ) {
+	if ( slap_startup( NULL ) != 0 ) {
 		rc = 1;
 		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 21 );
 		goto shutdown;
@@ -733,38 +896,10 @@ unhandled_option:;
 
 	Debug( LDAP_DEBUG_ANY, "slapd starting\n", 0, 0, 0 );
 
-
-	if ( slapd_pid_file != NULL ) {
-		FILE *fp = fopen( slapd_pid_file, "w" );
-
-		if( fp != NULL ) {
-			fprintf( fp, "%d\n", (int) getpid() );
-			fclose( fp );
-
-		} else {
-			free(slapd_pid_file);
-			slapd_pid_file = NULL;
-		}
-	}
-
-	if ( slapd_args_file != NULL ) {
-		FILE *fp = fopen( slapd_args_file, "w" );
-
-		if( fp != NULL ) {
-			for ( i = 0; i < g_argc; i++ ) {
-				fprintf( fp, "%s ", g_argv[i] );
-			}
-			fprintf( fp, "\n" );
-			fclose( fp );
-		} else {
-			free(slapd_args_file);
-			slapd_args_file = NULL;
-		}
-	}
-
 #ifdef HAVE_NT_EVENT_LOG
 	if (is_NT_Service)
-	lutil_LogStartedEvent( serverName, slap_debug, configfile, urls );
+	lutil_LogStartedEvent( serverName, slap_debug, configfile ?
+		configfile : SLAPD_DEFAULT_CONFIGFILE , urls );
 #endif
 
 	rc = slapd_daemon();
@@ -797,6 +932,9 @@ destroy:
 
 	extops_kill();
 
+	supported_feature_destroy();
+	entry_info_destroy();
+
 stop:
 #ifdef HAVE_NT_EVENT_LOG
 	if (is_NT_Service)
@@ -822,18 +960,32 @@ stop:
 	lutil_passwd_destroy();
 
 #ifdef HAVE_TLS
+	/* Setting it to itself decreases refcount, allowing it to be freed
+	 * when the LD is freed.
+	 */
+	if ( slap_tls_ld ) {
+		ldap_pvt_tls_set_option( slap_tls_ld, LDAP_OPT_X_TLS_CTX, slap_tls_ctx );
+		ldap_unbind( slap_tls_ld );
+	}
 	ldap_pvt_tls_destroy();
 #endif
 
-	if ( slapd_pid_file != NULL ) {
+	if ( slapd_pid_file_unlink ) {
 		unlink( slapd_pid_file );
 	}
-	if ( slapd_args_file != NULL ) {
+	if ( slapd_args_file_unlink ) {
 		unlink( slapd_args_file );
 	}
 
 	config_destroy();
 
+	if ( configfile )
+		ch_free( configfile );
+	if ( configdir )
+		ch_free( configdir );
+	if ( urls )
+		ch_free( urls );
+
 #ifdef CSRIMALLOC
 	mal_dumpleaktrace( leakfile );
 #endif
@@ -871,31 +1023,3 @@ wait4child( int sig )
 
 #endif /* LDAP_SIGCHLD */
 
-
-#ifdef LOG_LOCAL4
-
-/*
- *  Convert a string to an integer by means of a dispatcher table
- *  if the string is not in the table return the default
- */
-
-static int
-cnvt_str2int( char *stringVal, STRDISP_P dispatcher, int defaultVal )
-{
-    int	       retVal = defaultVal;
-    STRDISP_P  disp;
-
-    for (disp = dispatcher; disp->stringVal; disp++) {
-
-	if (!strncasecmp (stringVal, disp->stringVal, disp->abbr)) {
-
-	    retVal = disp->intVal;
-	    break;
-
-	}
-    }
-
-    return (retVal);
-}
-
-#endif	/* LOG_LOCAL4 */
