@@ -1,8 +1,8 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/ppolicy.c,v 1.31.2.2 2005/01/28 17:29:50 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/ppolicy.c,v 1.75 2006/01/19 21:16:23 hyc Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2004-2005 The OpenLDAP Foundation.
- * Portions Copyright 2004 Howard Chu, Symas Corporation.
+ * Copyright 2004-2006 The OpenLDAP Foundation.
+ * Portions Copyright 2004-2005 Howard Chu, Symas Corporation.
  * Portions Copyright 2004 Hewlett-Packard Company.
  * All rights reserved.
  *
@@ -23,19 +23,16 @@
 #include "portable.h"
 
 /* This file implements "Password Policy for LDAP Directories",
- * based on draft behera-ldap-password-policy-07
+ * based on draft behera-ldap-password-policy-09
  */
 
 #ifdef SLAPD_OVER_PPOLICY
 
-#if SLAPD_OVER_PPOLICY == SLAPD_MOD_DYNAMIC
-#define	SLAPD_IMPORT
-#endif
-
 #include <ldap.h>
 #include "lutil.h"
 #include "slap.h"
 #if SLAPD_MODULES
+#define LIBLTDL_DLL_IMPORT	/* Win32: don't re-export libltdl's symbols */
 #include <ltdl.h>
 #endif
 #include <ac/errno.h>
@@ -58,7 +55,7 @@ typedef struct pp_info {
  * used by all instances
  */
 typedef struct pw_conn {
-	int restrict;		/* TRUE if connection is restricted */
+	int restricted;		/* TRUE if connection is restricted */
 } pw_conn;
 
 static pw_conn *pwcons;
@@ -74,7 +71,7 @@ typedef struct pass_policy {
 	int pwdMinLength; /* minimum number of chars in password */
 	int pwdExpireWarning; /* number of seconds that warning controls are
 							sent before a password expires */
-	int pwdGraceLoginLimit; /* number of times you can log in with an
+	int pwdGraceAuthNLimit; /* number of times you can log in with an
 							expired password */
 	int pwdLockout; /* 0 = do not lockout passwords, 1 = lock them out */
 	int pwdLockoutDuration; /* time in seconds a password is locked out for */
@@ -101,8 +98,8 @@ typedef struct pw_hist {
 
 /* Operational attributes */
 static AttributeDescription *ad_pwdChangedTime, *ad_pwdAccountLockedTime,
-	*ad_pwdExpirationWarned, *ad_pwdFailureTime, *ad_pwdHistory,
-	*ad_pwdGraceUseTime, *ad_pwdReset, *ad_pwdPolicySubentry;
+	*ad_pwdFailureTime, *ad_pwdHistory, *ad_pwdGraceUseTime, *ad_pwdReset,
+	*ad_pwdPolicySubentry;
 
 static struct schema_info {
 	char *def;
@@ -114,7 +111,7 @@ static struct schema_info {
 		"EQUALITY generalizedTimeMatch "
 		"ORDERING generalizedTimeOrderingMatch "
 		"SYNTAX 1.3.6.1.4.1.1466.115.121.1.24 "
-		"SINGLE-VALUE USAGE directoryOperation )",
+		"SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
 		&ad_pwdChangedTime },
 	{	"( 1.3.6.1.4.1.42.2.27.8.1.17 "
 		"NAME ( 'pwdAccountLockedTime' ) "
@@ -122,37 +119,34 @@ static struct schema_info {
 		"EQUALITY generalizedTimeMatch "
 		"ORDERING generalizedTimeOrderingMatch "
 		"SYNTAX 1.3.6.1.4.1.1466.115.121.1.24 "
-		"SINGLE-VALUE USAGE directoryOperation )",
+		"SINGLE-VALUE "
+#if 0
+		/* Not until MANAGEDIT control is released */
+		"NO-USER-MODIFICATION "
+#endif
+		"USAGE directoryOperation )",
 		&ad_pwdAccountLockedTime },
-	{	"( 1.3.6.1.4.1.42.2.27.8.1.18 "
-		"NAME ( 'pwdExpirationWarned' ) "
-		"DESC 'The time the user was first warned about the coming expiration of the password' "
-		"EQUALITY generalizedTimeMatch "
-		"ORDERING generalizedTimeOrderingMatch "
-		"SYNTAX 1.3.6.1.4.1.1466.115.121.1.24 "
-		"SINGLE-VALUE USAGE directoryOperation )",
-		&ad_pwdExpirationWarned },
 	{	"( 1.3.6.1.4.1.42.2.27.8.1.19 "
 		"NAME ( 'pwdFailureTime' ) "
 		"DESC 'The timestamps of the last consecutive authentication failures' "
 		"EQUALITY generalizedTimeMatch "
 		"ORDERING generalizedTimeOrderingMatch "
 		"SYNTAX 1.3.6.1.4.1.1466.115.121.1.24 "
-		"USAGE directoryOperation )",
+		"NO-USER-MODIFICATION USAGE directoryOperation )",
 		&ad_pwdFailureTime },
 	{	"( 1.3.6.1.4.1.42.2.27.8.1.20 "
 		"NAME ( 'pwdHistory' ) "
 		"DESC 'The history of users passwords' "
 		"EQUALITY octetStringMatch "
 		"SYNTAX 1.3.6.1.4.1.1466.115.121.1.40 "
-		"USAGE directoryOperation )",
+		"NO-USER-MODIFICATION USAGE directoryOperation )",
 		&ad_pwdHistory },
 	{	"( 1.3.6.1.4.1.42.2.27.8.1.21 "
 		"NAME ( 'pwdGraceUseTime' ) "
 		"DESC 'The timestamps of the grace login once the password has expired' "
 		"EQUALITY generalizedTimeMatch "
 		"SYNTAX 1.3.6.1.4.1.1466.115.121.1.24 "
-		"USAGE directoryOperation )",
+		"NO-USER-MODIFICATION USAGE directoryOperation )",
 		&ad_pwdGraceUseTime }, 
 	{	"( 1.3.6.1.4.1.42.2.27.8.1.22 "
 		"NAME ( 'pwdReset' ) "
@@ -166,7 +160,12 @@ static struct schema_info {
 		"DESC 'The pwdPolicy subentry in effect for this object' "
 		"EQUALITY distinguishedNameMatch "
 		"SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 "
-		"SINGLE-VALUE USAGE directoryOperation )",
+		"SINGLE-VALUE "
+#if 0
+		/* Not until MANAGEDIT control is released */
+		"NO-USER-MODIFICATION "
+#endif
+		"USAGE directoryOperation )",
 		&ad_pwdPolicySubentry },
 	{ NULL, NULL }
 };
@@ -174,7 +173,7 @@ static struct schema_info {
 /* User attributes */
 static AttributeDescription *ad_pwdMinAge, *ad_pwdMaxAge, *ad_pwdInHistory,
 	*ad_pwdCheckQuality, *ad_pwdMinLength, *ad_pwdMaxFailure, 
-	*ad_pwdGraceLoginLimit, *ad_pwdExpireWarning, *ad_pwdLockoutDuration,
+	*ad_pwdGraceAuthNLimit, *ad_pwdExpireWarning, *ad_pwdLockoutDuration,
 	*ad_pwdFailureCountInterval, *ad_pwdCheckModule, *ad_pwdLockout,
 	*ad_pwdMustChange, *ad_pwdAllowUserChange, *ad_pwdSafeModify,
 	*ad_pwdAttribute;
@@ -189,7 +188,7 @@ static struct schema_info pwd_UsSchema[] = {
 	TAB(pwdCheckQuality),
 	TAB(pwdMinLength),
 	TAB(pwdMaxFailure),
-	TAB(pwdGraceLoginLimit),
+	TAB(pwdGraceAuthNLimit),
 	TAB(pwdExpireWarning),
 	TAB(pwdLockout),
 	TAB(pwdLockoutDuration),
@@ -222,10 +221,8 @@ account_locked( Operation *op, Entry *e,
 		PassPolicy *pp, Modifications **mod ) 
 {
 	Attribute       *la;
-	int rc;
-	Entry *de;
 
-	assert(mod);
+	assert(mod != NULL);
 
 	if ( (la = attr_find( e->e_attrs, ad_pwdAccountLockedTime )) != NULL ) {
 		BerVarray vals = la->a_nvals;
@@ -236,9 +233,11 @@ account_locked( Operation *op, Entry *e,
 		 */
 		if (vals[0].bv_val != NULL) {
 			time_t then, now;
-			struct berval bv;
 			Modifications *m;
 
+			if (!pp->pwdLockoutDuration)
+				return 1;
+
 			if ((then = parse_time( vals[0].bv_val )) == (time_t)0)
 				return 1;
 
@@ -249,6 +248,7 @@ account_locked( Operation *op, Entry *e,
 
 			m = ch_calloc( sizeof(Modifications), 1 );
 			m->sml_op = LDAP_MOD_DELETE;
+			m->sml_flags = 0;
 			m->sml_type = ad_pwdAccountLockedTime->ad_cname;
 			m->sml_desc = ad_pwdAccountLockedTime;
 			m->sml_next = *mod;
@@ -317,12 +317,11 @@ ppolicy_get( Operation *op, Entry *e, PassPolicy *pp )
 	pp_info *pi = on->on_bi.bi_private;
 	Attribute *a;
 	BerVarray vals;
-	int i, rc, nent;
+	int rc;
 	Entry *pe = NULL;
-	AttributeDescription *oca = slap_schema.si_ad_objectClass;
+#if 0
 	const char *text;
-	AttributeDescription *ad;
-	struct berval bv;
+#endif
 
 	memset( pp, 0, sizeof(PassPolicy) );
 
@@ -358,30 +357,40 @@ ppolicy_get( Operation *op, Entry *e, PassPolicy *pp )
 	pp->ad = slap_schema.si_ad_userPassword;
 #endif
 
-	if ((a = attr_find( pe->e_attrs, ad_pwdMinAge )))
-		pp->pwdMinAge = atoi(a->a_vals[0].bv_val );
-	if ((a = attr_find( pe->e_attrs, ad_pwdMaxAge )))
-		pp->pwdMaxAge = atoi(a->a_vals[0].bv_val );
-	if ((a = attr_find( pe->e_attrs, ad_pwdInHistory )))
-		pp->pwdInHistory = atoi(a->a_vals[0].bv_val );
-	if ((a = attr_find( pe->e_attrs, ad_pwdCheckQuality )))
-		pp->pwdCheckQuality = atoi(a->a_vals[0].bv_val );
-	if ((a = attr_find( pe->e_attrs, ad_pwdMinLength )))
-		pp->pwdMinLength = atoi(a->a_vals[0].bv_val );
-	if ((a = attr_find( pe->e_attrs, ad_pwdMaxFailure )))
-		pp->pwdMaxFailure = atoi(a->a_vals[0].bv_val );
-	if ((a = attr_find( pe->e_attrs, ad_pwdGraceLoginLimit )))
-		pp->pwdGraceLoginLimit = atoi(a->a_vals[0].bv_val );
-	if ((a = attr_find( pe->e_attrs, ad_pwdExpireWarning )))
-		pp->pwdExpireWarning = atoi(a->a_vals[0].bv_val );
-	if ((a = attr_find( pe->e_attrs, ad_pwdFailureCountInterval )))
-		pp->pwdFailureCountInterval = atoi(a->a_vals[0].bv_val );
-	if ((a = attr_find( pe->e_attrs, ad_pwdLockoutDuration )))
-		pp->pwdLockoutDuration = atoi(a->a_vals[0].bv_val );
+	if ( ( a = attr_find( pe->e_attrs, ad_pwdMinAge ) )
+			&& lutil_atoi( &pp->pwdMinAge, a->a_vals[0].bv_val ) != 0 )
+		goto defaultpol;
+	if ( ( a = attr_find( pe->e_attrs, ad_pwdMaxAge ) )
+			&& lutil_atoi( &pp->pwdMaxAge, a->a_vals[0].bv_val ) != 0 )
+		goto defaultpol;
+	if ( ( a = attr_find( pe->e_attrs, ad_pwdInHistory ) )
+			&& lutil_atoi( &pp->pwdInHistory, a->a_vals[0].bv_val ) != 0 )
+		goto defaultpol;
+	if ( ( a = attr_find( pe->e_attrs, ad_pwdCheckQuality ) )
+			&& lutil_atoi( &pp->pwdCheckQuality, a->a_vals[0].bv_val ) != 0 )
+		goto defaultpol;
+	if ( ( a = attr_find( pe->e_attrs, ad_pwdMinLength ) )
+			&& lutil_atoi( &pp->pwdMinLength, a->a_vals[0].bv_val ) != 0 )
+		goto defaultpol;
+	if ( ( a = attr_find( pe->e_attrs, ad_pwdMaxFailure ) )
+			&& lutil_atoi( &pp->pwdMaxFailure, a->a_vals[0].bv_val ) != 0 )
+		goto defaultpol;
+	if ( ( a = attr_find( pe->e_attrs, ad_pwdGraceAuthNLimit ) )
+			&& lutil_atoi( &pp->pwdGraceAuthNLimit, a->a_vals[0].bv_val ) != 0 )
+		goto defaultpol;
+	if ( ( a = attr_find( pe->e_attrs, ad_pwdExpireWarning ) )
+			&& lutil_atoi( &pp->pwdExpireWarning, a->a_vals[0].bv_val ) != 0 )
+		goto defaultpol;
+	if ( ( a = attr_find( pe->e_attrs, ad_pwdFailureCountInterval ) )
+			&& lutil_atoi( &pp->pwdFailureCountInterval, a->a_vals[0].bv_val ) != 0 )
+		goto defaultpol;
+	if ( ( a = attr_find( pe->e_attrs, ad_pwdLockoutDuration ) )
+			&& lutil_atoi( &pp->pwdLockoutDuration, a->a_vals[0].bv_val ) != 0 )
+		goto defaultpol;
 
-	if ((a = attr_find( pe->e_attrs, ad_pwdCheckModule ))) {
-		strncpy(pp->pwdCheckModule, a->a_vals[0].bv_val,
-			sizeof(pp->pwdCheckModule));
+	if ( ( a = attr_find( pe->e_attrs, ad_pwdCheckModule ) ) ) {
+		strncpy( pp->pwdCheckModule, a->a_vals[0].bv_val,
+			sizeof(pp->pwdCheckModule) );
 		pp->pwdCheckModule[sizeof(pp->pwdCheckModule)-1] = '\0';
 	}
 
@@ -423,12 +432,9 @@ password_scheme( struct berval *cred, struct berval *sch )
 
 	for(e = 1; cred->bv_val[e] && cred->bv_val[e] != '}'; e++);
 	if (cred->bv_val[e]) {
-		char *sc = ch_calloc( sizeof(char), e + 2);
-		sc[e + 1] = '\0'; /* terminate string */
-		strncpy( sc, cred->bv_val, e + 1);
-		e = lutil_passwd_scheme( sc );
-		free( sc );
-		if (e && sch) {
+		int rc;
+		rc = lutil_passwd_scheme( cred->bv_val );
+		if (rc && sch) {
 			sch->bv_val = cred->bv_val;
 			sch->bv_len = e;
 			return LDAP_SUCCESS;
@@ -442,7 +448,6 @@ check_password_quality( struct berval *cred, PassPolicy *pp, LDAPPasswordPolicyE
 {
 	int rc = LDAP_SUCCESS, ok = LDAP_SUCCESS;
 	char *ptr = cred->bv_val;
-	char *modpath;
 	struct berval sch;
 
 	assert( cred != NULL );
@@ -635,9 +640,9 @@ make_pwd_history_value( char *timebuf, struct berval *bv, Attribute *pa )
 	int nlen;
 
 	snprintf( str, MAX_PWD_HISTORY_SZ,
-		  "%s#%s#%d#", timebuf,
+		  "%s#%s#%lu#", timebuf,
 		  pa->a_desc->ad_type->sat_syntax->ssyn_oid,
-		  pa->a_nvals[0].bv_len );
+		  (unsigned long) pa->a_nvals[0].bv_len );
 	str[MAX_PWD_HISTORY_SZ-1] = 0;
 	nlen = strlen(str);
 
@@ -691,12 +696,11 @@ ppolicy_bind_resp( Operation *op, SlapReply *rs )
 	slap_overinst *on = ppb->on;
 	Modifications *mod = ppb->mod, *m;
 	int pwExpired = 0;
-	int ngut = -1, warn = -1, age, rc, i;
+	int ngut = -1, warn = -1, age, rc;
 	Attribute *a;
-	struct tm *tm;
-	time_t now, then, pwtime = (time_t)-1;
-	const char *txt;
+	time_t now, pwtime = (time_t)-1;
 	char nowstr[ LDAP_LUTIL_GENTIME_BUFSIZE ];
+	struct berval timestamp;
 	BackendInfo *bi = op->o_bd->bd_info;
 	Entry *e;
 
@@ -714,21 +718,23 @@ ppolicy_bind_resp( Operation *op, SlapReply *rs )
 	}
 
 	now = slap_get_time(); /* stored for later consideration */
-	ldap_pvt_thread_mutex_lock( &gmtime_mutex );
-	tm = gmtime(&now);
-	lutil_gentime( nowstr, sizeof(nowstr), tm );
-	ldap_pvt_thread_mutex_unlock( &gmtime_mutex );
+	timestamp.bv_val = nowstr;
+	timestamp.bv_len = sizeof(nowstr);
+	slap_timestamp( &now, &timestamp );
 
 	if ( rs->sr_err == LDAP_INVALID_CREDENTIALS ) {
 		int i = 0, fc = 0;
 
 		m = ch_calloc( sizeof(Modifications), 1 );
 		m->sml_op = LDAP_MOD_ADD;
+		m->sml_flags = 0;
 		m->sml_type = ad_pwdFailureTime->ad_cname;
 		m->sml_desc = ad_pwdFailureTime;
 		m->sml_values = ch_calloc( sizeof(struct berval), 2 );
+		m->sml_nvalues = ch_calloc( sizeof(struct berval), 2 );
 
-		ber_str2bv( nowstr, 0, 1, &m->sml_values[0] );
+		ber_dupbv( &m->sml_values[0], &timestamp );
+		ber_dupbv( &m->sml_nvalues[0], &timestamp );
 		m->sml_next = mod;
 		mod = m;
 
@@ -770,10 +776,13 @@ ppolicy_bind_resp( Operation *op, SlapReply *rs )
 			 */
 			m = ch_calloc( sizeof(Modifications), 1 );
 			m->sml_op = LDAP_MOD_REPLACE;
+			m->sml_flags = 0;
 			m->sml_type = ad_pwdAccountLockedTime->ad_cname;
 			m->sml_desc = ad_pwdAccountLockedTime;
 			m->sml_values = ch_calloc( sizeof(struct berval), 2 );
-			ber_str2bv( nowstr, 0, 1, &m->sml_values[0] );
+			m->sml_nvalues = ch_calloc( sizeof(struct berval), 2 );
+			ber_dupbv( &m->sml_values[0], &timestamp );
+			ber_dupbv( &m->sml_nvalues[0], &timestamp );
 			m->sml_next = mod;
 			mod = m;
 		}
@@ -785,6 +794,7 @@ ppolicy_bind_resp( Operation *op, SlapReply *rs )
 		if ( attr_find( e->e_attrs, ad_pwdFailureTime )) {
 			m = ch_calloc( sizeof(Modifications), 1 );
 			m->sml_op = LDAP_MOD_DELETE;
+			m->sml_flags = 0;
 			m->sml_type = ad_pwdFailureTime->ad_cname;
 			m->sml_desc = ad_pwdFailureTime;
 			m->sml_next = mod;
@@ -803,7 +813,7 @@ ppolicy_bind_resp( Operation *op, SlapReply *rs )
 			 * that we are disallowed from doing anything
 			 * other than change password.
 			 */
-			pwcons[op->o_conn->c_conn_idx].restrict = 1;
+			pwcons[op->o_conn->c_conn_idx].restricted = 1;
 
 			ppb->pErr = PP_changeAfterReset;
 
@@ -846,10 +856,10 @@ grace:
 		if (!pwExpired) goto check_expiring_password;
 		
 		if ((a = attr_find( e->e_attrs, ad_pwdGraceUseTime )) == NULL)
-			ngut = ppb->pp.pwdGraceLoginLimit;
+			ngut = ppb->pp.pwdGraceAuthNLimit;
 		else {
 			for(ngut=0; a->a_nvals[ngut].bv_val; ngut++);
-			ngut = ppb->pp.pwdGraceLoginLimit - ngut;
+			ngut = ppb->pp.pwdGraceAuthNLimit - ngut;
 		}
 
 		/*
@@ -870,10 +880,13 @@ grace:
 		 */
 		m = ch_calloc( sizeof(Modifications), 1 );
 		m->sml_op = LDAP_MOD_ADD;
+		m->sml_flags = 0;
 		m->sml_type = ad_pwdGraceUseTime->ad_cname;
 		m->sml_desc = ad_pwdGraceUseTime;
 		m->sml_values = ch_calloc( sizeof(struct berval), 2 );
-		ber_str2bv( nowstr, 0, 1, &m->sml_values[0] );
+		m->sml_nvalues = ch_calloc( sizeof(struct berval), 2 );
+		ber_dupbv( &m->sml_values[0], &timestamp );
+		ber_dupbv( &m->sml_nvalues[0], &timestamp );
 		m->sml_next = mod;
 		mod = m;
 
@@ -901,19 +914,8 @@ check_expiring_password:
 		 */
 		if (ppb->pp.pwdMaxAge - age < ppb->pp.pwdExpireWarning ) {
 			/*
-			 * Set the warning value, add expiration warned timestamp to the entry.
+			 * Set the warning value.
 			 */
-			if ((a = attr_find( e->e_attrs, ad_pwdExpirationWarned )) == NULL) {
-				m = ch_calloc( sizeof(Modifications), 1 );
-				m->sml_op = LDAP_MOD_ADD;
-				m->sml_type = ad_pwdExpirationWarned->ad_cname;
-				m->sml_desc = ad_pwdExpirationWarned;
-				m->sml_values = ch_calloc( sizeof(struct berval), 2 );
-				ber_str2bv( nowstr, 0, 1, &m->sml_values[0] );
-				m->sml_next = mod;
-				mod = m;
-			}
-			
 			warn = ppb->pp.pwdMaxAge - age; /* seconds left until expiry */
 			if (warn < 0) warn = 0; /* something weird here - why is pwExpired not set? */
 			
@@ -943,7 +945,7 @@ locked:
 		op2.o_ndn = op->o_bd->be_rootndn;
 		op2.o_bd->bd_info = (BackendInfo *)on->on_info;
 		rc = op->o_bd->be_modify( &op2, &r2 );
-		slap_mods_free( mod );
+		slap_mods_free( mod, 1 );
 	}
 
 	if ( ppb->send_ctrl ) {
@@ -968,10 +970,13 @@ ppolicy_bind( Operation *op, SlapReply *rs )
 {
 	slap_overinst *on = (slap_overinst *)op->o_bd->bd_info;
 
+	/* Reset lockout status on all Bind requests */
+	pwcons[op->o_conn->c_conn_idx].restricted = 0;
+
 	/* Root bypasses policy */
 	if ( !be_isroot_dn( op->o_bd, &op->o_req_ndn )) {
 		Entry *e;
-		int i, rc;
+		int rc;
 		ppbind *ppb;
 		slap_callback *cb;
 
@@ -1020,11 +1025,11 @@ ppolicy_bind( Operation *op, SlapReply *rs )
 	return SLAP_CB_CONTINUE;
 }
 
-/* Reset the restrict flag for the next session on this connection */
+/* Reset the restricted flag for the next session on this connection */
 static int
-ppolicy_unbind( Operation *op, SlapReply *rs )
+ppolicy_connection_destroy( BackendDB *bd, Connection *conn )
 {
-	pwcons[op->o_conn->c_conn_idx].restrict = 0;
+	pwcons[conn->c_conn_idx].restricted = 0;
 	return SLAP_CB_CONTINUE;
 }
 
@@ -1035,14 +1040,14 @@ ppolicy_restrict(
 	SlapReply *rs )
 {
 	slap_overinst *on = (slap_overinst *)op->o_bd->bd_info;
-	int i, send_ctrl = 0;
+	int send_ctrl = 0;
 
 	/* Did we receive a password policy request control? */
 	if ( op->o_ctrlflag[ppolicy_cid] ) {
 		send_ctrl = 1;
 	}
 
-	if ( op->o_conn && pwcons[op->o_conn->c_conn_idx].restrict ) {
+	if ( op->o_conn && pwcons[op->o_conn->c_conn_idx].restricted ) {
 		Debug( LDAP_DEBUG_TRACE,
 			"connection restricted to password changing only\n", 0, 0, 0);
 		if ( send_ctrl ) {
@@ -1054,7 +1059,7 @@ ppolicy_restrict(
 			rs->sr_ctrls = ctrls;
 		}
 		op->o_bd->bd_info = (BackendInfo *)on->on_info;
-		send_ldap_error( op, rs, LDAP_UNWILLING_TO_PERFORM,
+		send_ldap_error( op, rs, LDAP_INSUFFICIENT_ACCESS, 
 			"Operations are restricted to bind/unbind/abandon/StartTLS/modify password" );
 		return rs->sr_err;
 	}
@@ -1070,13 +1075,16 @@ ppolicy_add(
 	slap_overinst *on = (slap_overinst *)op->o_bd->bd_info;
 	pp_info *pi = on->on_bi.bi_private;
 	PassPolicy pp;
-	int pw;
 	Attribute *pa;
 	const char *txt;
 
 	if ( ppolicy_restrict( op, rs ) != SLAP_CB_CONTINUE )
 		return rs->sr_err;
 
+	/* If this is a replica, assume the master checked everything */
+	if ( be_shadow_update( op ))
+		return SLAP_CB_CONTINUE;
+
 	/* Check for password in entry */
 	if ((pa = attr_find( op->oq_add.rs_e->e_attrs,
 		slap_schema.si_ad_userPassword )))
@@ -1089,7 +1097,7 @@ ppolicy_add(
 		ppolicy_get( op, op->ora_e, &pp );
 		if (pp.pwdCheckQuality > 0 && !be_isroot( op )) {
 			struct berval *bv = &(pa->a_vals[0]);
-			int rc, i, send_ctrl = 0; 
+			int rc, send_ctrl = 0;
 			LDAPPasswordPolicyError pErr = PP_noError;
 
 			/* Did we receive a password policy request control? */
@@ -1110,54 +1118,50 @@ ppolicy_add(
 				send_ldap_error( op, rs, rc, "Password fails quality checking policy" );
 				return rs->sr_err;
 			}
-			    /*
-			     * A controversial bit. We hash cleartext
-			     * passwords provided via add and modify operations
-			     * You're not really supposed to do this, since
-			     * the X.500 model says "store attributes" as they
-			     * get provided. By default, this is what we do
-			     *
-			     * But if the hash_passwords flag is set, we hash
-			     * any cleartext password attribute values via the
-			     * default password hashing scheme.
-			     */
-			if ((pi->hash_passwords) &&
-				(password_scheme( &(pa->a_vals[0]), NULL ) != LDAP_SUCCESS)) {
-				struct berval hpw;
-
-				slap_passwd_hash( &(pa->a_vals[0]), &hpw, &txt );
-				if (hpw.bv_val == NULL) {
-				    /*
-				     * hashing didn't work. Emit an error.
-				     */
-					rs->sr_err = LDAP_OTHER;
-					rs->sr_text = txt;
-					send_ldap_error( op, rs, LDAP_OTHER, "Password hashing failed" );
-					return rs->sr_err;
-				}
-
-				memset( pa->a_vals[0].bv_val, 0, pa->a_vals[0].bv_len);
-				ber_memfree( pa->a_vals[0].bv_val );
-				pa->a_vals[0].bv_val = hpw.bv_val;
-				pa->a_vals[0].bv_len = hpw.bv_len;
-			}
 		}
+			/*
+			 * A controversial bit. We hash cleartext
+			 * passwords provided via add and modify operations
+			 * You're not really supposed to do this, since
+			 * the X.500 model says "store attributes" as they
+			 * get provided. By default, this is what we do
+			 *
+			 * But if the hash_passwords flag is set, we hash
+			 * any cleartext password attribute values via the
+			 * default password hashing scheme.
+			 */
+		if ((pi->hash_passwords) &&
+			(password_scheme( &(pa->a_vals[0]), NULL ) != LDAP_SUCCESS)) {
+			struct berval hpw;
+
+			slap_passwd_hash( &(pa->a_vals[0]), &hpw, &txt );
+			if (hpw.bv_val == NULL) {
+				/*
+				 * hashing didn't work. Emit an error.
+				 */
+				rs->sr_err = LDAP_OTHER;
+				rs->sr_text = txt;
+				send_ldap_error( op, rs, LDAP_OTHER, "Password hashing failed" );
+				return rs->sr_err;
+			}
+
+			memset( pa->a_vals[0].bv_val, 0, pa->a_vals[0].bv_len);
+			ber_memfree( pa->a_vals[0].bv_val );
+			pa->a_vals[0].bv_val = hpw.bv_val;
+			pa->a_vals[0].bv_len = hpw.bv_len;
+		}
+
 		/* If password aging is in effect, set the pwdChangedTime */
-		if (( pp.pwdMaxAge || pp.pwdMinAge ) && !be_shadow_update( op )) {
+		if ( pp.pwdMaxAge || pp.pwdMinAge ) {
 			struct berval timestamp;
 			char timebuf[ LDAP_LUTIL_GENTIME_BUFSIZE ];
-			struct tm *ltm;
 			time_t now = slap_get_time();
 
-			ldap_pvt_thread_mutex_lock( &gmtime_mutex );
-			ltm = gmtime( &now );
-			lutil_gentime( timebuf, sizeof(timebuf), ltm );
-			ldap_pvt_thread_mutex_unlock( &gmtime_mutex );
-
 			timestamp.bv_val = timebuf;
-			timestamp.bv_len = strlen(timebuf);
+			timestamp.bv_len = sizeof(timebuf);
+			slap_timestamp( &now, &timestamp );
 
-			attr_merge_one( op->ora_e, ad_pwdChangedTime, &timestamp, NULL );
+			attr_merge_one( op->ora_e, ad_pwdChangedTime, &timestamp, &timestamp );
 		}
 	}
 	return SLAP_CB_CONTINUE;
@@ -1168,11 +1172,12 @@ ppolicy_modify( Operation *op, SlapReply *rs )
 {
 	slap_overinst		*on = (slap_overinst *)op->o_bd->bd_info;
 	pp_info			*pi = on->on_bi.bi_private;
-	int			i, rc, mod_pw_only, pwmod, pwmop, deladd,
+	int			i, rc, mod_pw_only, pwmod, pwmop = -1, deladd,
 				hsize = 0;
 	PassPolicy		pp;
-	Modifications		*mods = NULL, *modtail, *ml, *delmod, *addmod;
-	Attribute		*pa, *ha, *ra, at;
+	Modifications		*mods = NULL, *modtail = NULL,
+				*ml, *delmod, *addmod;
+	Attribute		*pa, *ha, at;
 	const char		*txt;
 	pw_hist			*tl = NULL, *p;
 	int			zapReset, send_ctrl = 0;
@@ -1187,6 +1192,98 @@ ppolicy_modify( Operation *op, SlapReply *rs )
 
 	if ( rc != LDAP_SUCCESS ) return SLAP_CB_CONTINUE;
 
+	/* If this is a replica, we may need to tweak some of the
+	 * master's modifications. Otherwise, just pass it through.
+	 */
+	if ( be_shadow_update( op )) {
+		Modifications **prev;
+		int got_del_grace = 0, got_del_lock = 0, got_pw = 0, got_del_fail = 0;
+		Attribute *a_grace, *a_lock, *a_fail;
+
+		a_grace = attr_find( e->e_attrs, ad_pwdGraceUseTime );
+		a_lock = attr_find( e->e_attrs, ad_pwdAccountLockedTime );
+		a_fail = attr_find( e->e_attrs, ad_pwdFailureTime );
+
+		for( prev = &op->oq_modify.rs_modlist, ml = *prev; ml; ml = *prev ) {
+
+			if ( ml->sml_desc == slap_schema.si_ad_userPassword )
+				got_pw = 1;
+
+			/* If we're deleting an attr that didn't exist,
+			 * drop this delete op
+			 */
+			if ( ml->sml_op == LDAP_MOD_DELETE ) {
+				int drop = 0;
+
+				if ( ml->sml_desc == ad_pwdGraceUseTime ) {
+					got_del_grace = 1;
+					if ( !a_grace )
+						drop = 1;
+				} else
+				if ( ml->sml_desc == ad_pwdAccountLockedTime ) {
+					got_del_lock = 1;
+					if ( !a_lock )
+						drop = 1;
+				} else
+				if ( ml->sml_desc == ad_pwdFailureTime ) {
+					got_del_fail = 1;
+					if ( !a_fail )
+						drop = 1;
+				}
+				if ( drop ) {
+					*prev = ml->sml_next;
+					ml->sml_next = NULL;
+					slap_mods_free( ml, 1 );
+					continue;
+				}
+			}
+			prev = &ml->sml_next;
+		}
+
+		/* If we're resetting the password, make sure grace, accountlock,
+		 * and failure also get removed.
+		 */
+		if ( got_pw ) {
+			if ( a_grace && !got_del_grace ) {
+				ml = (Modifications *) ch_malloc( sizeof( Modifications ) );
+				ml->sml_op = LDAP_MOD_DELETE;
+				ml->sml_flags = SLAP_MOD_INTERNAL;
+				ml->sml_type.bv_val = NULL;
+				ml->sml_desc = ad_pwdGraceUseTime;
+				ml->sml_values = NULL;
+				ml->sml_nvalues = NULL;
+				ml->sml_next = NULL;
+				*prev = ml;
+				prev = &ml->sml_next;
+			}
+			if ( a_lock && !got_del_lock ) {
+				ml = (Modifications *) ch_malloc( sizeof( Modifications ) );
+				ml->sml_op = LDAP_MOD_DELETE;
+				ml->sml_flags = SLAP_MOD_INTERNAL;
+				ml->sml_type.bv_val = NULL;
+				ml->sml_desc = ad_pwdAccountLockedTime;
+				ml->sml_values = NULL;
+				ml->sml_nvalues = NULL;
+				ml->sml_next = NULL;
+				*prev = ml;
+			}
+			if ( a_fail && !got_del_fail ) {
+				ml = (Modifications *) ch_malloc( sizeof( Modifications ) );
+				ml->sml_op = LDAP_MOD_DELETE;
+				ml->sml_flags = SLAP_MOD_INTERNAL;
+				ml->sml_type.bv_val = NULL;
+				ml->sml_desc = ad_pwdFailureTime;
+				ml->sml_values = NULL;
+				ml->sml_nvalues = NULL;
+				ml->sml_next = NULL;
+				*prev = ml;
+			}
+		}
+		op->o_bd->bd_info = (BackendInfo *)on->on_info;
+		be_entry_release_r( op, e );
+		return SLAP_CB_CONTINUE;
+	}
+
 	/* Did we receive a password policy request control? */
 	if ( op->o_ctrlflag[ppolicy_cid] ) {
 		send_ctrl = 1;
@@ -1211,12 +1308,13 @@ ppolicy_modify( Operation *op, SlapReply *rs )
 
 	ppolicy_get( op, e, &pp );
 
-	for(ml = op->oq_modify.rs_modlist,
+	for ( ml = op->oq_modify.rs_modlist,
 			pwmod = 0, mod_pw_only = 1,
 			deladd = 0, delmod = NULL,
 			addmod = NULL,
 			zapReset = 1;
-		ml != NULL; modtail = ml, ml = ml->sml_next ) {
+		ml != NULL; modtail = ml, ml = ml->sml_next )
+	{
 		if ( ml->sml_desc == pp.ad ) {
 			pwmod = 1;
 			pwmop = ml->sml_op;
@@ -1255,10 +1353,10 @@ ppolicy_modify( Operation *op, SlapReply *rs )
 		}
 	}
 	
-	if (pwcons[op->o_conn->c_conn_idx].restrict && !mod_pw_only) {
+	if (pwcons[op->o_conn->c_conn_idx].restricted && !mod_pw_only) {
 		Debug( LDAP_DEBUG_TRACE,
 			"connection restricted to password changing only\n", 0, 0, 0 );
-		rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
+		rs->sr_err = LDAP_INSUFFICIENT_ACCESS; 
 		rs->sr_text = "Operations are restricted to bind/unbind/abandon/StartTLS/modify password";
 		pErr = PP_changeAfterReset;
 		goto return_results;
@@ -1294,7 +1392,6 @@ ppolicy_modify( Operation *op, SlapReply *rs )
 	if (pp.pwdInHistory > 0 && (ha = attr_find( e->e_attrs, ad_pwdHistory ))) {
 		struct berval oldpw;
 		time_t oldtime;
-		char *oid;
 
 		for(i=0; ha->a_nvals[i].bv_val; i++) {
 			rc = parse_pwdhistory( &(ha->a_nvals[i]), NULL,
@@ -1321,6 +1418,7 @@ ppolicy_modify( Operation *op, SlapReply *rs )
 	if (pp.pwdSafeModify && oldpw.bv_val ) {
 		ml = (Modifications *) ch_malloc( sizeof( Modifications ) );
 		ml->sml_op = LDAP_MOD_DELETE;
+		ml->sml_flags = SLAP_MOD_INTERNAL;
 		ml->sml_desc = pp.ad;
 		ml->sml_type = pp.ad->ad_cname;
 		ml->sml_values = (BerVarray) ch_malloc( 2 * sizeof( struct berval ) );
@@ -1338,14 +1436,14 @@ ppolicy_modify( Operation *op, SlapReply *rs )
 		Debug( LDAP_DEBUG_TRACE,
 			"change password must use DELETE followed by ADD/REPLACE\n",
 			0, 0, 0 );
-		rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
+		rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
 		rs->sr_text = "Must supply old password to be changed as well as new one";
 		pErr = PP_mustSupplyOldPassword;
 		goto return_results;
 	}
 
 	if (!pp.pwdAllowUserChange) {
-		rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
+		rs->sr_err = LDAP_INSUFFICIENT_ACCESS;
 		rs->sr_text = "User alteration of password is not allowed";
 		pErr = PP_passwordModNotAllowed;
 		goto return_results;
@@ -1360,7 +1458,7 @@ ppolicy_modify( Operation *op, SlapReply *rs )
 		now = slap_get_time();
 		age = (int)(now - pwtime);
 		if ((pwtime != (time_t)-1) && (age < pp.pwdMinAge)) {
-			rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
+			rs->sr_err = LDAP_CONSTRAINT_VIOLATION;
 			rs->sr_text = "Password is too young to change";
 			pErr = PP_passwordTooYoung;
 			goto return_results;
@@ -1456,39 +1554,37 @@ ppolicy_modify( Operation *op, SlapReply *rs )
 	}
 
 do_modify:
-	if ((pwmod) && (!be_shadow_update( op ))) {
+	if (pwmod) {
 		struct berval timestamp;
 		char timebuf[ LDAP_LUTIL_GENTIME_BUFSIZE ];
-		struct tm *ltm;
 		time_t now = slap_get_time();
-		Attribute *ga;
 		
 		/*
 		 * keep the necessary pwd.. operational attributes
 		 * up to date.
 		 */
 
-		ldap_pvt_thread_mutex_lock( &gmtime_mutex );
-		ltm = gmtime( &now );
-		lutil_gentime( timebuf, sizeof(timebuf), ltm );
-		ldap_pvt_thread_mutex_unlock( &gmtime_mutex );
-
 		timestamp.bv_val = timebuf;
-		timestamp.bv_len = strlen(timebuf);
+		timestamp.bv_len = sizeof(timebuf);
+		slap_timestamp( &now, &timestamp );
+
 		mods = (Modifications *) ch_malloc( sizeof( Modifications ) );
 		mods->sml_type.bv_val = NULL;
 		mods->sml_desc = ad_pwdChangedTime;
 		if (pwmop != LDAP_MOD_DELETE) {
 			mods->sml_op = LDAP_MOD_REPLACE;
 			mods->sml_values = (BerVarray) ch_malloc( 2 * sizeof( struct berval ) );
+			mods->sml_nvalues = (BerVarray) ch_malloc( 2 * sizeof( struct berval ) );
 			ber_dupbv( &mods->sml_values[0], &timestamp );
+			ber_dupbv( &mods->sml_nvalues[0], &timestamp );
 			mods->sml_values[1].bv_len = 0;
 			mods->sml_values[1].bv_val = NULL;
-			assert( mods->sml_values[0].bv_val );
+			assert( mods->sml_values[0].bv_val != NULL );
 		} else {
 			mods->sml_op = LDAP_MOD_DELETE;
 			mods->sml_values = NULL;
 		}
+		mods->sml_flags = SLAP_MOD_INTERNAL;
 		mods->sml_nvalues = NULL;
 		mods->sml_next = NULL;
 		modtail->sml_next = mods;
@@ -1497,6 +1593,7 @@ do_modify:
 		if (attr_find(e->e_attrs, ad_pwdGraceUseTime )) {
 			mods = (Modifications *) ch_malloc( sizeof( Modifications ) );
 			mods->sml_op = LDAP_MOD_DELETE;
+			mods->sml_flags = SLAP_MOD_INTERNAL;
 			mods->sml_type.bv_val = NULL;
 			mods->sml_desc = ad_pwdGraceUseTime;
 			mods->sml_values = NULL;
@@ -1506,11 +1603,25 @@ do_modify:
 			modtail = mods;
 		}
 
-		if (attr_find(e->e_attrs, ad_pwdExpirationWarned )) {
+		if (attr_find(e->e_attrs, ad_pwdAccountLockedTime )) {
 			mods = (Modifications *) ch_malloc( sizeof( Modifications ) );
 			mods->sml_op = LDAP_MOD_DELETE;
+			mods->sml_flags = SLAP_MOD_INTERNAL;
 			mods->sml_type.bv_val = NULL;
-			mods->sml_desc = ad_pwdExpirationWarned;
+			mods->sml_desc = ad_pwdAccountLockedTime;
+			mods->sml_values = NULL;
+			mods->sml_nvalues = NULL;
+			mods->sml_next = NULL;
+			modtail->sml_next = mods;
+			modtail = mods;
+		}
+
+		if (attr_find(e->e_attrs, ad_pwdFailureTime )) {
+			mods = (Modifications *) ch_malloc( sizeof( Modifications ) );
+			mods->sml_op = LDAP_MOD_DELETE;
+			mods->sml_flags = SLAP_MOD_INTERNAL;
+			mods->sml_type.bv_val = NULL;
+			mods->sml_desc = ad_pwdFailureTime;
 			mods->sml_values = NULL;
 			mods->sml_nvalues = NULL;
 			mods->sml_next = NULL;
@@ -1522,6 +1633,7 @@ do_modify:
 		if ((zapReset) && (attr_find(e->e_attrs, ad_pwdReset ))) {
 			mods = (Modifications *) ch_malloc( sizeof( Modifications ) );
 			mods->sml_op = LDAP_MOD_DELETE;
+			mods->sml_flags = SLAP_MOD_INTERNAL;
 			mods->sml_type.bv_val = NULL;
 			mods->sml_desc = ad_pwdReset;
 			mods->sml_values = NULL;
@@ -1551,6 +1663,7 @@ do_modify:
 				 */
 				mods = (Modifications *) ch_malloc( sizeof( Modifications ) );
 				mods->sml_op = LDAP_MOD_DELETE;
+				mods->sml_flags = SLAP_MOD_INTERNAL;
 				mods->sml_type.bv_val = NULL;
 				mods->sml_desc = ad_pwdHistory;
 				mods->sml_nvalues = NULL;
@@ -1583,6 +1696,7 @@ do_modify:
 			if ((pa = attr_find( e->e_attrs, pp.ad )) != NULL) {
 				mods = (Modifications *) ch_malloc( sizeof( Modifications ) );
 				mods->sml_op = LDAP_MOD_ADD;
+				mods->sml_flags = SLAP_MOD_INTERNAL;
 				mods->sml_type.bv_val = NULL;
 				mods->sml_desc = ad_pwdHistory;
 				mods->sml_nvalues = NULL;
@@ -1677,6 +1791,44 @@ ppolicy_parseCtrl(
 	return LDAP_SUCCESS;
 }
 
+static int
+attrPretty(
+	Syntax *syntax,
+	struct berval *val,
+	struct berval *out,
+	void *ctx )
+{
+	AttributeDescription *ad = NULL;
+	const char *err;
+	int code;
+
+	code = slap_bv2ad( val, &ad, &err );
+	if ( !code ) {
+		ber_dupbv_x( out, &ad->ad_type->sat_cname, ctx );
+	}
+	return code;
+}
+
+static int
+attrNormalize(
+	slap_mask_t use,
+	Syntax *syntax,
+	MatchingRule *mr,
+	struct berval *val,
+	struct berval *out,
+	void *ctx )
+{
+	AttributeDescription *ad = NULL;
+	const char *err;
+	int code;
+
+	code = slap_bv2ad( val, &ad, &err );
+	if ( !code ) {
+		ber_str2bv_x( ad->ad_type->sat_oid, 0, 1, out, ctx );
+	}
+	return code;
+}
+
 static int
 ppolicy_db_init(
 	BackendDB *be
@@ -1696,6 +1848,20 @@ ppolicy_db_init(
 				return code;
 			}
 		}
+		{
+			Syntax *syn;
+			MatchingRule *mr;
+
+			syn = ch_malloc( sizeof( Syntax ));
+			*syn = *ad_pwdAttribute->ad_type->sat_syntax;
+			syn->ssyn_pretty = attrPretty;
+			ad_pwdAttribute->ad_type->sat_syntax = syn;
+
+			mr = ch_malloc( sizeof( MatchingRule ));
+			*mr = *ad_pwdAttribute->ad_type->sat_equality;
+			mr->smr_normalize = attrNormalize;
+			ad_pwdAttribute->ad_type->sat_equality = mr;
+		}
 	}
 
 	on->on_bi.bi_private = ch_calloc( sizeof(pp_info), 1 );
@@ -1772,6 +1938,7 @@ ppolicy_config(
 			return ( 1 );
 		}
 		pi->hash_passwords = 1;
+		return 0;
 	}
 	return SLAP_CONF_UNKNOWN;
 }
@@ -1783,7 +1950,7 @@ static char *extops[] = {
 
 static slap_overinst ppolicy;
 
-int ppolicy_init()
+int ppolicy_initialize()
 {
 	LDAPAttributeType *at;
 	const char *err;
@@ -1797,7 +1964,7 @@ int ppolicy_init()
 				ldap_scherr2str(code), err );
 			return code;
 		}
-		code = at_add( at, &err );
+		code = at_add( at, 0, NULL, &err );
 		if ( !code ) {
 			slap_str2ad( at->at_names[0], pwd_OpSchema[i].ad, &err );
 		}
@@ -1807,6 +1974,11 @@ int ppolicy_init()
 				scherr2str(code), err );
 			return code;
 		}
+		/* Allow Manager to set these as needed */
+		if ( is_at_no_user_mod( (*pwd_OpSchema[i].ad)->ad_type )) {
+			(*pwd_OpSchema[i].ad)->ad_type->sat_flags |=
+				SLAP_AT_MANAGEABLE;
+		}
 	}
 
 	code = register_supported_control( LDAP_CONTROL_PASSWORDPOLICYREQUEST,
@@ -1827,18 +1999,18 @@ int ppolicy_init()
 
 	ppolicy.on_bi.bi_op_add = ppolicy_add;
 	ppolicy.on_bi.bi_op_bind = ppolicy_bind;
-	ppolicy.on_bi.bi_op_unbind = ppolicy_unbind;
 	ppolicy.on_bi.bi_op_compare = ppolicy_restrict;
 	ppolicy.on_bi.bi_op_delete = ppolicy_restrict;
 	ppolicy.on_bi.bi_op_modify = ppolicy_modify;
 	ppolicy.on_bi.bi_op_search = ppolicy_restrict;
+	ppolicy.on_bi.bi_connection_destroy = ppolicy_connection_destroy;
 
 	return overlay_register( &ppolicy );
 }
 
 #if SLAPD_OVER_PPOLICY == SLAPD_MOD_DYNAMIC
 int init_module(int argc, char *argv[]) {
-	return ppolicy_init();
+	return ppolicy_initialize();
 }
 #endif
 
