@@ -1,8 +1,8 @@
 /* refint.c - referential integrity module */
-/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/refint.c,v 1.7.2.2 2005/02/02 17:32:45 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/refint.c,v 1.19 2006/04/04 20:16:24 hyc Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2004-2005 The OpenLDAP Foundation.
+ * Copyright 2004-2006 The OpenLDAP Foundation.
  * Portions Copyright 2004 Symas Corporation.
  * All rights reserved.
  *
@@ -80,7 +80,7 @@ refint_db_init(
 {
 	slap_overinst *on = (slap_overinst *)be->bd_info;
 	refint_data *id = ch_malloc(sizeof(refint_data));
-	refint_attrs *ip;
+
 	id->message = "_init";
 	id->attrs = NULL;
 	id->newdn.bv_val = NULL;
@@ -242,9 +242,9 @@ refint_delete_cb(
 {
 	Attribute *a;
 	BerVarray b = NULL;
-	refint_data *id, *dd = op->o_callback->sc_private;
+	refint_data *dd = op->o_callback->sc_private;
 	refint_attrs *ia, *da = dd->attrs;
-	dependent_data *ip, *dp = NULL;
+	dependent_data *ip;
 	Modifications *mp, *ma;
 	int i;
 
@@ -291,6 +291,7 @@ refint_delete_cb(
 				mp->sml_values[1].bv_val = mp->sml_nvalues[1].bv_val = NULL;
 
 				mp->sml_op = LDAP_MOD_ADD;
+				mp->sml_flags = 0;
 				ber_dupbv(&mp->sml_values[0],  &dd->nothing);
 				ber_dupbv(&mp->sml_nvalues[0], &dd->nnothing);
 				mp->sml_next = ma;
@@ -305,6 +306,7 @@ refint_delete_cb(
 			mp->sml_values[1].bv_len = mp->sml_nvalues[1].bv_len = 0;
 			mp->sml_values[1].bv_val = mp->sml_nvalues[1].bv_val = NULL;
 			mp->sml_op = LDAP_MOD_DELETE;
+			mp->sml_flags = 0;
 			ber_dupbv(&mp->sml_values[0], &dd->dn);
 			ber_dupbv(&mp->sml_nvalues[0], &mp->sml_values[0]);
 			mp->sml_next = ma;
@@ -349,11 +351,11 @@ refint_modrdn_cb(
 {
 	Attribute *a;
 	BerVarray b = NULL;
-	refint_data *id, *dd = op->o_callback->sc_private;
+	refint_data *dd = op->o_callback->sc_private;
 	refint_attrs *ia, *da = dd->attrs;
-	dependent_data *ip = NULL, *dp = NULL;
+	dependent_data *ip = NULL;
 	Modifications *mp;
-	int i, j, fix;
+	int i, fix;
 
 	Debug(LDAP_DEBUG_TRACE, "refint_modrdn_cb <%s>\n",
 		rs->sr_entry ? rs->sr_entry->e_name.bv_val : "NOTHING", 0, 0);
@@ -389,6 +391,7 @@ refint_modrdn_cb(
 			}
 			mp = ch_malloc(sizeof(Modifications));
 			mp->sml_op = LDAP_MOD_ADD;
+			mp->sml_flags = 0;
 			mp->sml_desc = ia->attr;		/* XXX */
 			mp->sml_type = ia->attr->ad_cname;
 			mp->sml_values  = ch_malloc(2 * sizeof(BerValue));
@@ -401,6 +404,7 @@ refint_modrdn_cb(
 			ip->mm = mp;
 			mp = ch_malloc(sizeof(Modifications));
 			mp->sml_op = LDAP_MOD_DELETE;
+			mp->sml_flags = 0;
 			mp->sml_desc = ia->attr;		/* XXX */
 			mp->sml_type = ia->attr->ad_cname;
 			mp->sml_values  = ch_malloc(2 * sizeof(BerValue));
@@ -446,10 +450,9 @@ refint_response(
 	refint_data dd = *id;
 	refint_attrs *ip;
 	dependent_data *dp;
-	char *fstr, *key, *kp, **dnpp, **ndnpp, *cp;
-	BerValue ndn, moddn, pdn;
-	BerVarray b = NULL;
-	int rc, ac, i, j, ksize;
+	BerValue pdn;
+	int rc, ac;
+	Filter ftop, *fptr;
 
 	id->message = "_refint_response";
 
@@ -465,8 +468,8 @@ refint_response(
 
 	for(ip = id->attrs, ac = 0; ip; ip = ip->next, ac++);
 	if(!ac) {
-		rs->sr_err = LDAP_OTHER;
-		rs->sr_text = "refint_response called without any attributes";
+		Debug( LDAP_DEBUG_TRACE,
+			"refint_response called without any attributes\n", 0, 0, 0 );
 		return SLAP_CB_CONTINUE;
 	}
 
@@ -480,13 +483,15 @@ refint_response(
 
 	if(nop.o_bd) {
 		if (!nop.o_bd->be_search || !nop.o_bd->be_modify) {
-			rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
-			rs->sr_text = "backend missing search and/or modify";
+			Debug( LDAP_DEBUG_TRACE,
+				"refint_response: backend missing search and/or modify\n",
+				0, 0, 0 );
 			return SLAP_CB_CONTINUE;
 		}
 	} else {
-		rs->sr_err = LDAP_OTHER;
-		rs->sr_text = "no known backend? this shouldn't be happening!";
+		Debug( LDAP_DEBUG_TRACE,
+			"refint_response: no backend for our baseDN %s??\n",
+			id->dn.bv_val, 0, 0 );
 		return SLAP_CB_CONTINUE;
 	}
 
@@ -519,7 +524,6 @@ refint_response(
 	}
 
 	/*
-	** calculate the search key size and allocate it;
 	** build a search filter for all configured attributes;
 	** populate our Operation;
 	** pass our data (attr list, dn) to backend via sc_private;
@@ -529,18 +533,20 @@ refint_response(
 	**
 	*/
 
-	for(ksize = 16, ip = id->attrs; ip; ip = ip->next)
-		ksize += ip->attr->ad_cname.bv_len + op->o_req_dn.bv_len + 3;
-	kp = key = ch_malloc(ksize);
-	if(--ac) kp += sprintf(key, "(|");
-	for(ip = id->attrs; ip; ip = ip->next)
-		kp += sprintf(kp, "(%s=%s)",
-			ip->attr->ad_cname.bv_val, op->o_req_dn.bv_val);
-	if(ac) *kp++ = ')';
-	*kp = 0;
-
-	nop.ors_filter = str2filter_x(&nop, key);
-	ber_str2bv(key, 0, 0, &nop.ors_filterstr);
+	ftop.f_choice = LDAP_FILTER_OR;
+	ftop.f_next = NULL;
+	ftop.f_or = NULL;
+	nop.ors_filter = &ftop;
+	for(ip = id->attrs; ip; ip = ip->next) {
+		fptr = ch_malloc( sizeof(Filter) + sizeof(AttributeAssertion) );
+		fptr->f_choice = LDAP_FILTER_EQUALITY;
+		fptr->f_ava = (AttributeAssertion *)(fptr+1);
+		fptr->f_ava->aa_desc = ip->attr;
+		fptr->f_ava->aa_value = op->o_req_ndn;
+		fptr->f_next = ftop.f_or;
+		ftop.f_or = fptr;
+	}
+	filter2bv( nop.ors_filter, &nop.ors_filterstr );
 
 	/* callback gets the searched dn instead */
 	dd.dn = op->o_req_ndn;
@@ -565,23 +571,28 @@ refint_response(
 	/* search */
 	rc = nop.o_bd->be_search(&nop, &nrs);
 
-	filter_free_x(&nop, nop.ors_filter);
-	ch_free(key);
+	ch_free( nop.ors_filterstr.bv_val );
+	while ( (fptr = ftop.f_or) != NULL ) {
+		ftop.f_or = fptr->f_next;
+		ch_free( fptr );
+	}
 	ch_free(dd.nnewdn.bv_val);
 	ch_free(dd.newdn.bv_val);
 	dd.newdn.bv_val	= NULL;
 	dd.nnewdn.bv_val = NULL;
 
 	if(rc != LDAP_SUCCESS) {
-		rs->sr_err = nrs.sr_err;
-		rs->sr_text = "refint_response search failed";
+		Debug( LDAP_DEBUG_TRACE,
+			"refint_response: search failed: %d\n",
+			rc, 0, 0 );
 		goto done;
 	}
 
 	/* safety? paranoid just in case */
 	if(!cb.sc_private) {
-		rs->sr_err = LDAP_OTHER;
-		rs->sr_text = "whoa! refint_response callback wiped out sc_private?!";
+		Debug( LDAP_DEBUG_TRACE,
+			"refint_response: callback wiped out sc_private?!\n",
+			0, 0, 0 );
 		goto done;
 	}
 
@@ -614,30 +625,26 @@ refint_response(
 	*/
 
 	for(dp = dd.mods; dp; dp = dp->next) {
-		Modifications **tail, *m;
-
-		for(m = dp->mm; m && m->sml_next; m = m->sml_next);
-		tail = &m->sml_next;
 		nop.o_req_dn	= dp->dn;
 		nop.o_req_ndn	= dp->dn;
 		nop.o_bd = select_backend(&dp->dn, 0, 1);
 		if(!nop.o_bd) {
-			rs->sr_err = LDAP_OTHER;
-			rs->sr_text = "this should never happen either!";
+			Debug( LDAP_DEBUG_TRACE,
+				"refint_response: no backend for DN %s!\n",
+				dp->dn.bv_val, 0, 0 );
 			goto done;
 		}
 		nrs.sr_type	= REP_RESULT;
 		nop.orm_modlist = dp->mm;	/* callback did all the work */
 		nop.o_dn = refint_dn;
 		nop.o_ndn = refint_dn;
-		rs->sr_err = slap_mods_opattrs( &nop, nop.orm_modlist,
-			tail, &rs->sr_text, NULL, 0, 1 );
 		nop.o_dn = nop.o_bd->be_rootdn;
 		nop.o_ndn = nop.o_bd->be_rootndn;
 		if(rs->sr_err != LDAP_SUCCESS) goto done;
 		if((rc = nop.o_bd->be_modify(&nop, &nrs)) != LDAP_SUCCESS) {
-			rs->sr_err = nrs.sr_err;
-			rs->sr_text = "dependent modify failed";
+			Debug( LDAP_DEBUG_TRACE,
+				"refint_response: dependent modify failed: %d\n",
+				nrs.sr_err, 0, 0 );
 			goto done;
 		}
 	}
@@ -646,7 +653,7 @@ done:
 	for(dp = dd.mods; dp; dp = dd.mods) {
 		dd.mods = dp->next;
 		ch_free(dp->dn.bv_val);
-		slap_mods_free(dp->mm);
+		slap_mods_free(dp->mm, 1);
 	}
 	dd.mods = NULL;
 
@@ -658,7 +665,7 @@ done:
 ** it expects to be called automagically during dynamic module initialization
 */
 
-int refint_init() {
+int refint_initialize() {
 
 	/* statically declared just after the #includes at top */
 	refint.on_bi.bi_type = "refint";
@@ -673,7 +680,7 @@ int refint_init() {
 
 #if SLAPD_OVER_REFINT == SLAPD_MOD_DYNAMIC && defined(PIC)
 int init_module(int argc, char *argv[]) {
-	return refint_init();
+	return refint_initialize();
 }
 #endif
 
