@@ -1,8 +1,8 @@
 /* referral.c - muck with referrals */
-/* $OpenLDAP: pkg/ldap/servers/slapd/referral.c,v 1.23 2004/09/26 20:58:48 ando Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/referral.c,v 1.28.2.1 2006/04/11 16:07:17 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2004 The OpenLDAP Foundation.
+ * Copyright 1998-2006 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -108,12 +108,7 @@ static char * referral_dn_muck(
 			}
 
 			muck.bv_len = ntargetDN.bv_len + nrefDN.bv_len - nbaseDN.bv_len;
-			muck.bv_val = SLAP_MALLOC( muck.bv_len + 1 );
-			if( muck.bv_val == NULL ) {
-				Debug( LDAP_DEBUG_ANY,
-					"referral_dn_muck: SLAP_MALLOC failed\n", 0, 0, 0 );
-				return NULL;
-			}
+			muck.bv_val = ch_malloc( muck.bv_len + 1 );
 
 			strncpy( muck.bv_val, ntargetDN.bv_val,
 				ntargetDN.bv_len-nbaseDN.bv_len );
@@ -153,6 +148,7 @@ int validate_global_referral( const char *url )
 
 	case LDAP_URL_ERR_BADSCHEME:
 		/* not LDAP hence valid */
+		Debug( LDAP_DEBUG_CONFIG, "referral \"%s\": not LDAP.\n", url, 0, 0 );
 		return 0;
 
 	default:
@@ -200,63 +196,65 @@ BerVarray referral_rewrite(
 	struct berval *target,
 	int scope )
 {
-	int i;
-	BerVarray refs;
-	struct berval *iv, *jv;
+	int		i;
+	BerVarray	refs;
+	struct berval	*iv, *jv;
 
-	if( in == NULL ) return NULL;
-
-	for( i=0; in[i].bv_val != NULL ; i++ ) {
-		/* just count them */
-	}
-
-	if( i < 1 ) return NULL;
-
-	refs = SLAP_MALLOC( (i+1) * sizeof( struct berval ) );
-	if( refs == NULL ) {
-		Debug( LDAP_DEBUG_ANY,
-			"referral_rewrite: SLAP_MALLOC failed\n", 0, 0, 0 );
+	if ( in == NULL ) {
 		return NULL;
 	}
 
-	for( iv=in,jv=refs; iv->bv_val != NULL ; iv++ ) {
-		LDAPURLDesc *url;
-		int rc = ldap_url_parse_ext( iv->bv_val, &url );
+	for ( i = 0; !BER_BVISNULL( &in[i] ); i++ ) {
+		/* just count them */
+	}
 
-		if( rc == LDAP_URL_ERR_BADSCHEME ) {
+	if ( i < 1 ) {
+		return NULL;
+	}
+
+	refs = ch_malloc( ( i + 1 ) * sizeof( struct berval ) );
+
+	for ( iv = in, jv = refs; !BER_BVISNULL( iv ); iv++ ) {
+		LDAPURLDesc	*url;
+		char		*dn;
+		int		rc;
+		
+		rc = ldap_url_parse_ext( iv->bv_val, &url );
+		if ( rc == LDAP_URL_ERR_BADSCHEME ) {
 			ber_dupbv( jv++, iv );
 			continue;
 
-		} else if( rc != LDAP_URL_SUCCESS ) {
+		} else if ( rc != LDAP_URL_SUCCESS ) {
 			continue;
 		}
 
-		{
-			char *dn = url->lud_dn;
-			url->lud_dn = referral_dn_muck(
-				( dn && *dn ) ? dn : NULL,
+		dn = url->lud_dn;
+		url->lud_dn = referral_dn_muck( ( dn && *dn ) ? dn : NULL,
 				base, target );
+		ldap_memfree( dn );
 
-			ldap_memfree( dn );
-		}
-
-		if( url->lud_scope == LDAP_SCOPE_DEFAULT ) {
+		if ( url->lud_scope == LDAP_SCOPE_DEFAULT ) {
 			url->lud_scope = scope;
 		}
 
 		jv->bv_val = ldap_url_desc2str( url );
-		jv->bv_len = strlen( jv->bv_val );
+		if ( jv->bv_val != NULL ) {
+			jv->bv_len = strlen( jv->bv_val );
+
+		} else {
+			ber_dupbv( jv, iv );
+		}
+		jv++;
 
 		ldap_free_urldesc( url );
-		jv++;
 	}
 
-	if( jv == refs ) {
+	if ( jv == refs ) {
 		ch_free( refs );
 		refs = NULL;
 
 	} else {
-		jv->bv_val = NULL;
+		BER_BVZERO( jv );
 	}
 
 	return refs;
@@ -284,12 +282,7 @@ BerVarray get_entry_referrals(
 
 	if( i < 1 ) return NULL;
 
-	refs = SLAP_MALLOC( (i + 1) * sizeof(struct berval));
-	if( refs == NULL ) {
-		Debug( LDAP_DEBUG_ANY,
-			"get_entry_referrals: SLAP_MALLOC failed\n", 0, 0, 0 );
-		return NULL;
-	}
+	refs = ch_malloc( (i + 1) * sizeof(struct berval));
 
 	for( iv=attr->a_vals, jv=refs; iv->bv_val != NULL; iv++ ) {
 		unsigned k;
