@@ -1,8 +1,8 @@
 /* result.c - routines to send ldap results, errors, and referrals */
-/* $OpenLDAP: pkg/ldap/servers/slapd/result.c,v 1.244.2.2 2005/02/02 17:32:44 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/result.c,v 1.289.2.1 2006/04/11 16:07:17 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2005 The OpenLDAP Foundation.
+ * Copyright 1998-2006 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -36,10 +36,7 @@
 #include <ac/unistd.h>
 
 #include "slap.h"
-
-#ifdef LDAP_SLAPI
-#include "slapi/slapi.h"
-#endif
+#include "lutil.h"
 
 const struct berval slap_dummy_bv = BER_BVNULL;
 
@@ -50,7 +47,7 @@ int slap_null_cb( Operation *op, SlapReply *rs )
 
 int slap_freeself_cb( Operation *op, SlapReply *rs )
 {
-	assert( op->o_callback );
+	assert( op->o_callback != NULL );
 
 	op->o_tmpfree( op->o_callback, op->o_tmpmemctx );
 	op->o_callback = NULL;
@@ -86,11 +83,7 @@ static char *v2ref( BerVarray ref, const char *text )
 		}
 	}
 
-	v2 = SLAP_MALLOC( len+i+sizeof("Referral:") );
-	if( v2 == NULL ) {
-		Debug( LDAP_DEBUG_ANY, "v2ref: SLAP_MALLOC failed", 0, 0, 0 );
-		return NULL;
-	}
+	v2 = ch_malloc( len+i+sizeof("Referral:") );
 
 	if( text != NULL ) {
 		strcpy(v2, text);
@@ -102,11 +95,7 @@ static char *v2ref( BerVarray ref, const char *text )
 	len += sizeof("Referral:");
 
 	for( i=0; ref[i].bv_val != NULL; i++ ) {
-		v2 = SLAP_REALLOC( v2, len + ref[i].bv_len + 1 );
-		if( v2 == NULL ) {
-			Debug( LDAP_DEBUG_ANY, "v2ref: SLAP_MALLOC failed", 0, 0, 0 );
-			return NULL;
-		}
+		v2 = ch_realloc( v2, len + ref[i].bv_len + 1 );
 		v2[len-1] = '\n';
 		AC_MEMCPY(&v2[len], ref[i].bv_val, ref[i].bv_len );
 		len += ref[i].bv_len;
@@ -119,7 +108,8 @@ static char *v2ref( BerVarray ref, const char *text )
 	return v2;
 }
 
-static ber_tag_t req2res( ber_tag_t tag )
+ber_tag_t
+slap_req2res( ber_tag_t tag )
 {
 	switch( tag ) {
 	case LDAP_REQ_ADD:
@@ -177,7 +167,7 @@ static long send_ldap_ber(
 			return 0;
 		}
 
-		if ( ber_flush( conn->c_sb, ber, 0 ) == 0 ) {
+		if ( ber_flush2( conn->c_sb, ber, LBER_FLUSH_FREE_NEVER ) == 0 ) {
 			break;
 		}
 
@@ -189,11 +179,11 @@ static long send_ldap_ber(
 		 * it's a hard error and return.
 		 */
 
-		Debug( LDAP_DEBUG_CONNS, "ber_flush failed errno=%d reason=\"%s\"\n",
+		Debug( LDAP_DEBUG_CONNS, "ber_flush2 failed errno=%d reason=\"%s\"\n",
 		    err, sock_errstr(err), 0 );
 
 		if ( err != EWOULDBLOCK && err != EAGAIN ) {
-			connection_closing( conn );
+			connection_closing( conn, "connection lost on write" );
 
 			ldap_pvt_thread_mutex_unlock( &conn->c_mutex );
 			ldap_pvt_thread_mutex_unlock( &conn->c_write_mutex );
@@ -246,48 +236,186 @@ static int
 send_ldap_controls( Operation *o, BerElement *ber, LDAPControl **c )
 {
 	int rc;
-#ifdef LDAP_SLAPI
-	LDAPControl **sctrls = NULL;
 
-	/*
-	 * Retrieve any additional controls that may be set by the
-	 * plugin.
-	 */
-
-	if ( o->o_pb && slapi_pblock_get( o->o_pb, SLAPI_RESCONTROLS, &sctrls ) != 0 ) {
-		sctrls = NULL;
-	}
-
-	if ( c == NULL && sctrls == NULL ) return 0;
-#else
-	if( c == NULL ) return 0;
-#endif /* LDAP_SLAPI */
+	if( c == NULL )
+		return 0;
 
 	rc = ber_printf( ber, "t{"/*}*/, LDAP_TAG_CONTROLS );
 	if( rc == -1 ) return rc;
 
-#ifdef LDAP_SLAPI
-	if ( c != NULL )
-#endif /* LDAP_SLAPI */
 	for( ; *c != NULL; c++) {
 		rc = send_ldap_control( ber, *c );
 		if( rc == -1 ) return rc;
 	}
 
-#ifdef LDAP_SLAPI
-	if ( sctrls != NULL ) {
-		for ( c = sctrls; *c != NULL; c++ ) {
-			rc = send_ldap_control( ber, *c );
-			if( rc == -1 ) return rc;
+#ifdef LDAP_DEVEL
+	/* this is a hack to avoid having to modify op->s_ctrls */
+	if( o->o_sortedresults ) {
+		BerElementBuffer berbuf;
+		BerElement *sber = (BerElement *) &berbuf;
+		LDAPControl sorted;
+		BER_BVZERO( &sorted.ldctl_value );
+		sorted.ldctl_oid = LDAP_CONTROL_SORTRESPONSE;
+		sorted.ldctl_iscritical = 0;
+
+		ber_init2( sber, NULL, LBER_USE_DER );
+
+		ber_printf( sber, "{e}", LDAP_UNWILLING_TO_PERFORM );
+
+		if( ber_flatten2( ber, &sorted.ldctl_value, 0 ) == -1 ) {
+			return -1;
 		}
+
+		(void) ber_free_buf( ber );
+
+		rc = send_ldap_control( ber, &sorted );
+		if( rc == -1 ) return rc;
 	}
-#endif /* LDAP_SLAPI */
+#endif
 
 	rc = ber_printf( ber, /*{*/"N}" );
 
 	return rc;
 }
 
+/*
+ * slap_response_play()
+ *
+ * plays the callback list; rationale: a callback can
+ *   - remove itself from the list, by setting op->o_callback = NULL;
+ *     malloc()'ed callbacks should free themselves from inside the
+ *     sc_response() function.
+ *   - replace itself with another (list of) callback(s), by setting
+ *     op->o_callback = a new (list of) callback(s); in this case, it
+ *     is the callback's responsibility to to append existing subsequent
+ *     callbacks to the end of the list that is passed to the sc_response()
+ *     function.
+ *   - modify the list of subsequent callbacks by modifying the value
+ *     of the sc_next field from inside the sc_response() function; this
+ *     case does not require any handling from inside slap_response_play()
+ *
+ * To stop execution of the playlist, the sc_response() function must return
+ * a value different from SLAP_SC_CONTINUE.
+ *
+ * The same applies to slap_cleanup_play(); only, there is no means to stop
+ * execution of the playlist, since all cleanup functions must be called.
+ */
+static int
+slap_response_play(
+	Operation *op,
+	SlapReply *rs )
+{
+	int rc;
+
+#ifdef LDAP_DEVEL
+	slap_callback	*sc = op->o_callback, **scp;
+
+	rc = SLAP_CB_CONTINUE;
+	for ( scp = &sc; *scp; ) {
+		slap_callback *sc_next = (*scp)->sc_next, **sc_nextp = &(*scp)->sc_next;
+
+		op->o_callback = *scp;
+		if ( op->o_callback->sc_response ) {
+			rc = op->o_callback->sc_response( op, rs );
+			if ( op->o_callback == NULL ) {
+				/* the callback has been removed;
+				 * repair the list */
+				*scp = sc_next;
+				sc_nextp = scp;
+
+			} else if ( op->o_callback != *scp ) {
+				/* a new callback has been inserted
+				 * in place of the existing one; repair the list */
+				*scp = op->o_callback;
+				sc_nextp = scp;
+			}
+			if ( rc != SLAP_CB_CONTINUE ) break;
+		}
+		scp = sc_nextp;
+	}
+
+	op->o_callback = sc;
+#else /* ! LDAP_DEVEL */
+	slap_callback	*sc = op->o_callback, **sc_prev = &sc, *sc_next;
+
+	rc = SLAP_CB_CONTINUE;
+	for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
+		sc_next = op->o_callback->sc_next;
+		if ( op->o_callback->sc_response ) {
+			slap_callback *sc2 = op->o_callback;
+			rc = op->o_callback->sc_response( op, rs );
+			if ( op->o_callback != sc2 ) {
+				*sc_prev = op->o_callback;
+			}
+			if ( rc != SLAP_CB_CONTINUE || !op->o_callback ) break;
+			if ( op->o_callback != sc2 ) continue;
+		}
+		sc_prev = &op->o_callback->sc_next;
+	}
+
+	op->o_callback = sc;
+#endif /* ! LDAP_DEVEL */
+
+	return rc;
+}
+
+static int
+slap_cleanup_play(
+	Operation *op,
+	SlapReply *rs )
+{
+#ifdef LDAP_DEVEL
+	slap_callback	*sc = op->o_callback, **scp;
+
+	for ( scp = &sc; *scp; ) {
+		slap_callback *sc_next = (*scp)->sc_next, **sc_nextp = &(*scp)->sc_next;
+
+		op->o_callback = *scp;
+		if ( op->o_callback->sc_cleanup ) {
+			(void)op->o_callback->sc_cleanup( op, rs );
+			if ( op->o_callback == NULL ) {
+				/* the callback has been removed;
+				 * repair the list */
+				*scp = sc_next;
+				sc_nextp = scp;
+
+			} else if ( op->o_callback != *scp ) {
+				/* a new callback has been inserted
+				 * after the existing one; repair the list */
+				/* a new callback has been inserted
+				 * in place of the existing one; repair the list */
+				*scp = op->o_callback;
+				sc_nextp = scp;
+			}
+			/* don't care about the result; do all cleanup */
+		}
+		scp = sc_nextp;
+	}
+
+	op->o_callback = sc;
+#else /* ! LDAP_DEVEL */
+	slap_callback	*sc = op->o_callback, **sc_prev = &sc, *sc_next;
+
+	for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
+		sc_next = op->o_callback->sc_next;
+		if ( op->o_callback->sc_cleanup ) {
+			slap_callback *sc2 = op->o_callback;
+			(void)op->o_callback->sc_cleanup( op, rs );
+			if ( op->o_callback != sc2 ) {
+				*sc_prev = op->o_callback;
+			}
+			if ( !op->o_callback ) break;
+			if ( op->o_callback != sc2 ) continue;
+		}
+		sc_prev = &op->o_callback->sc_next;
+	}
+
+	op->o_callback = sc;
+#endif /* ! LDAP_DEVEL */
+
+	return LDAP_SUCCESS;
+}
+
 static int
 send_ldap_response(
 	Operation *op,
@@ -298,26 +426,16 @@ send_ldap_response(
 	int		rc = LDAP_SUCCESS;
 	long	bytes;
 
+	if ( rs->sr_err == SLAPD_ABANDON || op->o_abandon ) {
+		rc = SLAPD_ABANDON;
+		goto clean2;
+	}
+
 	if ( op->o_callback ) {
-		int		first = 1;
-		slap_callback	*sc = op->o_callback,
-				*sc_next = op->o_callback;
-
-		rc = SLAP_CB_CONTINUE;
-		for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
-			sc_next = op->o_callback->sc_next;
-			if ( op->o_callback->sc_response ) {
-				rc = op->o_callback->sc_response( op, rs );
-				if ( first && op->o_callback == NULL ) {
-					sc = NULL;
-				}
-				if ( rc != SLAP_CB_CONTINUE ) break;
-			}
-			first = 0;
+		rc = slap_response_play( op, rs );
+		if ( rc != SLAP_CB_CONTINUE ) {
+			goto clean2;
 		}
-
-		op->o_callback = sc;
-		if ( rc != SLAP_CB_CONTINUE ) goto clean2;
 	}
 
 #ifdef LDAP_CONNECTIONLESS
@@ -442,15 +560,6 @@ send_ldap_response(
 		goto cleanup;
 	}
 
-#ifdef LDAP_SLAPI
-	if ( op->o_pb ) {
-		slapi_pblock_set( op->o_pb, SLAPI_RESULT_CODE, (void *)rs->sr_err );
-		slapi_pblock_set( op->o_pb, SLAPI_RESULT_MATCHED,
-			(void *)rs->sr_matched );
-		slapi_pblock_set( op->o_pb, SLAPI_RESULT_TEXT, (void *)rs->sr_text );
-	}
-#endif /* LDAP_SLAPI */
-
 	ldap_pvt_thread_mutex_lock( &slap_counters.sc_sent_mutex );
 	ldap_pvt_mp_add_ulong( slap_counters.sc_pdu, 1 );
 	ldap_pvt_mp_add_ulong( slap_counters.sc_bytes, (unsigned long)bytes );
@@ -464,25 +573,9 @@ cleanup:;
 
 clean2:;
 	if ( op->o_callback ) {
-		int		first = 1;
-		slap_callback	*sc = op->o_callback,
-				*sc_next = op->o_callback;
-
-		for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
-			sc_next = op->o_callback->sc_next;
-			if ( op->o_callback->sc_cleanup ) {
-				(void)op->o_callback->sc_cleanup( op, rs );
-				if ( first && op->o_callback == NULL ) {
-					sc = NULL;
-				}
-			}
-			first = 0;
-		}
-
-		op->o_callback = sc;
+		(void)slap_cleanup_play( op, rs );
 	}
 
-
 	if ( rs->sr_matched && rs->sr_flags & REP_MATCHED_MUSTBEFREED ) {
 		free( (char *)rs->sr_matched );
 		rs->sr_matched = NULL;
@@ -515,7 +608,7 @@ send_ldap_disconnect( Operation	*op, SlapReply *rs )
 
 	if ( op->o_protocol < LDAP_VERSION3 ) {
 		rs->sr_rspoid = NULL;
-		rs->sr_tag = req2res( op->o_tag );
+		rs->sr_tag = slap_req2res( op->o_tag );
 		rs->sr_msgid = (rs->sr_tag != LBER_SEQUENCE) ? op->o_msgid : 0;
 
 	} else {
@@ -541,7 +634,11 @@ slap_send_ldap_result( Operation *op, SlapReply *rs )
 
 	rs->sr_type = REP_RESULT;
 
-	assert( !LDAP_API_ERROR( rs->sr_err ));
+	/* Propagate Abandons so that cleanup callbacks can be processed */
+	if ( rs->sr_err == SLAPD_ABANDON || op->o_abandon )
+		goto abandon;
+
+	assert( !LDAP_API_ERROR( rs->sr_err ) );
 
 	Debug( LDAP_DEBUG_TRACE,
 		"send_ldap_result: %s p=%d\n",
@@ -563,11 +660,8 @@ slap_send_ldap_result( Operation *op, SlapReply *rs )
 	assert( rs->sr_err != LDAP_PARTIAL_RESULTS );
 
 	if ( rs->sr_err == LDAP_REFERRAL ) {
-#ifdef LDAP_CONTROL_X_DOMAIN_SCOPE
-		if( op->o_domain_scope ) {
-			rs->sr_ref = NULL;
-		}
-#endif
+		if( op->o_domain_scope ) rs->sr_ref = NULL;
+
 		if( rs->sr_ref == NULL ) {
 			rs->sr_err = LDAP_NO_SUCH_OBJECT;
 		} else if ( op->o_protocol < LDAP_VERSION3 ) {
@@ -575,35 +669,16 @@ slap_send_ldap_result( Operation *op, SlapReply *rs )
 		}
 	}
 
-#ifdef LDAP_SLAPI
-	/*
-	 * Call pre-result plugins. To avoid infinite recursion plugins
-	 * should just set SLAPI_RESULT_CODE rather than sending a
-	 * result if they wish to change the result.
-	 */
-	if ( op->o_pb != NULL ) {
-		slapi_int_pblock_set_operation( op->o_pb, op );
-		slapi_pblock_set( op->o_pb, SLAPI_RESULT_CODE,
-			(void *)rs->sr_err );
-		slapi_pblock_set( op->o_pb, SLAPI_RESULT_TEXT,
-			(void *)rs->sr_text );
-		slapi_pblock_set( op->o_pb, SLAPI_RESULT_MATCHED,
-			(void *)rs->sr_matched );
-
-		(void) slapi_int_call_plugins( op->o_bd, SLAPI_PLUGIN_PRE_RESULT_FN,
-			op->o_pb );
-	}
-#endif /* LDAP_SLAPI */
-
 	if ( op->o_protocol < LDAP_VERSION3 ) {
 		tmp = v2ref( rs->sr_ref, rs->sr_text );
 		rs->sr_text = tmp;
 		rs->sr_ref = NULL;
 	}
 
-	rs->sr_tag = req2res( op->o_tag );
+	rs->sr_tag = slap_req2res( op->o_tag );
 	rs->sr_msgid = (rs->sr_tag != LBER_SEQUENCE) ? op->o_msgid : 0;
 
+abandon:
 	if ( send_ldap_response( op, rs ) == SLAP_CB_CONTINUE ) {
 		if ( op->o_tag == LDAP_REQ_SEARCH ) {
 			char nbuf[64];
@@ -635,10 +710,15 @@ send_ldap_sasl( Operation *op, SlapReply *rs )
 		rs->sr_err,
 		rs->sr_sasldata ? (long) rs->sr_sasldata->bv_len : -1, NULL );
 
-	rs->sr_tag = req2res( op->o_tag );
+	rs->sr_tag = slap_req2res( op->o_tag );
 	rs->sr_msgid = (rs->sr_tag != LBER_SEQUENCE) ? op->o_msgid : 0;
 
-	send_ldap_response( op, rs );
+	if ( send_ldap_response( op, rs ) == SLAP_CB_CONTINUE ) {
+		Statslog( LDAP_DEBUG_STATS,
+			"%s RESULT tag=%lu err=%d text=%s\n",
+			op->o_log_prefix, rs->sr_tag, rs->sr_err,
+			rs->sr_text ? rs->sr_text : "", 0 );
+	}
 }
 
 void
@@ -652,10 +732,15 @@ slap_send_ldap_extended( Operation *op, SlapReply *rs )
 		rs->sr_rspoid ? rs->sr_rspoid : "",
 		rs->sr_rspdata != NULL ? rs->sr_rspdata->bv_len : 0 );
 
-	rs->sr_tag = req2res( op->o_tag );
+	rs->sr_tag = slap_req2res( op->o_tag );
 	rs->sr_msgid = (rs->sr_tag != LBER_SEQUENCE) ? op->o_msgid : 0;
 
-	send_ldap_response( op, rs );
+	if ( send_ldap_response( op, rs ) == SLAP_CB_CONTINUE ) {
+		Statslog( LDAP_DEBUG_STATS,
+			"%s RESULT oid=%s err=%d text=%s\n",
+			op->o_log_prefix, rs->sr_rspoid ? rs->sr_rspoid : "",
+			rs->sr_err, rs->sr_text ? rs->sr_text : "", 0 );
+	}
 }
 
 void
@@ -669,24 +754,35 @@ slap_send_ldap_intermediate( Operation *op, SlapReply *rs )
 		rs->sr_rspdata != NULL ? rs->sr_rspdata->bv_len : 0 );
 	rs->sr_tag = LDAP_RES_INTERMEDIATE;
 	rs->sr_msgid = op->o_msgid;
-	send_ldap_response( op, rs );
+	if ( send_ldap_response( op, rs ) == SLAP_CB_CONTINUE ) {
+		Statslog( LDAP_DEBUG_STATS2,
+			"%s INTERM oid=%s\n",
+			op->o_log_prefix,
+			rs->sr_rspoid ? rs->sr_rspoid : "", 0, 0, 0 );
+	}
 }
 
+/*
+ * returns:
+ *
+ * LDAP_SUCCESS			entry sent
+ * LDAP_OTHER			entry not sent (other)
+ * LDAP_INSUFFICIENT_ACCESS	entry not sent (ACL)
+ * LDAP_UNAVAILABLE		entry not sent (connection closed)
+ * LDAP_SIZELIMIT_EXCEEDED	entry not sent (caller must send sizelimitExceeded)
+ */
+
 int
 slap_send_search_entry( Operation *op, SlapReply *rs )
 {
 	BerElementBuffer berbuf;
 	BerElement	*ber = (BerElement *) &berbuf;
 	Attribute	*a;
-	int		i, j, rc=-1, bytes;
+	int		i, j, rc = LDAP_UNAVAILABLE, bytes;
 	char		*edn;
 	int		userattrs;
 	AccessControlState acl_state = ACL_STATE_INIT;
-#ifdef LDAP_SLAPI
-	/* Support for computed attribute plugins */
-	computed_attr_context	 ctx;
-	AttributeName	*anp;
-#endif
+	int			 attrsonly;
 	AttributeDescription *ad_entry = slap_schema.si_ad_entry;
 
 	/* a_flags: array of flags telling if the i-th element will be
@@ -694,7 +790,11 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 	 * e_flags: array of a_flags
 	 */
 	char **e_flags = NULL;
-	
+
+	if ( op->ors_slimit >= 0 && rs->sr_nentries >= op->ors_slimit ) {
+		return LDAP_SIZELIMIT_EXCEEDED;
+	}
+
 	rs->sr_type = REP_SEARCH;
 
 	/* eventually will loop through generated operational attribute types
@@ -703,7 +803,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 	/* NOTE: moved before overlays callback circling because
 	 * they may modify entry and other stuff in rs */
 	/* check for special all operational attributes ("+") type */
-	/* FIXME: maybe we could se this flag at the operation level;
+	/* FIXME: maybe we could set this flag at the operation level;
 	 * however, in principle the caller of send_search_entry() may
 	 * change the attribute list at each call */
 	rs->sr_attr_flags = slap_attr_flags( rs->sr_attrs );
@@ -714,37 +814,24 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 	}
 
 	if ( op->o_callback ) {
-		int		first = 1;
-		slap_callback	*sc = op->o_callback,
-				*sc_next = op->o_callback;
-
-		rc = SLAP_CB_CONTINUE;
-		for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
-			sc_next = op->o_callback->sc_next;
-			if ( op->o_callback->sc_response ) {
-				rc = op->o_callback->sc_response( op, rs );
-				if ( first && op->o_callback == NULL ) {
-					sc = NULL;
-				}
-				if ( rc != SLAP_CB_CONTINUE ) break;
-			}
-			first = 0;
+		rc = slap_response_play( op, rs );
+		if ( rc != SLAP_CB_CONTINUE ) {
+			goto error_return;
 		}
-
-		op->o_callback = sc;
-		if ( rc != SLAP_CB_CONTINUE ) goto error_return;
 	}
 
 	Debug( LDAP_DEBUG_TRACE, "=> send_search_entry: conn %lu dn=\"%s\"%s\n",
 		op->o_connid, rs->sr_entry->e_name.bv_val,
 		op->ors_attrsonly ? " (attrsOnly)" : "" );
 
+	attrsonly = op->ors_attrsonly;
+
 	if ( !access_allowed( op, rs->sr_entry, ad_entry, NULL, ACL_READ, NULL )) {
 		Debug( LDAP_DEBUG_ACL,
 			"send_search_entry: conn %lu access to entry (%s) not allowed\n", 
 			op->o_connid, rs->sr_entry->e_name.bv_val, 0 );
 
-		rc = 1;
+		rc = LDAP_INSUFFICIENT_ACCESS;
 		goto error_return;
 	}
 
@@ -754,12 +841,10 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 		/* read back control or LDAP_CONNECTIONLESS */
 	    ber = op->o_res_ber;
 	} else {
-		ber_len_t	siz, len;
 		struct berval	bv;
 
-		entry_flatsize( rs->sr_entry, &siz, &len, 0 );
-		bv.bv_len = siz + len;
-		bv.bv_val = op->o_tmpalloc(bv.bv_len, op->o_tmpmemctx );
+		bv.bv_len = entry_flatsize( rs->sr_entry, 0 );
+		bv.bv_val = op->o_tmpalloc( bv.bv_len, op->o_tmpmemctx );
 
 		ber_init2( ber, &bv, LBER_USE_DER );
 		ber_set_option( ber, LBER_OPT_BER_MEMCTX, &op->o_tmpmemctx );
@@ -792,6 +877,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 
 		if ( op->o_res_ber == NULL ) ber_free_buf( ber );
 		send_ldap_error( op, rs, LDAP_OTHER, "encoding DN error" );
+		rc = rs->sr_err;
 		goto error_return;
 	}
 
@@ -839,6 +925,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 				if ( op->o_res_ber == NULL ) ber_free_buf( ber );
 				send_ldap_error( op, rs, LDAP_OTHER,
 					"matched values filtering error" );
+				rc = rs->sr_err;
 				goto error_return;
 			}
 		}
@@ -849,7 +936,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 		int finish = 0;
 
 		if ( rs->sr_attrs == NULL ) {
-			/* all attrs request, skip operational attributes */
+			/* all user attrs request, skip operational attributes */
 			if( is_at_operational( desc->ad_type ) ) {
 				continue;
 			}
@@ -858,20 +945,19 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 			/* specific attrs requested */
 			if ( is_at_operational( desc->ad_type ) ) {
 				if ( !SLAP_OPATTRS( rs->sr_attr_flags ) &&
-						!ad_inlist( desc, rs->sr_attrs ) )
+					!ad_inlist( desc, rs->sr_attrs ) )
 				{
 					continue;
 				}
 
 			} else {
-				if ( !userattrs && !ad_inlist( desc, rs->sr_attrs ) )
-				{
+				if ( !userattrs && !ad_inlist( desc, rs->sr_attrs ) ) {
 					continue;
 				}
 			}
 		}
 
-		if ( op->ors_attrsonly ) {
+		if ( attrsonly ) {
 			if ( ! access_allowed( op, rs->sr_entry, desc, NULL,
 				ACL_READ, &acl_state ) )
 			{
@@ -889,6 +975,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 				if ( op->o_res_ber == NULL ) ber_free_buf( ber );
 				send_ldap_error( op, rs, LDAP_OTHER,
 					"encoding description error");
+				rc = rs->sr_err;
 				goto error_return;
 			}
 			finish = 1;
@@ -922,6 +1009,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 						if ( op->o_res_ber == NULL ) ber_free_buf( ber );
 						send_ldap_error( op, rs, LDAP_OTHER,
 							"encoding description error");
+						rc = rs->sr_err;
 						goto error_return;
 					}
 				}
@@ -933,6 +1021,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 					if ( op->o_res_ber == NULL ) ber_free_buf( ber );
 					send_ldap_error( op, rs, LDAP_OTHER,
 						"encoding values error" );
+					rc = rs->sr_err;
 					goto error_return;
 				}
 			}
@@ -945,6 +1034,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 
 			if ( op->o_res_ber == NULL ) ber_free_buf( ber );
 			send_ldap_error( op, rs, LDAP_OTHER, "encode end error" );
+			rc = rs->sr_err;
 			goto error_return;
 		}
 	}
@@ -998,6 +1088,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 				if ( op->o_res_ber == NULL ) ber_free_buf( ber );
 				send_ldap_error( op, rs, LDAP_OTHER,
 					"matched values filtering error" );
+				rc = rs->sr_err;
 				goto error_return;
 			}
 		}
@@ -1007,7 +1098,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 		AttributeDescription *desc = a->a_desc;
 
 		if ( rs->sr_attrs == NULL ) {
-			/* all attrs request, skip operational attributes */
+			/* all user attrs request, skip operational attributes */
 			if( is_at_operational( desc->ad_type ) ) {
 				continue;
 			}
@@ -1016,7 +1107,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 			/* specific attrs requested */
 			if( is_at_operational( desc->ad_type ) ) {
 				if ( !SLAP_OPATTRS( rs->sr_attr_flags ) && 
-						!ad_inlist( desc, rs->sr_attrs ) )
+					!ad_inlist( desc, rs->sr_attrs ) )
 				{
 					continue;
 				}
@@ -1047,10 +1138,11 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 			if ( op->o_res_ber == NULL ) ber_free_buf( ber );
 			send_ldap_error( op, rs, LDAP_OTHER,
 				"encoding description error" );
+			rc = rs->sr_err;
 			goto error_return;
 		}
 
-		if ( ! op->ors_attrsonly ) {
+		if ( ! attrsonly ) {
 			for ( i = 0; a->a_vals[i].bv_val != NULL; i++ ) {
 				if ( ! access_allowed( op, rs->sr_entry,
 					desc, &a->a_vals[i], ACL_READ, &acl_state ) )
@@ -1075,6 +1167,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 					if ( op->o_res_ber == NULL ) ber_free_buf( ber );
 					send_ldap_error( op, rs, LDAP_OTHER,
 						"encoding values error" );
+					rc = rs->sr_err;
 					goto error_return;
 				}
 			}
@@ -1087,50 +1180,11 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 
 			if ( op->o_res_ber == NULL ) ber_free_buf( ber );
 			send_ldap_error( op, rs, LDAP_OTHER, "encode end error" );
+			rc = rs->sr_err;
 			goto error_return;
 		}
 	}
 
-#ifdef LDAP_SLAPI
-	/*
-	 * First, setup the computed attribute context that is
-	 * passed to all plugins.
-	 */
-	if ( op->o_pb ) {
-		ctx.cac_pb = op->o_pb;
-		ctx.cac_attrs = rs->sr_attrs;
-		ctx.cac_attrsonly = op->ors_attrsonly;
-		ctx.cac_userattrs = userattrs;
-		ctx.cac_opattrs = rs->sr_attr_flags;
-		ctx.cac_acl_state = acl_state;
-		ctx.cac_private = (void *)ber;
-
-		/*
-		 * For each client requested attribute, call the plugins.
-		 */
-		if ( rs->sr_attrs != NULL ) {
-			for ( anp = rs->sr_attrs; anp->an_name.bv_val != NULL; anp++ ) {
-				rc = compute_evaluator( &ctx, anp->an_name.bv_val,
-					rs->sr_entry, slapi_int_compute_output_ber );
-				if ( rc == 1 ) break;
-			}
-		} else {
-			/*
-			 * Technically we shouldn't be returning operational attributes
-			 * when the user requested only user attributes. We'll let the
-			 * plugin decide whether to be naughty or not.
-			 */
-			rc = compute_evaluator( &ctx, "*",
-				rs->sr_entry, slapi_int_compute_output_ber );
-		}
-		if ( rc == 1 ) {
-			if ( op->o_res_ber == NULL ) ber_free_buf( ber );
-			send_ldap_error( op, rs, LDAP_OTHER, "computed attribute error" );
-			goto error_return;
-		}
-	}
-#endif /* LDAP_SLAPI */
-
 	/* free e_flags */
 	if ( e_flags ) {
 		slap_sl_free( e_flags, op->o_tmpmemctx );
@@ -1161,10 +1215,16 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 
 		if ( op->o_res_ber == NULL ) ber_free_buf( ber );
 		send_ldap_error( op, rs, LDAP_OTHER, "encode entry end error" );
-		rc = 1;
+		rc = rs->sr_err;
 		goto error_return;
 	}
 
+	if ( rs->sr_flags & REP_ENTRY_MUSTRELEASE ) {
+		be_entry_release_rw( op, rs->sr_entry, 0 );
+		rs->sr_flags ^= REP_ENTRY_MUSTRELEASE;
+		rs->sr_entry = NULL;
+	}
+
 	if ( op->o_res_ber == NULL ) {
 		bytes = send_ldap_ber( op->o_conn, ber );
 		ber_free_buf( ber );
@@ -1174,7 +1234,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 				"send_search_entry: conn %lu  ber write failed.\n", 
 				op->o_connid, 0, 0 );
 
-			rc = -1;
+			rc = LDAP_UNAVAILABLE;
 			goto error_return;
 		}
 		rs->sr_nentries++;
@@ -1187,31 +1247,16 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 	}
 
 	Statslog( LDAP_DEBUG_STATS2, "%s ENTRY dn=\"%s\"\n",
-	    op->o_log_prefix, rs->sr_entry->e_dn, 0, 0, 0 );
+	    op->o_log_prefix, edn, 0, 0, 0 );
 
 	Debug( LDAP_DEBUG_TRACE,
 		"<= send_search_entry: conn %lu exit.\n", op->o_connid, 0, 0 );
 
-	rc = 0;
+	rc = LDAP_SUCCESS;
 
 error_return:;
 	if ( op->o_callback ) {
-		int		first = 1;
-		slap_callback	*sc = op->o_callback,
-				*sc_next = op->o_callback;
-
-		for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
-			sc_next = op->o_callback->sc_next;
-			if ( op->o_callback->sc_cleanup ) {
-				(void)op->o_callback->sc_cleanup( op, rs );
-				if ( first && op->o_callback == NULL ) {
-					sc = NULL;
-				}
-			}
-			first = 0;
-		}
-
-		op->o_callback = sc;
+		(void)slap_cleanup_play( op, rs );
 	}
 
 	if ( e_flags ) {
@@ -1255,25 +1300,10 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 
 	rs->sr_type = REP_SEARCHREF;
 	if ( op->o_callback ) {
-		int		first = 1;
-		slap_callback	*sc = op->o_callback,
-				*sc_next = op->o_callback;
-
-		rc = SLAP_CB_CONTINUE;
-		for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
-			sc_next = op->o_callback->sc_next;
-			if ( op->o_callback->sc_response ) {
-				rc = op->o_callback->sc_response( op, rs );
-				if ( first && op->o_callback == NULL ) {
-					sc = NULL;
-				}
-				if ( rc != SLAP_CB_CONTINUE ) break;
-			}
-			first = 0;
+		rc = slap_response_play( op, rs );
+		if ( rc != SLAP_CB_CONTINUE ) {
+			goto rel;
 		}
-
-		op->o_callback = sc;
-		if ( rc != SLAP_CB_CONTINUE ) goto rel;
 	}
 
 	Debug( LDAP_DEBUG_TRACE,
@@ -1301,7 +1331,6 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 		goto rel;
 	}
 
-#ifdef LDAP_CONTROL_X_DOMAIN_SCOPE
 	if( op->o_domain_scope ) {
 		Debug( LDAP_DEBUG_ANY,
 			"send_search_reference: domainScope control in (%s)\n", 
@@ -1309,7 +1338,6 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 		rc = 0;
 		goto rel;
 	}
-#endif
 
 	if( rs->sr_ref == NULL ) {
 		Debug( LDAP_DEBUG_ANY,
@@ -1362,6 +1390,13 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 		goto rel;
 	}
 
+	rc = 0;
+	if ( rs->sr_flags & REP_ENTRY_MUSTRELEASE ) {
+		be_entry_release_rw( op, rs->sr_entry, 0 );
+		rs->sr_flags ^= REP_ENTRY_MUSTRELEASE;
+		rs->sr_entry = NULL;
+	}
+
 #ifdef LDAP_CONNECTIONLESS
 	if (!op->o_conn || op->o_conn->c_is_udp == 0) {
 #endif
@@ -1376,31 +1411,25 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 #ifdef LDAP_CONNECTIONLESS
 	}
 #endif
+	if ( rs->sr_ref != NULL ) {
+		int	r;
 
-	Statslog( LDAP_DEBUG_STATS2, "%s REF dn=\"%s\"\n",
-		op->o_log_prefix, rs->sr_entry ? rs->sr_entry->e_dn : "(null)",
-		0, 0, 0 );
+		for ( r = 0; !BER_BVISNULL( &rs->sr_ref[ r ] ); r++ ) {
+			Statslog( LDAP_DEBUG_STATS2, "%s REF #%d \"%s\"\n",
+				op->o_log_prefix, r, rs->sr_ref[0].bv_val,
+				0, 0 );
+		}
+
+	} else {
+		Statslog( LDAP_DEBUG_STATS2, "%s REF \"(null)\"\n",
+			op->o_log_prefix, 0, 0, 0, 0 );
+	}
 
 	Debug( LDAP_DEBUG_TRACE, "<= send_search_reference\n", 0, 0, 0 );
 
 rel:
 	if ( op->o_callback ) {
-		int		first = 1;
-		slap_callback	*sc = op->o_callback,
-				*sc_next = op->o_callback;
-
-		for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
-			sc_next = op->o_callback->sc_next;
-			if ( op->o_callback->sc_cleanup ) {
-				(void)op->o_callback->sc_cleanup( op, rs );
-				if ( first && op->o_callback == NULL ) {
-					sc = NULL;
-				}
-			}
-			first = 0;
-		}
-
-		op->o_callback = sc;
+		(void)slap_cleanup_play( op, rs );
 	}
 
 	return rc;
@@ -1411,8 +1440,7 @@ str2result(
     char	*s,
     int		*code,
     char	**matched,
-    char	**info
-)
+    char	**info )
 {
 	int	rc;
 	char	*c;
@@ -1439,8 +1467,8 @@ str2result(
 		}
 
 		if ( strncasecmp( s, "code", STRLENOF( "code" ) ) == 0 ) {
-			if ( c != NULL ) {
-				*code = atoi( c );
+			if ( c != NULL && lutil_atoi( code, c ) != 0 ) {
+				goto bailout;
 			}
 		} else if ( strncasecmp( s, "matched", STRLENOF( "matched" ) ) == 0 ) {
 			if ( c != NULL ) {
@@ -1451,6 +1479,7 @@ str2result(
 				*info = c;
 			}
 		} else {
+bailout:;
 			Debug( LDAP_DEBUG_ANY, "str2result (%s) unknown\n",
 			    s, 0, 0 );
 
@@ -1473,7 +1502,6 @@ int slap_read_controls(
 	BerElementBuffer berbuf;
 	BerElement *ber = (BerElement *) &berbuf;
 	LDAPControl c;
-	ber_len_t	siz, len;
 	Operation myop;
 
 	Debug( LDAP_DEBUG_ANY, "slap_read_controls: (%s) %s\n",
@@ -1483,8 +1511,7 @@ int slap_read_controls(
 	rs->sr_attrs = ( oid == &slap_pre_read_bv ) ?
 		op->o_preread_attrs : op->o_postread_attrs; 
 
-	entry_flatsize( rs->sr_entry, &siz, &len, 0 );
-	bv.bv_len = siz + len;
+	bv.bv_len = entry_flatsize( rs->sr_entry, 0 );
 	bv.bv_val = op->o_tmpalloc(bv.bv_len, op->o_tmpmemctx );
 
 	ber_init2( ber, &bv, LBER_USE_DER );
@@ -1494,6 +1521,8 @@ int slap_read_controls(
 	myop = *op;
 	myop.o_bd = NULL;
 	myop.o_res_ber = ber;
+	myop.o_callback = NULL;
+	myop.ors_slimit = 1;
 
 	rc = slap_send_search_entry( &myop, rs );
 	if( rc ) return rc;
@@ -1505,7 +1534,7 @@ int slap_read_controls(
 	c.ldctl_oid = oid->bv_val;
 	c.ldctl_iscritical = 0;
 
-	if ( ctrl == NULL ) {
+	if ( *ctrl == NULL ) {
 		/* first try */
 		*ctrl = (LDAPControl *) slap_sl_calloc( 1, sizeof(LDAPControl), NULL );
 	} else {
@@ -1573,8 +1602,10 @@ slap_attr_flags( AttributeName *an )
 		flags |= ( SLAP_OPATTRS_NO | SLAP_USERATTRS_YES );
 
 	} else {
-		flags |= an_find( an, &AllOper ) ?  SLAP_OPATTRS_YES : SLAP_OPATTRS_NO;
-		flags |= an_find( an, &AllUser ) ?  SLAP_USERATTRS_YES : SLAP_USERATTRS_NO;
+		flags |= an_find( an, &AllOper )
+			? SLAP_OPATTRS_YES : SLAP_OPATTRS_NO;
+		flags |= an_find( an, &AllUser )
+			? SLAP_USERATTRS_YES : SLAP_USERATTRS_NO;
 	}
 
 	return flags;
