@@ -1,8 +1,8 @@
 /* slapcommon.c - common routine for the slap tools */
-/* $OpenLDAP: pkg/ldap/servers/slapd/slapcommon.c,v 1.23.2.3 2005/01/28 17:29:49 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/slapcommon.c,v 1.73 2006/02/02 07:30:56 hyc Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2005 The OpenLDAP Foundation.
+ * Copyright 1998-2006 The OpenLDAP Foundation.
  * Portions Copyright 1998-2003 Kurt D. Zeilenga.
  * Portions Copyright 2003 IBM Corporation.
  * All rights reserved.
@@ -36,6 +36,7 @@
 
 #include "slapcommon.h"
 #include "lutil.h"
+#include "ldif.h"
 
 tool_vars tool_globals;
 
@@ -44,23 +45,25 @@ static char *leakfilename;
 static FILE *leakfile;
 #endif
 
+static LDIFFP dummy;
+
 static void
 usage( int tool, const char *progname )
 {
 	char *options = NULL;
 	fprintf( stderr,
-		"usage: %s [-v] [-c] [-d debuglevel] [-f configfile]",
+		"usage: %s [-v] [-d debuglevel] [-f configfile] [-F configdir]",
 		progname );
 
 	switch( tool ) {
 	case SLAPACL:
-		options = "\n\t[-U authcID | -D authcDN]"
-			" -b DN [attr[/access][:value]] [...]\n";
+		options = "\n\t[-U authcID | -D authcDN] [-X authzID | -o authzDN=<DN>]"
+			"\n\t-b DN -o <var>[=<val>] [-u] [attr[/access][:value]] [...]\n";
 		break;
 
 	case SLAPADD:
-		options = "\n\t[-n databasenumber | -b suffix]\n"
-			"\t[-l ldiffile] [-q] [-u] [-w]\n";
+		options = " [-c]\n\t[-g] [-n databasenumber | -b suffix]\n"
+			"\t[-l ldiffile] [-q] [-u] [-s] [-w]\n";
 		break;
 
 	case SLAPAUTH:
@@ -68,16 +71,16 @@ usage( int tool, const char *progname )
 		break;
 
 	case SLAPCAT:
-		options = "\n\t[-n databasenumber | -b suffix]"
+		options = " [-c]\n\t[-g] [-n databasenumber | -b suffix]"
 			" [-l ldiffile] [-a filter]\n";
 		break;
 
 	case SLAPDN:
-		options = " DN [...]\n";
+		options = "\n\t[-N | -P] DN [...]\n";
 		break;
 
 	case SLAPINDEX:
-		options = "\n\t[-n databasenumber | -b suffix] [-q]\n";
+		options = " [-c]\n\t[-g] [-n databasenumber | -b suffix] [-q]\n";
 		break;
 
 	case SLAPTEST:
@@ -91,6 +94,77 @@ usage( int tool, const char *progname )
 	exit( EXIT_FAILURE );
 }
 
+static int
+parse_slapacl( void )
+{
+	size_t	len;
+	char	*p;
+
+	p = strchr( optarg, '=' );
+	if ( p == NULL ) {
+		return -1;
+	}
+
+	len = p - optarg;
+	p++;
+
+	if ( strncasecmp( optarg, "sockurl", len ) == 0 ) {
+		if ( !BER_BVISNULL( &listener_url ) ) {
+			ber_memfree( listener_url.bv_val );
+		}
+		ber_str2bv( p, 0, 1, &listener_url );
+
+	} else if ( strncasecmp( optarg, "domain", len ) == 0 ) {
+		if ( !BER_BVISNULL( &peer_domain ) ) {
+			ber_memfree( peer_domain.bv_val );
+		}
+		ber_str2bv( p, 0, 1, &peer_domain );
+
+	} else if ( strncasecmp( optarg, "peername", len ) == 0 ) {
+		if ( !BER_BVISNULL( &peer_name ) ) {
+			ber_memfree( peer_name.bv_val );
+		}
+		ber_str2bv( p, 0, 1, &peer_name );
+
+	} else if ( strncasecmp( optarg, "sockname", len ) == 0 ) {
+		if ( !BER_BVISNULL( &sock_name ) ) {
+			ber_memfree( sock_name.bv_val );
+		}
+		ber_str2bv( p, 0, 1, &sock_name );
+
+	} else if ( strncasecmp( optarg, "ssf", len ) == 0 ) {
+		if ( lutil_atou( &ssf, p ) ) {
+			Debug( LDAP_DEBUG_ANY, "unable to parse ssf=\"%s\".\n", p, 0, 0 );
+			return -1;
+		}
+
+	} else if ( strncasecmp( optarg, "transport_ssf", len ) == 0 ) {
+		if ( lutil_atou( &transport_ssf, p ) ) {
+			Debug( LDAP_DEBUG_ANY, "unable to parse transport_ssf=\"%s\".\n", p, 0, 0 );
+			return -1;
+		}
+
+	} else if ( strncasecmp( optarg, "tls_ssf", len ) == 0 ) {
+		if ( lutil_atou( &tls_ssf, p ) ) {
+			Debug( LDAP_DEBUG_ANY, "unable to parse tls_ssf=\"%s\".\n", p, 0, 0 );
+			return -1;
+		}
+
+	} else if ( strncasecmp( optarg, "sasl_ssf", len ) == 0 ) {
+		if ( lutil_atou( &sasl_ssf, p ) ) {
+			Debug( LDAP_DEBUG_ANY, "unable to parse sasl_ssf=\"%s\".\n", p, 0, 0 );
+			return -1;
+		}
+
+	} else if ( strncasecmp( optarg, "authzDN", len ) == 0 ) {
+		ber_str2bv( p, 0, 1, &authzDN );
+
+	} else {
+		return -1;
+	}
+
+	return 0;
+}
 
 /*
  * slap_tool_init - initialize slap utility, handle program options.
@@ -100,6 +174,8 @@ usage( int tool, const char *progname )
  *	argc, argv	command line arguments
  */
 
+static int need_shutdown;
+
 void
 slap_tool_init(
 	const char* progname,
@@ -107,14 +183,23 @@ slap_tool_init(
 	int argc, char **argv )
 {
 	char *options;
-	char *conffile = SLAPD_DEFAULT_CONFIGFILE;
+	char *conffile = NULL;
+	char *confdir = NULL;
 	struct berval base = BER_BVNULL;
 	char *filterstr = NULL;
 	char *subtree = NULL;
 	char *ldiffile	= NULL;
+	char **debug_unknowns = NULL;
 	int rc, i, dbnum;
 	int mode = SLAP_TOOL_MODE;
 	int truncatemode = 0;
+	int use_glue = 1;
+
+#ifdef LDAP_DEBUG
+	/* tools default to "none", so that at least LDAP_DEBUG_ANY 
+	 * messages show up; use -d 0 to reset */
+	slap_debug = LDAP_DEBUG_NONE;
+#endif
 
 #ifdef CSRIMALLOC
 	leakfilename = malloc( strlen( progname ) + STRLENOF( ".leak" ) + 1 );
@@ -127,36 +212,36 @@ slap_tool_init(
 
 	switch( tool ) {
 	case SLAPADD:
-		options = "b:cd:f:l:n:qtuvw";
+		options = "b:cd:f:F:gl:n:qstuvw";
 		break;
 
 	case SLAPCAT:
-		options = "a:b:cd:f:l:n:s:v";
+		options = "a:b:cd:f:F:gl:n:s:v";
 		mode |= SLAP_TOOL_READMAIN | SLAP_TOOL_READONLY;
 		break;
 
 	case SLAPDN:
-		options = "d:f:v";
+		options = "d:f:F:NPv";
 		mode |= SLAP_TOOL_READMAIN | SLAP_TOOL_READONLY;
 		break;
 
 	case SLAPTEST:
-		options = "d:f:uv";
+		options = "d:f:F:uv";
 		mode |= SLAP_TOOL_READMAIN | SLAP_TOOL_READONLY;
 		break;
 
 	case SLAPAUTH:
-		options = "d:f:M:R:U:vX:";
+		options = "d:f:F:M:R:U:vX:";
 		mode |= SLAP_TOOL_READMAIN | SLAP_TOOL_READONLY;
 		break;
 
 	case SLAPINDEX:
-		options = "b:cd:f:n:qv";
+		options = "b:cd:f:F:gn:qv";
 		mode |= SLAP_TOOL_READMAIN;
 		break;
 
 	case SLAPACL:
-		options = "b:D:d:f:U:v";
+		options = "b:D:d:f:F:o:uU:vX:";
 		mode |= SLAP_TOOL_READMAIN | SLAP_TOOL_READONLY;
 		break;
 
@@ -180,9 +265,26 @@ slap_tool_init(
 			continuemode++;
 			break;
 
-		case 'd':	/* turn on debugging */
-			ldap_debug += atoi( optarg );
-			break;
+		case 'd': {	/* turn on debugging */
+			int	level = 0;
+
+			if ( parse_debug_level( optarg, &level, &debug_unknowns ) ) {
+				usage( tool, progname );
+			}
+#ifdef LDAP_DEBUG
+			if ( level == 0 ) {
+				/* allow to reset log level */
+				slap_debug = 0;
+
+			} else {
+				slap_debug |= level;
+			}
+#else
+			if ( level != 0 )
+				fputs( "must compile with LDAP_DEBUG for debugging\n",
+				       stderr );
+#endif
+			} break;
 
 		case 'D':
 			ber_str2bv( optarg, 0, 1, &authcDN );
@@ -192,6 +294,14 @@ slap_tool_init(
 			conffile = strdup( optarg );
 			break;
 
+		case 'F':	/* specify a conf dir */
+			confdir = strdup( optarg );
+			break;
+
+		case 'g':	/* disable subordinate glue */
+			use_glue = 0;
+			break;
+
 		case 'l':	/* LDIF file */
 			ldiffile = strdup( optarg );
 			break;
@@ -200,8 +310,30 @@ slap_tool_init(
 			ber_str2bv( optarg, 0, 0, &mech );
 			break;
 
+		case 'N':
+			if ( dn_mode && dn_mode != SLAP_TOOL_LDAPDN_NORMAL ) {
+				usage( tool, progname );
+			}
+			dn_mode = SLAP_TOOL_LDAPDN_NORMAL;
+			break;
+
 		case 'n':	/* which config file db to index */
-			dbnum = atoi( optarg ) - 1;
+			if ( lutil_atoi( &dbnum, optarg ) ) {
+				usage( tool, progname );
+			}
+			break;
+
+		case 'o':
+			if ( parse_slapacl() ) {
+				usage( tool, progname );
+			}
+			break;
+
+		case 'P':
+			if ( dn_mode && dn_mode != SLAP_TOOL_LDAPDN_PRETTY ) {
+				usage( tool, progname );
+			}
+			dn_mode = SLAP_TOOL_LDAPDN_PRETTY;
 			break;
 
 		case 'q':	/* turn on quick */
@@ -213,7 +345,10 @@ slap_tool_init(
 			break;
 
 		case 's':	/* dump subtree */
-			subtree = strdup( optarg );
+			if ( tool == SLAPADD )
+				mode |= SLAP_TOOL_NO_SCHEMA_CHECK;
+			else if ( tool == SLAPCAT )
+				subtree = strdup( optarg );
 			break;
 
 		case 't':	/* turn on truncate */
@@ -289,10 +424,13 @@ slap_tool_init(
 		break;
 	}
 
-	if ( ldiffile == NULL ) {
-		ldiffp = tool == SLAPCAT ? stdout : stdin;
+	ldap_syslog = 0;
 
-	} else if ((ldiffp = fopen( ldiffile, tool == SLAPCAT ? "w" : "r" ))
+	if ( ldiffile == NULL ) {
+		dummy.fp = tool == SLAPCAT ? stdout : stdin;
+		ldiffp = &dummy;
+
+	} else if ((ldiffp = ldif_open( ldiffile, tool == SLAPCAT ? "w" : "r" ))
 		== NULL )
 	{
 		perror( ldiffile );
@@ -303,46 +441,29 @@ slap_tool_init(
 	 * initialize stuff and figure out which backend we're dealing with
 	 */
 
-#ifdef SLAPD_MODULES
-	if ( module_init() != 0 ) {
-		fprintf( stderr, "%s: module_init failed!\n", progname );
-		exit( EXIT_FAILURE );
-	}
-#endif
-		
 	rc = slap_init( mode, progname );
-
 	if ( rc != 0 ) {
 		fprintf( stderr, "%s: slap_init failed!\n", progname );
 		exit( EXIT_FAILURE );
 	}
 
-	rc = slap_schema_init();
+	rc = read_config( conffile, confdir );
 
 	if ( rc != 0 ) {
-		fprintf( stderr, "%s: slap_schema_init failed!\n", progname );
+		fprintf( stderr, "%s: bad configuration %s!\n",
+			progname, confdir ? "directory" : "file" );
 		exit( EXIT_FAILURE );
 	}
 
-	if ( frontend_init() ) {
-		fprintf( stderr, "%s: frontend_init failed!\n", progname );
-		exit( EXIT_FAILURE );
-	}
-
-	if ( overlay_init() ) {
-		fprintf( stderr, "%s: overlay_init failed!\n", progname );
-		exit( EXIT_FAILURE );
-	}
-
-	rc = read_config( conffile, 0 );
-
-	if ( rc != 0 ) {
-		fprintf( stderr, "%s: bad configuration file!\n", progname );
-		exit( EXIT_FAILURE );
+	if ( debug_unknowns ) {
+		rc = parse_debug_unknowns( debug_unknowns, &slap_debug );
+		ldap_charray_free( debug_unknowns );
+		debug_unknowns = NULL;
+		if ( rc )
+			exit( EXIT_FAILURE );
 	}
 
 	at_oc_cache = 1;
-	ldap_syslog = 0;
 
 	switch ( tool ) {
 	case SLAPADD:
@@ -359,6 +480,16 @@ slap_tool_init(
 		break;
 	}
 
+	if ( use_glue ) {
+		rc = glue_sub_attach();
+
+		if ( rc != 0 ) {
+			fprintf( stderr,
+				"%s: subordinate configuration error\n", progname );
+			exit( EXIT_FAILURE );
+		}
+	}
+
 	rc = slap_schema_check();
 
 	if ( rc != 0 ) {
@@ -431,52 +562,63 @@ slap_tool_init(
 		/* If the named base is a glue master, operate on the
 		 * entire context
 		 */
-		if (SLAP_GLUE_INSTANCE(be)) {
+		if ( SLAP_GLUE_INSTANCE( be ) ) {
 			nosubordinates = 1;
 		}
 
 	} else if ( dbnum == -1 ) {
+		/* no suffix and no dbnum specified, just default to
+		 * the first available database
+		 */
 		if ( nbackends <= 0 ) {
 			fprintf( stderr, "No available databases\n" );
 			exit( EXIT_FAILURE );
 		}
+		LDAP_STAILQ_FOREACH( be, &backendDB, be_next ) {
+			dbnum++;
+
+			/* db #0 is cn=config, don't select it as a default */
+			if ( dbnum < 1 ) continue;
 		
-		be = &backends[dbnum=0];
+		 	if ( SLAP_MONITOR(be))
+				continue;
+
 		/* If just doing the first by default and it is a
 		 * glue subordinate, find the master.
 		 */
-		while (SLAP_GLUE_SUBORDINATE(be) || SLAP_MONITOR(be)) {
-			if (SLAP_GLUE_SUBORDINATE(be)) {
+			if ( SLAP_GLUE_SUBORDINATE(be) ) {
 				nosubordinates = 1;
+				continue;
 			}
-			be++;
-			dbnum++;
+			break;
 		}
 
-
-		if ( dbnum >= nbackends ) {
+		if ( !be ) {
 			fprintf( stderr, "Available database(s) "
 					"do not allow %s\n", progname );
 			exit( EXIT_FAILURE );
 		}
 		
-		if ( nosubordinates == 0 && dbnum > 0 ) {
+		if ( nosubordinates == 0 && dbnum > 1 ) {
 			Debug( LDAP_DEBUG_ANY,
 				"The first database does not allow %s;"
 				" using the first available one (%d)\n",
-				progname, dbnum + 1, 0 );
+				progname, dbnum, 0 );
 		}
 
 	} else if ( dbnum < 0 || dbnum > (nbackends-1) ) {
 		fprintf( stderr,
 			"Database number selected via -n is out of range\n"
-			"Must be in the range 1 to %d"
-			" (number of databases in the config file)\n",
-			nbackends );
+			"Must be in the range 0 to %d"
+			" (number of configured databases)\n",
+			nbackends-1 );
 		exit( EXIT_FAILURE );
 
 	} else {
-		be = &backends[dbnum];
+		LDAP_STAILQ_FOREACH( be, &backendDB, be_next ) {
+			if ( dbnum == 0 ) break;
+			dbnum--;
+		}
 	}
 
 startup:;
@@ -485,30 +627,44 @@ startup:;
 	mal_leaktrace(1);
 #endif
 
-	if ( !dryrun && slap_startup( be ) ) {
+	if ( conffile != NULL ) {
+		ch_free( conffile );
+	}
 
-		switch ( tool ) {
-		case SLAPTEST:
-			fprintf( stderr, "slap_startup failed "
-					"(test would succeed using "
-					"the -u switch)\n" );
-			break;
+	if ( ldiffile != NULL ) {
+		ch_free( ldiffile );
+	}
 
-		default:
-			fprintf( stderr, "slap_startup failed\n" );
-			break;
+	/* slapdn doesn't specify a backend to startup */
+	if ( !dryrun && tool != SLAPDN ) {
+		need_shutdown = 1;
+
+		if ( slap_startup( be ) ) {
+			switch ( tool ) {
+			case SLAPTEST:
+				fprintf( stderr, "slap_startup failed "
+						"(test would succeed using "
+						"the -u switch)\n" );
+				break;
+
+			default:
+				fprintf( stderr, "slap_startup failed\n" );
+				break;
+			}
+
+			exit( EXIT_FAILURE );
 		}
-		
-		exit( EXIT_FAILURE );
 	}
 }
 
 void slap_tool_destroy( void )
 {
-	if ( !dryrun && be != NULL ) {
-		slap_shutdown( be );
+	if ( !dryrun ) {
+		if ( need_shutdown ) {
+			slap_shutdown( be );
+		}
+		slap_destroy();
 	}
-	slap_destroy();
 #ifdef SLAPD_MODULES
 	if ( slapMode == SLAP_SERVER_MODE ) {
 	/* always false. just pulls in necessary symbol references. */
