@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slurpd/main.c,v 1.42.2.1 2004/12/30 21:29:59 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slurpd/main.c,v 1.48 2006/01/03 22:12:26 kurt Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2004 The OpenLDAP Foundation.
+ * Copyright 1998-2006 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -155,11 +155,21 @@ int main( int argc, char **argv )
     }
 
 #ifdef HAVE_TLS
-	if( ldap_pvt_tls_init() || ldap_pvt_tls_init_def_ctx() ) {
-		fprintf( stderr, "TLS Initialization failed.\n" );
-		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 20 );
-		rc = 1;
-		goto stop;
+	if( ldap_pvt_tls_init() || ldap_pvt_tls_init_def_ctx( 0 ) ) {
+		rc = 0;
+		/* See if we actually need TLS */
+		for ( i=0; i < sglob->num_replicas; i++ ) {
+			if ( sglob->replicas[i]->ri_tls || ( sglob->replicas[i]->ri_uri &&
+				!strncmp( sglob->replicas[i]->ri_uri, "ldaps:", 6 ))) {
+				rc = 1;
+				break;
+			}
+		}
+		if ( rc ) {
+			fprintf( stderr, "TLS Initialization failed.\n" );
+			SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 20 );
+			goto stop;
+		}
 	}
 #endif
 
@@ -205,31 +215,57 @@ int main( int argc, char **argv )
 	}
 #endif
 
-	if ( slurpd_pid_file != NULL ) {
-		FILE *fp = fopen( slurpd_pid_file, "w" );
+	/*
+	 * don't open pid/args file in one-shot mode (ITS#4152)
+	 *
+	 * bail out if files were specified but cannot be opened (ITS#4074)
+	 */
+	if ( !sglob->one_shot_mode) {
+		if ( slurpd_pid_file != NULL ) {
+			FILE *fp = fopen( slurpd_pid_file, "w" );
+
+			if ( fp == NULL ) {
+				int save_errno = errno;
+
+				fprintf( stderr, "unable to open pid file "
+					"\"%s\": %d (%s)\n",
+					slurpd_pid_file,
+					save_errno, strerror( save_errno ) );
+
+				free( slurpd_pid_file );
+				slurpd_pid_file = NULL;
+
+				rc = 1;
+				goto stop;
+			}
 
-		if( fp != NULL ) {
 			fprintf( fp, "%d\n", (int) getpid() );
 			fclose( fp );
-
-		} else {
-		free(slurpd_pid_file);
-		slurpd_pid_file = NULL;
 		}
-	}
 
-	if ( slurpd_args_file != NULL ) {
-		FILE *fp = fopen( slurpd_args_file, "w" );
+		if ( slurpd_args_file != NULL ) {
+			FILE *fp = fopen( slurpd_args_file, "w" );
+
+			if ( fp == NULL ) {
+				int save_errno = errno;
+
+				fprintf( stderr, "unable to open args file "
+					"\"%s\": %d (%s)\n",
+					slurpd_args_file,
+					save_errno, strerror( save_errno ) );
+
+				free( slurpd_args_file );
+				slurpd_pid_file = NULL;
+
+				rc = 1;
+				goto stop;
+			}
 
-		if( fp != NULL ) {
 			for ( i = 0; i < argc; i++ ) {
 				fprintf( fp, "%s ", argv[i] );
 			}
 			fprintf( fp, "\n" );
 			fclose( fp );
-		} else {
-			free(slurpd_args_file);
-			slurpd_args_file = NULL;
 		}
 	}
 
